# NESGuard SRI Monitor Module for Drupal

## Overview

This module integrates your Drupal site with the NESGuard.cloud service to provide enhanced security for external JavaScript assets using Subresource Integrity (SRI) and real-time monitoring via WebSockets. It also includes optional Content Security Policy (CSP) header management.

**Key Features:**

- Automatically fetches monitored script lists and their corresponding SRI hashes from your NESGuard account via API.
- Applies `integrity` and `crossorigin` attributes to monitored scripts loaded via Drupal's Library API (`hook_library_info_alter`).
- Establishes a WebSocket connection to NESGuard.cloud for real-time hash update notifications.
- Client-side JavaScript updates SRI attributes instantly when NESGuard detects a hash change.
- Includes fallback mechanisms for Stripe JS (`https://js.stripe.com/v3`) monitoring and hash calculation if the NESGuard API is temporarily unavailable (requires an active subscription).
- Applies initial SRI hashes based on backend data for immediate protection upon page load.
- Optionally manages and enforces Content Security Policy (CSP) headers.
- Uses a service-based architecture following Drupal best practices.

## Requirements

- Drupal 9.4 or higher / Drupal 10+.
- PHP 7.4 or higher (check dependencies like Guzzle for specific requirements).
- A NESGuard.cloud account with an active subscription.
- Your NESGuard Agency Identifier and API Key.
- PHP `openssl` extension enabled (for API Key encryption).
- Ability for your Drupal server to make outbound HTTPS requests to `https://nesguard.cloud` and `wss://nesguard.cloud`.
- Ability for your Drupal server to make outbound HTTPS requests to `https://js.stripe.com` (for the fallback mechanism).

## Installation

1.  Place the module directory (`nesguard_drupal`) in your Drupal installation's modules directory (e.g., `modules/custom/nesguard_drupal`).
2.  **Crucially: Set Encryption Key.** Before enabling the module or saving credentials, you MUST add a secure, random encryption key to your site's `settings.php` file. This key is used to encrypt your NESGuard API Key in the Drupal configuration. **Protect this key carefully.**
    - Generate a strong key (e.g., using `php -r 'echo base64_encode(random_bytes(32));'` on the command line).
    - Add the following line to your `settings.php` (typically located at `sites/default/settings.php`), replacing `YOUR_SECURE_RANDOM_KEY_HERE` with the key you generated:
      ```php
      $settings['nesguard_encryption_key'] = 'YOUR_SECURE_RANDOM_KEY_HERE';
      ```
    - **Security Warning:** Ensure `settings.php` has appropriate file permissions and is not publicly accessible or committed to version control with the key value exposed. Consider using environment variables or other secure methods to manage this key in production environments.
3.  Enable the module using the Drupal UI (`/admin/modules`) or Drush (`drush en nesguard_drupal`).
4.  Clear Drupal's cache (`drush cr` or via UI at `/admin/config/development/performance`).

## Configuration

1.  Navigate to the NESGuard settings page: Administration > Configuration > System > NESGuard SRI Settings (`/admin/config/system/nesguard-sri`).
2.  **Enable NESGuard Integration:** Check this box to activate the module's core functionality.
3.  **API Credentials:**
    - **NESGuard Account Email (Reference):** Enter the email associated with your NESGuard account (for reference purposes).
    - **Agency Identifier:** Enter your unique NESGuard Agency Identifier (e.g., "NES-SIMON").
    - **Enter/Update API Key:** Enter your NESGuard API Key. It will be encrypted upon saving using the key from your `settings.php`. Leave blank if the key is already configured and you don't need to change it.
4.  **SRI & Subscription Status:** This section shows the status detected via the API (requires credentials to be saved and valid). SRI protection is automatically enabled only if the module is enabled AND an active subscription is detected. The list of monitored scripts fetched from the API (or the Stripe fallback) is displayed here.
5.  **Content Security Policy (CSP) Management (Optional):**
    - **Warning:** Only enable this if you are _not_ managing CSP through other means (like the `csp` module, server headers, or meta tags), as conflicts can occur.
    - **Warning:** Only enable this if you are _not_ managing CSP through other means (like the `csp` module, server headers, or meta tags), as conflicts can occur.
    - Check **Enable CSP Header Management...** to have this module add CSP headers.
    - Configure directives using the checkboxes for common services (Google Fonts, Stripe, NESGuard) and the textareas for additional custom sources.
6.  **Advanced Module Settings:**
    - **API Data Cache Lifetime:** How long Drupal caches results from NESGuard API calls (subscription status, script list/hashes). Default is 1 hour.
    - **Cron Cache Refresh Interval:** How often the Drupal cron job forces a refresh of the cached API data. Default is 6 hours.
    - **Enable Debug Mode:** Check to log detailed information (including potentially sensitive data like URLs, non-redacted tokens in some messages - use with caution) to Drupal logs and the browser console.
7.  **Save configuration.**

## How it Works

1.  **SRI Application (Server-Side):** For scripts loaded via Drupal libraries, `hook_library_info_alter` uses the `NesguardSriService` to get the latest known hash (from API cache or Stripe fallback) and adds `integrity`/`crossorigin` attributes to the library definition before the page is rendered.
2.  **Settings & Fallback Data:** `hook_page_attachments` fetches required configuration (auth token, agency ID, etc.) and the latest known script/hash list (including Stripe fallback if needed) via services and passes it securely to the client-side JavaScript via `drupalSettings`.
3.  **Client-Side JS:** An external JS file (`js/nesguard-client.js`), loaded via the Drupal Library API:
    - Applies initial SRI attributes based on the data received in `drupalSettings`.
    - Connects to the NESGuard WebSocket using the provided token and identifiers.
    - Listens for an initial `welcome` message containing the authoritative script/hash list from the server, replaces its internal data, and re-applies SRI attributes to ensure accuracy.
    - Listens for real-time `hash_change` messages and instantly updates the `integrity` attributes on the relevant script tags in the page's DOM.
    - Includes logic for WebSocket reconnection and uses the last known good data if the connection fails.
4.  **CSP (Optional):** If enabled, `hook_page_attachments` constructs and adds the configured `Content-Security-Policy` or `Content-Security-Policy-Report-Only` header.

## Troubleshooting

- **SRI Attributes Not Appearing:**
  - Verify the module is enabled globally AND the "Enable NESGuard Integration" checkbox is checked in settings.
  - Verify API Credentials (Agency ID, API Key) are correct in settings.
  - Verify the `nesguard_encryption_key` is correctly set in `settings.php` _before_ saving the API key.
  - Check the "Subscription Status" on the module settings page. SRI is only active if the subscription is active.
  - Check Drupal logs (`/admin/reports/dblog`) for errors from `NesguardAuthService`, `NesguardSubscriptionService`, or `NesguardSriService`. Enable Debug Mode for more detail.
  - Ensure the script you expect to have SRI is actually being monitored according to the list displayed on the settings page (or the Stripe fallback is active).
  - If the script is loaded via a library, clear Drupal's caches thoroughly (`drush cr`). Library info is heavily cached.
- **JavaScript Not Loading / `drupalSettings` Error:**
  - Ensure the module is fully enabled.
  - Check PHP logs for errors during page load, especially in `hook_page_attachments`.
  - Clear Drupal caches (`drush cr`).
  - Verify `nesguard_drupal.libraries.yml` syntax is correct.
  - Verify `js/nesguard-client.js` exists and has correct permissions.
  - Temporarily disable JS aggregation (`/admin/config/development/performance`) to see if the individual file loads. If it does, check `js/nesguard-client.js` for syntax errors using a linter.
  - Ensure your theme renders `{{ page_bottom }}`.
- **WebSocket Not Connecting:**
  - Check browser console logs (enable Debug Mode for more detail).
  - Verify `accessToken`, `agencyId`, and `clientId` are present in `drupalSettings.nesguard` in the page source.
  - Check browser Network tab (filter by WS) for connection attempts and status codes.
  - Ensure no firewalls are blocking outbound WSS connections from the browser to `wss://nesguard.cloud`.
- **CSP Blocking Resources:**
  - If you enabled CSP Management, use "Report-Only Mode" first to identify blocked resources.
  - Add necessary sources via the checkboxes or "Additional Custom Sources" textareas in the module settings.
  - Consider conflicts if another system (module, server header) is also setting CSP headers. It's usually best to have only one system manage CSP.

## Security Considerations

- **API Key:** Treat your NESGuard API Key like a password. It grants access to the NESGuard API.
- **Token Storage (State API):** This module stores the short-lived access token and the potentially longer-lived refresh token obtained from the NESGuard API in Drupal's State API. While convenient, State API is typically stored in the database unencrypted. For environments requiring higher security for session/API tokens, consider modifying the `NesguardAuthService` to integrate with Drupal's Key module and Encrypt module (or Lockr) to store these tokens in a more secure, encrypted manner.
- **JavaScript (`js/nesguard-client.js`):** The client-side code handles data received from the WebSocket. While care has been taken, review this code if you have extreme security requirements regarding client-side execution. It primarily manipulates `integrity` and `crossorigin` attributes based on data originating from the authenticated WebSocket connection or backend settings.
- **CSP:** If using the module's CSP management, ensure your policy is appropriately strict but allows all necessary resources for your site to function. Test thoroughly using Report-Only mode.
