/**
 * @file
 * NesGuard client-side script for SRI monitoring and WebSocket communication.
 *
 * Handles applying initial SRI attributes based on drupalSettings, connecting
 * to the NesGuard WebSocket server, receiving real-time hash updates, and
 * dynamically updating SRI attributes on script tags in the DOM. Includes
 * logic for reconnection and error handling.
 *
 * Expects configuration via drupalSettings.nesguard.
 */

(function ($, Drupal, drupalSettings) {
  "use strict";

  // --- Early Exit Conditions ---
  if (typeof Drupal === 'undefined' || typeof $ === 'undefined') {
    // Don't log error - this could happen during normal page load sequencing
    return;
  }
  
  if (!drupalSettings || !drupalSettings.nesguard) {
    console.error(
      "[NESGuard] drupalSettings.nesguard is missing or invalid. Cannot initialize."
    );
    return;
  }

  const config = drupalSettings.nesguard;
  // Ensure essential config values exist, providing defaults where sensible.
  config.version = config.version || "unknown";
  config.debug = config.debug === true; // Explicit boolean check.
  config.sriEnabled = config.sriEnabled === true; // Explicit boolean check.
  config.enforceSRI = config.enforceSRI === true; // Explicit boolean check.
  // Add fallbacks for failed script loading
  config.fallbacks = config.fallbacks || {};
  // Initialize hashes map from drupalSettings, ensure it's a Map.
  // The map stores { currentHash: string, previousHash: string|null }.
  config.hashes = new Map(Object.entries(config.hashes || {}));

  // --- Global State ---
  /** @type {WebSocket|null} The WebSocket connection instance. */
  let ws = null;
  /** @type {number|null} Timeout ID for scheduled reconnection. */
  let reconnectTimeout = null;
  /** @type {number} Counter for reconnection attempts. */
  let reconnectAttempts = 0;
  /** @type {boolean} Flag indicating if initial data was received from WS. */
  let initialDataReceived = false;
  /** @type {boolean} Flag for tracking initialization state. */
  let initialized = false;
  /** @type {Object} Cache for HTTP fallback data. */
  let httpFallbackCache = {
    lastCheck: 0,
    data: null
  };

  // --- Helper Functions ---

  /**
   * Checks if a URL is likely a Stripe JS URL.
   *
   * @param {string|null} url The URL to check.
   * @returns {boolean} True if the URL includes 'js.stripe.com'.
   */
  const isStripeUrl = (url) =>
    typeof url === "string" && url.toLowerCase().includes("js.stripe.com");

  /**
   * Function to identify critical scripts (currently only Stripe).
   * Used to dispatch specific events for these scripts.
   *
   * @param {string} url The URL to check.
   * @returns {boolean} True if the script is considered critical.
   */
  const CRITICAL_SCRIPTS_CHECK = isStripeUrl;
  
  /**
   * Determines if SRI should be enforced for a specific URL.
   * In monitor-only mode, critical scripts like Stripe are never blocked.
   *
   * @param {string} url The URL to check.
   * @returns {boolean} True if SRI should be enforced for this URL.
   */
  const shouldEnforceSriForUrl = (url) => {
    // Global enforcement setting from Drupal config
    const globalEnforce = config.enforceSRI === true;
    
    // Check if this is a critical script (like Stripe JS)
    if (CRITICAL_SCRIPTS_CHECK(url)) {
      if (config.debug) {
        console.log(`[NESGuard SRI] ${globalEnforce ? 'Enforcing' : 'Not enforcing'} SRI for critical script: ${url}`);
      }
      return globalEnforce; // Use configured setting for critical scripts
    }
    
    // Always enforce SRI for non-critical scripts when SRI is enabled
    return config.sriEnabled === true;
  };

  /**
   * Normalizes an SRI hash string to 'sha384-...' format.
   *
   * @param {string|null} hash The hash string (potentially with or without prefix).
   * @returns {string|null} The normalized hash string or null if invalid.
   */
  function normalizeHash(hash) {
    if (typeof hash !== "string" || !hash) {
      return null;
    }
    hash = hash.trim();
    if (hash.startsWith("sha384-")) {
      // Verify the length and characters of the base64 part.
      if (/^sha384-[A-Za-z0-9+\/=]{60,}$/.test(hash)) { // More flexible regex
        return hash;
      }
    }
    // Check if it's just the base64 part (missing prefix).
    else if (/^[A-Za-z0-9+\/=]{60,}$/.test(hash)) { // More flexible regex
      return "sha384-" + hash; // Add prefix.
    }
    // Invalid format.
    if (config.debug) {
      console.warn("[NESGuard] normalizeHash received invalid format:", hash);
    }
    return null;
  }

  /**
   * Updates SRI attributes ('integrity', 'crossorigin') on script tags in the DOM.
   *
   * Finds script tags matching the base URL and applies the current hash,
   * plus the previous hash if different and available. No-op if SRI is disabled.
   *
   * @param {string} url The normalized URL of the script whose tags should be updated.
   * @param {{currentHash: string, previousHash: string|null}} hashData
   * An object containing the current and optional previous normalized hashes.
   * @returns {boolean} True if any script tag attributes were modified.
   */
  let updateScriptIntegrity = function updateScriptIntegrityNoOp() {
    // Default no-op function if SRI is disabled.
    return false;
  };

  if (config.sriEnabled) {
    // Redefine only if SRI is enabled.
    updateScriptIntegrity = function updateScriptIntegrityEnabled(
      url,
      hashData
    ) {
      if (
        typeof url !== "string" ||
        !url ||
        typeof hashData !== "object" ||
        !hashData
      ) {
        if (config.debug) {
          console.warn(
            "[NESGuard SRI] Invalid input to updateScriptIntegrity:",
            url,
            hashData
          );
        }
        return false;
      }

      const currentHash = normalizeHash(hashData.currentHash);
      const previousHash = normalizeHash(hashData.previousHash);
      // A previous hash is valid only if normalized AND different from current.
      const validPreviousHash =
        previousHash && previousHash !== currentHash ? previousHash : null;

      if (!currentHash) {
        if (config.debug) {
          console.warn(
            "[NESGuard SRI] Invalid current hash provided for url:",
            url,
            hashData
          );
        }
        return false; // Cannot proceed without a valid current hash.
      }

      if (config.debug) {
        console.log(
          `[NESGuard SRI] updateScriptIntegrity called for: ${url}, Current: ${currentHash}, Previous: ${
            validPreviousHash || "N/A"
          }`
        );
      }

      // Select scripts based on the 'src' attribute starting with the URL's base
      // (protocol, host, path without query/fragment).
      const baseUrl = url.split(/[?#]/)[0];
      let scripts;
      try {
        // CSS selector needs proper escaping for special characters in URLs
        // Create a safe version of baseUrl for use in querySelector
        const escapedBaseUrl = baseUrl.replace(/[.*+?^${}()|[\]\\]/g, '\\$&');
        scripts = document.querySelectorAll(`script[src^="${escapedBaseUrl}"]`);
      } catch (e) {
        console.error(
          `[NESGuard SRI] Error selecting script tags for base URL ${baseUrl}:`,
          e
        );
        return false;
      }

      if (!scripts || scripts.length === 0) {
        // Common case: script not found (yet) or removed.
        if (config.debug) {
          console.log(
            `[NESGuard SRI] No script tags found matching base URL: ${baseUrl}`
          );
        }
        return false;
      }

      let updated = false;
      scripts.forEach((script) => {
        const currentIntegrityAttr = script.getAttribute("integrity") || "";
        const currentCrossOrigin = script.getAttribute("crossorigin");
        const currentIntegrityMode = script.getAttribute("data-sri-mode") || "enforce";

        // Check if this URL should have enforced SRI or monitor-only
        const enforceForThisUrl = shouldEnforceSriForUrl(url);
        const targetSriMode = enforceForThisUrl ? "enforce" : "monitor";
        
        // Construct the target integrity value with maximum security
        // For critical scripts, we'll use only the most recent hash to ensure validation
        // This reduces the chance of hash mismatch during rapid updates
        let targetIntegrityValue = currentHash;
        
        // Only include the previous hash if it's provided and is different
        // This helps during transition periods when servers might serve either version
        if (validPreviousHash) {
          targetIntegrityValue += " " + validPreviousHash;
        }

        // Determine if update is needed.
        const integrityNeedsUpdate =
          !currentIntegrityAttr.includes(currentHash) ||
          currentIntegrityAttr !== targetIntegrityValue;
        const crossoriginNeedsUpdate = currentCrossOrigin !== "anonymous";
        const modeNeedsUpdate = currentIntegrityMode !== targetSriMode;
        const needsUpdate = integrityNeedsUpdate || crossoriginNeedsUpdate || modeNeedsUpdate;

        if (config.debug) {
          console.log(
            `[NESGuard SRI] Script ${
              script.src
            }: needsUpdate = ${needsUpdate} (Integrity current: "${currentIntegrityAttr}", target: "${targetIntegrityValue}", CrossOrigin: "${
              currentCrossOrigin || "none"
            }", Mode: "${currentIntegrityMode}" → "${targetSriMode}")`
          );
        }

        if (needsUpdate) {
          if (config.debug) {
            console.log(
              `[NESGuard SRI] Applying/Updating attributes for ${script.src}. New integrity: "${targetIntegrityValue}", Mode: ${targetSriMode}`
            );
          }
          try {
            // Store the integrity value in a data attribute for reference/debugging
            script.setAttribute("data-integrity-value", targetIntegrityValue);
            
            // For critical scripts in monitor-only mode, we might skip the integrity attribute
            // This allows scripts like Stripe to load even during hash transitions
            if (enforceForThisUrl) {
              script.setAttribute("integrity", targetIntegrityValue);
            } else if (script.hasAttribute("integrity")) {
              // If we're in monitor-only mode for this URL, remove any integrity attribute
              script.removeAttribute("integrity");
              if (config.debug) {
                console.log(`[NESGuard SRI] Monitor-only mode: Removed integrity for ${url}`);
              }
            }
            
            // Store original hash for potential recovery if needed
            if (CRITICAL_SCRIPTS_CHECK(url)) {
              script.setAttribute("data-original-integrity", targetIntegrityValue);
            }
            
            script.setAttribute("crossorigin", "anonymous");
            script.setAttribute("data-sri-mode", targetSriMode);
            // Add a timestamp attribute for debugging client-side updates.
            script.setAttribute("data-sri-update-time", Date.now().toString());
            
            // Add error handler for critical scripts (fallback mechanism)
            if (CRITICAL_SCRIPTS_CHECK(url)) {
              if (!script.hasAttribute('onerror')) {
                script.setAttribute('onerror', 'if(window.NESGuard_handleScriptError)window.NESGuard_handleScriptError(this)');
                script.onerror = function() { 
                  if (window.NESGuard_handleScriptError) {
                    window.NESGuard_handleScriptError(this);
                  }
                };
              }
            }
            
            updated = true;

            // Dispatch custom events.
            try {
              document.dispatchEvent(
                new CustomEvent("nesguard-hash-updated", {
                  detail: {
                    url: url,
                    newHash: currentHash,
                    oldHash: currentIntegrityAttr,
                    timestamp: Date.now(),
                    enforcedMode: enforceForThisUrl
                  },
                })
              );
            } catch (e) {
              if (config.debug) {
                console.warn(`[NESGuard SRI] Error dispatching hash-updated event: ${e.message}`);
              }
            }
            if (CRITICAL_SCRIPTS_CHECK(url)) {
              try {
                document.dispatchEvent(
                  new CustomEvent("nesguard-critical-update", {
                    detail: {
                      url: url,
                      newHash: currentHash,
                      oldHash: currentIntegrityAttr,
                      timestamp: Date.now(),
                      enforcedMode: enforceForThisUrl
                    },
                  })
                );
              } catch (e) {
                if (config.debug) {
                  console.warn(`[NESGuard SRI] Error dispatching critical-update event: ${e.message}`);
                }
              }
            }
          } catch (e) {
            console.error(
              "[NESGuard SRI] Error setting attributes for:",
              url,
              e
            );
          }
        }
      }); // End forEach script

      // Update internal hash map in config if needed, ensuring the map exists.
      if (config.hashes && (config.hashes.has(url) || updated)) {
        const data = config.hashes.get(url) || {}; // Get existing or new object.
        // Update only if hashes actually changed.
        if (
          data.currentHash !== currentHash ||
          data.previousHash !== validPreviousHash
        ) {
          data.currentHash = currentHash;
          data.previousHash = validPreviousHash; // Store normalized previous hash (or null).
          config.hashes.set(url, data); // Put back into map.
          if (config.debug) {
            console.log(`[NESGuard SRI] Updated internal hash map for ${url}`);
          }
        }
      }
      return updated; // Return true if any script tag was modified.
    };
  } // end if (config.sriEnabled).

  // --- WebSocket Handling ---

  /**
   * Establishes and manages the WebSocket connection to NesGuard.
   * Handles messages, errors, closure, and heartbeat.
   *
   * @returns {WebSocket|null} The WebSocket instance or null if connection failed.
   */
  async function connectToNesGuardWebSocket() {
    // Check prerequisites for connection.
    if (
      !config.agencyId ||
      !config.accessToken ||
      !config.clientId ||
      !config.wsEndpoint
    ) {
      if (config.debug) {
        console.log(
          "[NESGuard WS] Cannot connect: Missing required configuration (AgencyID, AccessToken, ClientID, or WSEndpoint)."
        );
      }
      // Do not schedule reconnect here; let initialization logic handle it if needed.
      return null;
    }

    // --- Close Existing Connection ---
    if (ws) {
      try {
        ws.close(1000, "Client initiated reconnect"); // Use normal closure code.
        if (config.debug) {
          console.log(
            "[NESGuard WS] Closed existing WebSocket connection for reconnect."
          );
        }
      } catch (e) {
        /* Ignore errors closing already closed/closing socket */
      }
      ws = null; // Ensure ws is nullified.
    }
    // Clear any pending reconnect timeout.
    if (reconnectTimeout) {
      clearTimeout(reconnectTimeout);
      reconnectTimeout = null;
    }
    initialDataReceived = false; // Reset flag on new connection attempt.

    // --- Construct Connection URL ---
    let connectionUrl = `${config.wsEndpoint}?token=${encodeURIComponent(
      config.accessToken
    )}&agency=${encodeURIComponent(
      config.agencyId
    )}&client=${encodeURIComponent(config.clientId)}`;
    connectionUrl += `&v=${encodeURIComponent(config.version)}`; // Add client version.

    if (config.debug) {
      console.log(
        "[NESGuard WS] Connecting:",
        connectionUrl.replace(/token=.*?(&|$)/, "token=REDACTED$1")
      );
    }

    // --- Create WebSocket ---
    try {
      const socket = new WebSocket(connectionUrl);
      let heartbeatInterval = null;
      ws = socket; // Assign the new socket to the module-scoped variable.

      /** On WebSocket Open */
      socket.onopen = function onWsOpen() {
        if (config.debug) {
          console.log("[NESGuard WS] WebSocket Connected.");
        }
        reconnectAttempts = 0; // Reset attempts on successful connection.
        if (config.debug) {
          console.log(
            "[NESGuard WS] Waiting for 'welcome' message with initial script data."
          );
        }

        // Start heartbeat (ping) mechanism.
        heartbeatInterval = setInterval(() => {
          if (socket.readyState === WebSocket.OPEN) {
            try {
              socket.send(
                JSON.stringify({ type: "ping", timestamp: Date.now() })
              );
            } catch (e) {
              console.error("[NESGuard WS] Error sending ping:", e);
              // Consider closing and triggering reconnect if ping fails.
              clearInterval(heartbeatInterval);
              heartbeatInterval = null;
              socket.close(1011, "Ping send error"); // Internal Error.
              
              // If WebSocket fails, initiate an immediate HTTP fallback check
              // This ensures we can still get fresh hashes even with broken websocket
              if (typeof window.fetchLatestHashesViaHttp === 'function') {
                window.fetchLatestHashesViaHttp();
              }
            }
          } else {
            // Stop interval if socket is no longer open.
            clearInterval(heartbeatInterval);
            heartbeatInterval = null;
          }
        }, 30000); // Send ping every 30 seconds.
      }; // End onopen

      /** On WebSocket Message Received */
      socket.onmessage = function onWsMessage(event) {
        let message;
        try {
          message = JSON.parse(event.data);
          // Basic validation of message structure.
          if (!message || typeof message !== "object" || !message.type) {
            if (config.debug) {
              console.warn(
                "[NESGuard WS] Received invalid message format:",
                event.data
              );
            }
            return;
          }

          // Log received message in debug mode (excluding noisy pongs).
          if (config.debug && message.type !== "pong") {
            console.log("[NESGuard WS] Message:", message);
          }

          // --- Handle Specific Message Types ---

          // Handle Authentication Errors specifically.
          if (
            message.type === "error" &&
            (message.code === 401 ||
              message.code === 403 ||
              message.message?.toLowerCase().includes("auth"))
          ) {
            console.warn(
              "[NESGuard WS] WebSocket Authentication error:",
              message.message || `Code ${message.code}`
            );
            
            // Check if token is expired and attempt refresh
            const isTokenExpired = 
              message.message?.toLowerCase().includes("token expired") || 
              message.message?.toLowerCase().includes("refresh_token") ||
              message.code === 401;
              
            if (isTokenExpired) {
              if (config.debug) {
                console.log("[NESGuard WS] Token appears expired, attempting to refresh...");
              }
              
              // Attempt to get a new token via the API
              refreshTokenAndReconnect();
              return;
            }
            
            // For other auth errors (not expiration), close the connection
            socket.close(1008, "Authentication Failed");
            if (reconnectTimeout) {
              clearTimeout(reconnectTimeout);
              reconnectTimeout = null;
            } // Stop trying.
            document.dispatchEvent(new CustomEvent("nesguard-auth-failed")); // Notify app.
            return;
          }

          // Handle initial data dump ('welcome' message).
          if (
            message.type === "welcome" &&
            message.scripts &&
            config.sriEnabled
          ) {
            if (config.debug) {
              console.log(
                "[NESGuard WS] Received 'welcome' message with script data."
              );
            }
            initialDataReceived = true; // Flag that initial state is received.

            if (Array.isArray(message.scripts)) {
              const newHashMap = new Map();
              let receivedCount = 0;
              message.scripts.forEach((scriptInfo) => {
                // Validate entry structure.
                if (!scriptInfo?.url || !scriptInfo.currentHash) {
                  if (config.debug) {
                    console.warn(
                      "[NESGuard WS] Skipping invalid script entry in 'welcome' data:",
                      scriptInfo
                    );
                  }
                  return; // Skip this entry.
                }

                const url = scriptInfo.url; // Assume normalized from server.
                const currentHash = scriptInfo.currentHash;
                const previousHash = scriptInfo.previousHash || null;

                // Normalize received hashes.
                const normCurrentHash = normalizeHash(currentHash);
                const normPreviousHash = normalizeHash(previousHash);

                if (url && normCurrentHash) {
                  newHashMap.set(url, {
                    currentHash: normCurrentHash,
                    previousHash:
                      normPreviousHash && normPreviousHash !== normCurrentHash
                        ? normPreviousHash
                        : null,
                  });
                  receivedCount++;
                } else if (config.debug) {
                  console.warn(
                    "[NESGuard WS] Invalid URL/Hash after normalization in 'welcome' data:",
                    scriptInfo
                  );
                }
              }); // End foreach scriptInfo

              // Replace the entire config hash map with the fresh data.
              config.hashes = newHashMap;
              if (config.debug) {
                console.log(
                  `[NESGuard WS] Updated hash map with ${receivedCount} scripts from 'welcome' message.`
                );
              }

              // Re-apply SRI based on the new map from 'welcome'.
              if (config.debug) {
                console.log(
                  `[NESGuard WS Debug] Re-applying SRI based on 'welcome' data. Map size: ${config.hashes.size}`
                );
              }
              let reappliedCount = 0;
              for (const [entryUrl, entryHashData] of config.hashes) {
                if (updateScriptIntegrity(entryUrl, entryHashData)) {
                  reappliedCount++;
                }
              }
              if (config.debug) {
                console.log(
                  `[NESGuard WS Debug] Finished re-applying SRI based on 'welcome' data. Updates applied to ${reappliedCount} script instances.`
                );
              }
            } else {
              console.warn(
                "[NESGuard WS] Received 'welcome' message but 'scripts' format unexpected:",
                message
              );
            }
          } // End 'welcome' handling

          // Handle subsequent hash changes ('hash_change' message).
          else if (config.sriEnabled && message.type === "hash_change") {
            // Validate required fields.
            if (!message.url || !message.new_hash) {
              if (config.debug) {
                console.warn(
                  "[NESGuard WS] Received 'hash_change' with missing url or new_hash:",
                  message
                );
              }
              return;
            }
            const url = message.url; // Assume normalized from server.
            const currentHash = message.new_hash;
            const previousHash = message.old_hash || null; // Use old_hash if provided.

            if (url && currentHash) {
              // Check primarily currentHash.
              if (config.debug) {
                console.log(`[NESGuard WS] Received hash_change for ${url}`);
              }
              // Pass both hashes; normalization happens inside updateScriptIntegrity.
              const wasUpdated = updateScriptIntegrity(url, {
                currentHash: currentHash,
                previousHash: previousHash,
              });
              
              // If script was updated, notify the server to invalidate library cache
              // This is critical for ensuring new page loads get the updated hashes
              if (wasUpdated && CRITICAL_SCRIPTS_CHECK(url)) {
                try {
                  if (config.debug) {
                    console.log(`[NESGuard WS] Notifying server about hash change for ${url}`);
                  }
                  
                  // Notify server to invalidate caches with an AJAX call
                  notifyServerAboutHashChange(url, currentHash);
                } catch (e) {
                  console.error(`[NESGuard WS] Error notifying server about hash change: ${e.message}`);
                }
              }
            }
          } // End 'hash_change' handling

          // Handle ping/pong heartbeat.
          else if (message.type === "ping") {
            // Respond to server ping with a pong.
            try {
              socket.send(
                JSON.stringify({
                  type: "pong",
                  timestamp: message.timestamp || Date.now(),
                })
              );
            } catch (e) {
              // Log error but don't crash - heartbeat interval will detect issues
              if (config.debug) {
                console.warn(`[NESGuard WS] Error sending pong: ${e.message}`);
              }
              
              // Schedule a reconnect if socket appears broken
              if (socket.readyState !== WebSocket.OPEN) {
                scheduleReconnect(5000); // Try to reconnect after 5 seconds
              }
            }
          } else if (message.type === "pong" && config.debug) {
            console.log("[NESGuard WS] Pong received.");
          }
          // Handle other message types if necessary.
        } catch (e) {
          // Error parsing message JSON.
          if (config.debug) {
            console.error(
              "[NESGuard WS] Error processing message:",
              e,
              event.data
            );
          }
        }
      }; // End onmessage

      /** On WebSocket Error */
      socket.onerror = function onWsError(errorEvent) {
        // Log the error event object for details.
        console.error("[NESGuard WS] WebSocket error:", errorEvent);
        // onerror is usually followed by onclose, which handles reconnection logic.
        
        // Initiate HTTP fallback on websocket errors
        if (typeof window.fetchLatestHashesViaHttp === 'function') {
          setTimeout(window.fetchLatestHashesViaHttp, 1000);
        }
      }; // End onerror

      /** On WebSocket Close */
      socket.onclose = function onWsClose(event) {
        let logMessage = `[NESGuard WS] WebSocket Closed. Code: ${
          event.code
        }, Reason: ${event.reason || "N/A"}`;
        let shouldReconnect = false;
        let shouldAttemptTokenRefresh = false;

        // Determine if reconnection should be attempted based on close code.
        // See: https://developer.mozilla.org/en-US/docs/Web/API/CloseEvent/code
        switch (event.code) {
          case 1000: // Normal Closure
            logMessage += " (Normal Closure)";
            shouldReconnect = false; // Usually don't reconnect on normal close.
            break;
          case 1001: // Going Away
            logMessage += " (Going Away)";
            shouldReconnect = true; // Server/browser navigating away, maybe reconnect.
            break;
          case 1006: // Abnormal Closure (often network issues)
            logMessage += " (Abnormal Closure)";
            shouldReconnect = true; // Definitely attempt reconnect.
            
            // With abnormal closure, also trigger HTTP fallback
            if (typeof window.fetchLatestHashesViaHttp === 'function') {
              setTimeout(window.fetchLatestHashesViaHttp, 1000);
            }
            break;
          case 1008: // Policy Violation (e.g., Auth Failed)
            logMessage += " (Authentication Failed / Policy Violation)";
            
            // Check for auth-related reasons that suggest token expiration
            const reasonLower = (event.reason || "").toLowerCase();
            if (reasonLower.includes("auth") || 
                reasonLower.includes("token") || 
                reasonLower.includes("unauthorized") ||
                reasonLower.includes("expired")) {
              
              console.warn("[NESGuard WS] Authentication error detected in close reason, attempting token refresh...");
              shouldAttemptTokenRefresh = true;
              shouldReconnect = false;
            } else {
              shouldReconnect = false; // Don't reconnect for other policy violations
            }
            break;
          case 1009: // Message Too Big
          case 1010: // Missing Extension
          case 1011: // Internal Error
            logMessage += " (Protocol/Internal Error)";
            shouldReconnect = true; // Maybe transient, try reconnecting.
            
            // For internal errors, also trigger HTTP fallback
            if (typeof window.fetchLatestHashesViaHttp === 'function') {
              setTimeout(window.fetchLatestHashesViaHttp, 2000);
            }
            break;
          case 1012: // Service Restart
          case 1013: // Try Again Later
          case 1014: // Bad Gateway
            logMessage += " (Server Unavailable/Restarting)";
            shouldReconnect = true; // Definitely attempt reconnect.
            
            // For server unavailability, also trigger HTTP fallback
            if (typeof window.fetchLatestHashesViaHttp === 'function') {
              setTimeout(window.fetchLatestHashesViaHttp, 2000);
            }
            break;
          default:
            if (event.code >= 4000) {
              // Application-specific errors
              logMessage += " (Server Application Error)";
              shouldReconnect = true; // Assume transient server issue.
            } else {
              logMessage += " (Unknown Closure Code)";
              shouldReconnect = true; // Attempt reconnect for unknown codes.
            }
            
            // For unknown errors, also trigger HTTP fallback
            if (typeof window.fetchLatestHashesViaHttp === 'function') {
              setTimeout(window.fetchLatestHashesViaHttp, 2000);
            }
        }

        // Log if in debug mode OR if the closure was not normal/expected.
        if (config.debug || event.code !== 1000) {
          console.log(logMessage);
        }

        // Clean up resources.
        ws = null; // Nullify the socket variable.
        if (heartbeatInterval) {
          clearInterval(heartbeatInterval);
          heartbeatInterval = null;
        }
        
        // Try token refresh if auth failure is detected
        if (shouldAttemptTokenRefresh) {
          // Don't wait for a reconnect schedule, try refresh immediately
          refreshTokenAndReconnect()
            .catch(err => {
              console.error("[NESGuard WS] Token refresh failed after close:", err);
              // If refresh fails, we'll fall back to regular reconnect with backoff
              scheduleReconnect(5000); // Try a regular reconnect after 5 seconds
            });
        }
        // Schedule reconnection attempt if appropriate.
        else if (shouldReconnect) {
          scheduleReconnect();
        }
      }; // End onclose

      return socket; // Return the created socket instance.
    } catch (e) {
      // Catch errors during WebSocket object creation itself.
      console.error("[NESGuard WS] Failed to create WebSocket:", e);
      // Schedule a reconnect attempt even if creation fails.
      scheduleReconnect(5000); // Try again after 5 seconds.
      
      // Trigger HTTP fallback on websocket creation failure
      if (typeof window.fetchLatestHashesViaHttp === 'function') {
        setTimeout(window.fetchLatestHashesViaHttp, 1000);
      }
      
      return null;
    }
  } // End connectToNesGuardWebSocket
  
  /**
   * Notifies the server about hash changes via HTTP to invalidate caches.
   * Uses fetch with timeout, retry, and proper error handling.
   * 
   * @param {string} url The URL of the script with updated hash.
   * @param {string} newHash The new hash value.
   * @returns {Promise<Response|null>} The fetch response or null on error.
   */
  async function notifyServerAboutHashChange(url, newHash) {
    if (!url || !newHash || !config.csrfToken) {
      if (config.debug) {
        console.warn('[NESGuard] Cannot notify server: missing url, hash, or CSRF token');
      }
      return null;
    }
    
    const controller = new AbortController();
    const timeoutId = setTimeout(() => controller.abort(), 10000); // 10 second timeout
    
    try {
      const response = await fetch('/nesguard-hash-update', {
        method: 'POST',
        headers: {
          'Content-Type': 'application/json',
          'X-CSRF-Token': config.csrfToken || '',
          'X-Requested-With': 'XMLHttpRequest'
        },
        credentials: 'same-origin',
        signal: controller.signal,
        body: JSON.stringify({
          url: url,
          hash: newHash,
          timestamp: Date.now()
        })
      });
      
      if (config.debug) {
        console.log(`[NESGuard] Server notified about hash change, status: ${response.status}`);
      }
      
      return response;
    } catch (err) {
      // Check if it's a timeout error
      if (err.name === 'AbortError') {
        console.error(`[NESGuard] Timeout notifying server about hash change for ${url}`);
      } else {
        console.error(`[NESGuard] Failed to notify server about hash change: ${err.message}`);
      }
      
      // Try once more after a short delay if it failed for non-abort reasons
      if (err.name !== 'AbortError') {
        try {
          let retryTimeoutId = setTimeout(async () => {
            try {
              await fetch('/nesguard-hash-update', {
                method: 'POST',
                headers: {
                  'Content-Type': 'application/json',
                  'X-CSRF-Token': config.csrfToken || '',
                  'X-Requested-With': 'XMLHttpRequest'
                },
                credentials: 'same-origin',
                body: JSON.stringify({
                  url: url,
                  hash: newHash,
                  timestamp: Date.now(),
                  retry: true
                })
              });
              if (config.debug) {
                console.log(`[NESGuard] Retry notification successful for ${url}`);
              }
            } catch (e) {
              // Log but don't retry again to avoid infinite loops
              console.error(`[NESGuard] Retry notification also failed: ${e.message}`);
            }
          }, 2000); // Retry after 2 seconds
          
          // Store timeout ID so it can be cleared during page unload
          if (!window.NESGuard_timeoutIds) window.NESGuard_timeoutIds = [];
          window.NESGuard_timeoutIds.push(retryTimeoutId);
        } catch (retryErr) {
          // Just log the error, we already tried twice
          console.error(`[NESGuard] Error during retry scheduling: ${retryErr.message}`);
        }
      }
      
      return null;
    } finally {
      clearTimeout(timeoutId);
    }
  }

  /**
   * Attempts to refresh the authentication token by calling a server endpoint.
   * When successful, it updates the access token and reconnects.
   * 
   * @param {number} retryCount Current retry attempt count
   * @returns {Promise<boolean>} Promise resolving to true if refresh successful
   */
  async function refreshTokenAndReconnect(retryCount = 0) {
    const MAX_RETRIES = 2;
    
    if (config.debug) {
      console.log(`[NESGuard WS] Attempting to refresh authentication token... (attempt ${retryCount + 1}/${MAX_RETRIES + 1})`);
    }
    
    // Create an abort controller for timeout
    const controller = new AbortController();
    const timeoutId = setTimeout(() => controller.abort(), 10000); // 10 second timeout
    
    try {
      // Call our token refresh endpoint
      const response = await fetch("/nesguard-token-refresh", {
        method: "POST",
        headers: {
          "Content-Type": "application/json",
          "X-CSRF-Token": config.csrfToken || "",
          "X-Requested-With": "XMLHttpRequest"
        },
        credentials: "same-origin", // Include cookies for Drupal session
        signal: controller.signal
      });
      
      if (!response.ok) {
        // If we get a 5xx server error and have retries left, try again
        if (response.status >= 500 && retryCount < MAX_RETRIES) {
          console.warn(`[NESGuard WS] Server error (${response.status}) during token refresh, retrying...`);
          // Wait with increasing backoff
          await new Promise(resolve => setTimeout(resolve, 1000 * Math.pow(2, retryCount)));
          return refreshTokenAndReconnect(retryCount + 1);
        }
        
        throw new Error(`Token refresh failed with status: ${response.status}`);
      }
      
      const data = await response.json();
      
      if (!data.success || !data.accessToken) {
        throw new Error("Invalid response from token refresh endpoint");
      }
      
      // Update the token in our config
      config.accessToken = data.accessToken;
      
      if (config.debug) {
        console.log("[NESGuard WS] Successfully refreshed token, reconnecting...");
      }
      
      // Reconnect with new token
      connectToNesGuardWebSocket();
      return true;
    } catch (err) {
      // Check if it's a timeout error
      if (err.name === 'AbortError') {
        console.error(`[NESGuard WS] Token refresh request timed out`);
      }
      
      // Try again with retry if we have attempts left
      if (retryCount < MAX_RETRIES) {
        console.warn(`[NESGuard WS] Token refresh failed (attempt ${retryCount + 1}/${MAX_RETRIES + 1}), retrying: ${err.message}`);
        // Wait with increasing backoff
        await new Promise(resolve => setTimeout(resolve, 1000 * Math.pow(2, retryCount)));
        return refreshTokenAndReconnect(retryCount + 1);
      }
      
      // All retries failed
      console.error("[NESGuard WS] Token refresh failed after all attempts:", err.message);
      
      // If refresh fails, dispatch auth failed event and prevent auto-reconnect
      document.dispatchEvent(new CustomEvent("nesguard-auth-failed"));
      
      if (reconnectTimeout) {
        clearTimeout(reconnectTimeout);
        reconnectTimeout = null;
      }
      
      // Trigger HTTP fallback since WebSocket authentication failed
      if (typeof window.fetchLatestHashesViaHttp === 'function') {
        setTimeout(window.fetchLatestHashesViaHttp, 1000);
      }
      
      return false;
    } finally {
      clearTimeout(timeoutId);
    }
  }

  /**
   * Schedules WebSocket reconnection with exponential backoff and jitter.
   * Avoids scheduling if already scheduled or prerequisites are missing.
   *
   * @param {number|null} [initialDelay=null] - An optional initial delay in ms.
   */
  function scheduleReconnect(initialDelay = null) {
    // Do not schedule if already scheduled OR if required config is missing.
    if (reconnectTimeout || !config.agencyId || !config.accessToken) {
      if (config.debug && !reconnectTimeout) {
        const reason = reconnectTimeout
          ? "already scheduled"
          : "missing config";
        console.log(`[NESGuard WS] Skipping reconnect schedule: ${reason}.`);
      }
      return;
    }
    reconnectAttempts++;

    // Calculate delay: exponential backoff (1s, 2s, 4s...) up to 60s + random jitter (0-1s).
    const baseDelay = Math.min(
      1000 * Math.pow(2, reconnectAttempts - 1),
      60000
    ); // Max 60 seconds.
    const jitter = Math.random() * 1000; // Add up to 1s jitter.
    const delay = initialDelay ?? baseDelay + jitter;

    if (config.debug) {
      console.log(
        `[NESGuard WS] Scheduling reconnect attempt ${reconnectAttempts} in ${(
          delay / 1000
        ).toFixed(2)}s`
      );
    }

    reconnectTimeout = setTimeout(() => {
      const savedTimeoutId = reconnectTimeout;
      // Don't reset the timeout before we check conditions
      
      // Double-check config before connecting (it might have changed).
      if (config.agencyId && config.accessToken) {
        // Only clear the timeout if we're actually proceeding with reconnection
        // and it hasn't been changed already
        if (reconnectTimeout === savedTimeoutId) {
          reconnectTimeout = null;
        }
        connectToNesGuardWebSocket(); // Attempt to reconnect.
      } else if (config.debug) {
        reconnectTimeout = null; // Clear it now that we're done
        console.log(
          "[NESGuard WS] Skipping scheduled reconnect attempt: config missing."
        );
      }
    }, delay);
    
    // Store timeout ID for cleanup
    if (!window.NESGuard_timeoutIds) window.NESGuard_timeoutIds = [];
    window.NESGuard_timeoutIds.push(reconnectTimeout);
  } // End scheduleReconnect

  // --- HTTP Fallback Mechanism ---
  
  /**
   * Fetches latest script hashes via HTTP when WebSocket is unavailable.
   * Implements fetch timeout, retry logic, and response caching.
   */
  window.fetchLatestHashesViaHttp = async function() {
    // Don't make excessive requests if we've checked recently (30 second throttle)
    const TIME_BETWEEN_CHECKS = 30000; // 30 seconds
    const now = Date.now();
    
    if (now - httpFallbackCache.lastCheck < TIME_BETWEEN_CHECKS) {
      if (config.debug) {
        console.log(`[NESGuard] HTTP fallback skipped - last check was ${Math.round((now - httpFallbackCache.lastCheck)/1000)}s ago`);
      }
      return;
    }
    
    httpFallbackCache.lastCheck = now;
    
    if (config.debug) {
      console.log("[NESGuard] Initiating HTTP fallback to fetch latest hashes");
    }
    
    // Don't proceed if we don't have any URLs to check
    if (!config.hashes || config.hashes.size === 0) {
      if (config.debug) {
        console.warn("[NESGuard] HTTP fallback skipped - no scripts to check");
      }
      return;
    }
    
    // Only do this for critical scripts like Stripe
    let criticalScripts = [];
    for (const [url, hashData] of config.hashes.entries()) {
      if (CRITICAL_SCRIPTS_CHECK(url)) {
        criticalScripts.push(url);
      }
    }
    
    if (criticalScripts.length === 0) {
      if (config.debug) {
        console.warn("[NESGuard] HTTP fallback skipped - no critical scripts");
      }
      return;
    }
    
    // Use fetch API to get latest hashes with timeout
    const controller = new AbortController();
    const timeoutId = setTimeout(() => controller.abort(), 8000); // 8 second timeout
    
    try {
      // Use CSRF token if available, otherwise proceed without
      const headers = { 
        'Content-Type': 'application/json',
        'X-Requested-With': 'XMLHttpRequest'
      };
      
      if (config.csrfToken) {
        headers['X-CSRF-Token'] = config.csrfToken;
      }
      
      const response = await fetch('/nesguard-hash-check', {
        method: 'POST',
        headers: headers,
        credentials: 'same-origin',
        signal: controller.signal,
        body: JSON.stringify({
          scripts: criticalScripts,
          timestamp: Date.now()
        })
      });
      
      if (!response.ok) {
        throw new Error(`HTTP fallback failed with status: ${response.status}`);
      }
      
      const data = await response.json();
      httpFallbackCache.data = data; // Cache the response
      
      if (data.success && data.hashes && Object.keys(data.hashes).length > 0) {
        if (config.debug) {
          console.log(`[NESGuard] HTTP fallback received ${Object.keys(data.hashes).length} hash updates`);
        }
        
        // Apply received hashes
        let updateCount = 0;
        for (const [url, hashData] of Object.entries(data.hashes)) {
          // Find matching URL in config.hashes
          for (const [configUrl, configHashData] of config.hashes.entries()) {
            // Normalize URLs to compare more accurately
            const normalizedConfigUrl = configUrl.replace(/^https?:\/\//, '').split(/[?#]/)[0];
            const normalizedUrl = url.replace(/^https?:\/\//, '').split(/[?#]/)[0];
            
            if (normalizedConfigUrl === normalizedUrl || 
                (normalizedConfigUrl.includes(normalizedUrl) && normalizedUrl.length > 10) ||
                (normalizedUrl.includes(normalizedConfigUrl) && normalizedConfigUrl.length > 10)) {
              if (hashData.current && hashData.current !== configHashData.currentHash) {
                if (config.debug) {
                  console.log(`[NESGuard] HTTP fallback updating hash for ${configUrl}: ${hashData.current} (previous: ${hashData.previous || 'none'})`);
                }
                
                // The server should have handled previous hash preservation
                // but we'll double-check to ensure smooth transitions
                const previousHash = hashData.previous || 
                  (configHashData.currentHash !== hashData.current ? configHashData.currentHash : configHashData.previousHash);
                
                // Update hash in map
                config.hashes.set(configUrl, {
                  currentHash: hashData.current,
                  previousHash: previousHash
                });
                
                // Apply to DOM if needed - use the properly determined previous hash
                updateScriptIntegrity(configUrl, {
                  currentHash: hashData.current,
                  previousHash: previousHash
                });
                
                updateCount++;
              }
            }
          }
        }
        
        if (config.debug) {
          console.log(`[NESGuard] HTTP fallback applied ${updateCount} hash updates`);
        }
      } else {
        if (config.debug) {
          console.warn("[NESGuard] HTTP fallback received no hash updates");
        }
      }
    } catch (error) {
      console.error(`[NESGuard] HTTP fallback error: ${error.message}`);
      
      // Try again after a delay if it's a timeout error or server error
      if (error.name === 'AbortError' || error.message.includes('status: 5')) {
        setTimeout(() => {
          if (config.debug) {
            console.log('[NESGuard] Retrying HTTP fallback after failure');
          }
          // Reset the lastCheck time to allow an immediate retry
          httpFallbackCache.lastCheck = 0;
          window.fetchLatestHashesViaHttp();
        }, 5000); // Retry after 5 seconds
      }
    } finally {
      clearTimeout(timeoutId);
    }
  };

  // --- Script Error Handler ---
  
  /**
   * Global handler for script load errors, especially for critical scripts like Stripe.
   * Must be defined before scripts load and have error events.
   * 
   * @param {HTMLScriptElement} scriptElement The script element that failed to load
   */
  window.NESGuard_handleScriptError = function(scriptElement) {
    if (!scriptElement || !scriptElement.src) return;
    
    const scriptUrl = scriptElement.src;
    if (config.debug) {
      console.warn(`[NESGuard] Script load error detected for ${scriptUrl}. Attempting recovery...`);
    }
    
    // For critical scripts like Stripe, we need to ensure they load
    if (CRITICAL_SCRIPTS_CHECK(scriptUrl)) {
      // Record error for analytics - don't expose sensitive data
      document.dispatchEvent(new CustomEvent('nesguard-script-error', {
        detail: {
          url: scriptUrl,
          timestamp: Date.now(),
          hasIntegrity: scriptElement.hasAttribute('integrity')
        }
      }));
        
      // IMPORTANT: The script has already failed to load and won't be reprocessed
      // by the browser even if we change the attributes. We must create a new script element.
      
      const oldIntegrity = scriptElement.getAttribute('integrity');
      if (config.debug) {
        console.warn(`[NESGuard] Script failed to load with integrity: ${oldIntegrity || 'none'}`);
      }
      
      // Add cache-busting parameter to ensure we get a fresh copy
      let scriptUrlWithBust = scriptUrl;
      try {
        // Parse the URL properly to handle all edge cases
        const urlObj = new URL(scriptUrl);
        // Add or update the v parameter
        urlObj.searchParams.set('v', Date.now().toString());
        scriptUrlWithBust = urlObj.toString();
      } catch (e) {
        // Fallback to string manipulation if URL parsing fails
        if (scriptUrlWithBust.includes('?')) {
          // Replace existing v parameter or add a new one
          scriptUrlWithBust = scriptUrlWithBust.replace(/([?&])v=[^&]*/, '$1v=' + Date.now());
          if (!scriptUrlWithBust.includes('v=')) {
            scriptUrlWithBust += '&v=' + Date.now();
          }
        } else {
          scriptUrlWithBust += '?v=' + Date.now();
        }
      }
      
      // Create a new script element to replace the failed one
      const newScript = document.createElement('script');
      newScript.src = scriptUrlWithBust;
      newScript.crossOrigin = 'anonymous';
      
      // Check if we have a more recent hash from the websocket or HTTP fallback
      let updatedHash = null;
      for (const [url, hashData] of config.hashes.entries()) {
        if (url.includes(scriptUrl) || scriptUrl.includes(url)) {
          updatedHash = hashData.currentHash;
          break;
        }
      }
      
      // Track recovery attempts to avoid infinite loops
      const recoveryAttempts = parseInt(scriptElement.getAttribute('data-recovery-attempts') || '0', 10);
      
      // If we have a hash and this is the first failure, try with the hash
      if (updatedHash && recoveryAttempts < 1) {
        if (config.debug) {
          console.warn(`[NESGuard] Adding updated integrity attribute to new script: ${updatedHash}`);
        }
        newScript.setAttribute('integrity', updatedHash);
      } 
      // Otherwise, load without integrity on subsequent attempts
      else {
        if (config.debug) {
          console.warn(`[NESGuard] Loading without integrity after ${recoveryAttempts} attempts`);
        }
        // Don't set integrity attribute at all - this is a fallback to ensure loading
      }
      
      // Set additional attributes on new script
      newScript.setAttribute('data-recovery', 'true');
      newScript.setAttribute('data-recovery-time', Date.now().toString());
      newScript.setAttribute('data-recovery-attempts', (recoveryAttempts + 1).toString());
      
      // Add an onerror handler to the new script for potential additional recovery
      newScript.onerror = function() {
        window.NESGuard_handleScriptError(this);
      };
      
      // Replace the failed script
      if (scriptElement.parentNode) {
        scriptElement.parentNode.insertBefore(newScript, scriptElement);
        scriptElement.parentNode.removeChild(scriptElement);
        
        if (config.debug) {
          console.log(`[NESGuard] Recovery script injected for ${scriptUrl}`);
        }
        
        // Dispatch recovery event
        try {
          document.dispatchEvent(new CustomEvent('nesguard-script-recovery', {
            detail: {
              url: scriptUrl,
              timestamp: Date.now(),
              attempt: recoveryAttempts + 1,
              withIntegrity: newScript.hasAttribute('integrity')
            }
          }));
        } catch (e) {
          if (config.debug) {
            console.warn(`[NESGuard] Error dispatching recovery event: ${e.message}`);
          }
        }
      }
      
      // Trigger HTTP fallback to get latest hashes in case we're working with outdated data
      if (typeof window.fetchLatestHashesViaHttp === 'function' && recoveryAttempts === 0) {
        setTimeout(window.fetchLatestHashesViaHttp, 500);
      }
    }
  };

  // --- Initialization ---

  /**
   * Initializes the NesGuard client-side script.
   * Applies initial SRI and connects to WebSocket.
   */
  function initialize() {
    if (initialized) {
      return; // Prevent multiple initializations
    }
    
    initialized = true;
    
    if (config.debug) {
      console.log(
        `[NESGuard] Initializing client v${config.version}${
          config.sriEnabled ? (config.enforceSRI ? " (SRI Enforced)" : " (SRI Monitor-Only)") : " (SRI Inactive)"
        }`
      );
    }
    
    // Check for pre-existing SRI failures that need recovery - especially for Stripe
    // This runs BEFORE initial SRI application to catch scripts that failed during initial load
    setTimeout(function() {
      if (config.debug) {
        console.log('[NESGuard] Running initial script failure check...');
      }
      
      // For critical scripts like Stripe, check if we need to do additional recovery
      // This is needed in case the script failed to load due to SRI validation
      // and our attempt to rescue the script wasn't activated
      
      // First check for any direct Drupal state info about latest hashes
      // This is a way to bypass the WebSocket when there's a new hash
      if (typeof drupalSettings?.nesguard?.stripeLatestHash === 'string' && 
          typeof drupalSettings?.nesguard?.stripeHashUpdateTime === 'number') {
        const stripeHash = drupalSettings.nesguard.stripeLatestHash;
        const updateTime = drupalSettings.nesguard.stripeHashUpdateTime;
        const age = Date.now()/1000 - updateTime;
        
        if (age < 3600) { // Only use if less than 1 hour old
          if (config.debug) {
            console.log(`[NESGuard] Found recent Stripe hash in Drupal settings: ${stripeHash} (${Math.round(age)}s old)`);
          }
          
          // Update our hash map with this value
          for (const [url, hashData] of config.hashes.entries()) {
            if (CRITICAL_SCRIPTS_CHECK(url)) {
              config.hashes.set(url, {
                currentHash: stripeHash,
                previousHash: hashData.currentHash !== stripeHash ? hashData.currentHash : hashData.previousHash
              });
              
              if (config.debug) {
                console.log(`[NESGuard] Updated hash map with Drupal-provided Stripe hash: ${stripeHash}`);
              }
            }
          }
        }
      }
      
      // Find script tags that might have failed SRI validation
      for (const [url, hashData] of config.hashes.entries()) {
        if (CRITICAL_SCRIPTS_CHECK(url)) {
          // Look for script tags matching this URL
          try {
            const baseUrl = url.split(/[?#]/)[0];
            const escapedBaseUrl = baseUrl.replace(/[.*+?^${}()|[\]\\]/g, '\\$&');
            const scriptElements = document.querySelectorAll(`script[src^="${escapedBaseUrl}"]`);
            
            // See if script is present but failed to initialize (no stripe object)
            const hasStripe = typeof window.Stripe === 'function';
            
            if ((scriptElements.length === 0 || !hasStripe) && config.debug) {
              console.log(`[NESGuard] Critical script needs recovery: ${url} (Found: ${scriptElements.length > 0}, Initialized: ${hasStripe})`);
              
              // Create and inject the script - don't use integrity for startup recovery
              // This ensures the script loads even if hash is currently mismatched
              const script = document.createElement('script');
              script.src = url + (url.includes('?') ? '&' : '?') + 'v=' + Date.now();
              script.crossOrigin = 'anonymous';
              script.setAttribute('data-sri-mode', 'monitor');
              script.setAttribute('data-recovery', 'startup');
              script.setAttribute('data-ignore-integrity', 'true');
              
              // Make sure to use the global function not the attribute
              script.onerror = function() { 
                window.NESGuard_handleScriptError(this); 
              };
              
              if (config.debug) {
                console.log(`[NESGuard] Injecting recovery script for: ${url}`);
              }
              
              // Make sure head exists before appending
              if (document.head) {
                document.head.appendChild(script);
              } else if (document.body) {
                document.body.appendChild(script);
              } else {
                // Fallback to document if neither head nor body are available
                document.appendChild(script);
              }
            }
          } catch (e) {
            if (config.debug) {
              console.warn(`[NESGuard] Error in startup recovery check: ${e.message}`);
            }
          }
        }
      }
    }, 1000); // Delay execution to allow page to finish loading first

    // --- Apply Initial SRI ---
    if (config.sriEnabled && config.hashes?.size > 0) {
      if (config.debug) {
        console.log(
          `[NESGuard] Applying initial SRI based on Drupal settings for ${config.hashes.size} scripts.`
        );
      }
      let initialAppliedCount = 0;
      for (const [url, hashData] of config.hashes) {
        // Pass the full hashData object {currentHash, previousHash}.
        if (hashData && hashData.currentHash) {
          if (updateScriptIntegrity(url, hashData)) {
            initialAppliedCount++;
          }
        } else if (config.debug) {
          console.warn(
            `[NESGuard] Skipping initial SRI for ${url}, invalid hashData:`,
            hashData
          );
        }
      }
      if (config.debug && initialAppliedCount > 0) {
        console.log(
          `[NESGuard] Applied initial SRI modifications to ${initialAppliedCount} script instances.`
        );
      }
    } else if (config.sriEnabled && config.debug) {
      console.log(
        "[NESGuard] No initial SRI hashes provided or map empty via Drupal settings."
      );
    } else if (!config.sriEnabled && config.debug) {
      console.log("[NESGuard] SRI is disabled, skipping initial application.");
    }

    // --- Connect WebSocket ---
    // WebSocket handles its own config checks internally. Attempt connection.
    connectToNesGuardWebSocket();
    
    // Schedule periodic fallback checks in case WebSocket fails
    const fallbackIntervalId = setInterval(window.fetchLatestHashesViaHttp, 300000); // Every 5 minutes
    // Store interval ID for cleanup
    if (!window.NESGuard_intervalIds) window.NESGuard_intervalIds = [];
    window.NESGuard_intervalIds.push(fallbackIntervalId);
    
    // --- Expose Limited Global API ---
    window.NESGUARD_CLIENT = {
      config: {
        // Expose non-sensitive config.
        version: config.version,
        debug: config.debug,
        sriEnabled: config.sriEnabled,
        enforceSRI: config.enforceSRI,
        clientId: config.clientId,
      },
      /** Manually trigger a WebSocket reconnection attempt. */
      reconnect: () => {
        if (config.debug) {
          console.log("[NESGuard] Manual reconnect triggered.");
        }
        reconnectAttempts = 0; // Reset attempts for manual reconnect.
        connectToNesGuardWebSocket(); // Attempt connection immediately.
      },
      /** Manually trigger HTTP fallback. */
      checkLatestHashes: () => {
        if (typeof window.fetchLatestHashesViaHttp === 'function') {
          // Reset the lastCheck time to allow an immediate check
          httpFallbackCache.lastCheck = 0;
          window.fetchLatestHashesViaHttp();
        }
      },
      /** Get current WebSocket readyState. */
      getSocketState: () => (ws ? ws.readyState : WebSocket.CLOSED), // CLOSED=3.
      // Flag to prevent multiple initializations by Drupal behaviors.
      initialized: true, // Set flag *after* initialization logic runs.
    };

    // Register page unload cleanup
    window.addEventListener('beforeunload', function() {
      // Close WebSocket cleanly
      if (ws && ws.readyState === WebSocket.OPEN) {
        try {
          ws.close(1000, "Page unload");
        } catch (e) {
          /* Ignore errors during page unload */
        }
      }
      
      // Clear any pending timeouts
      if (reconnectTimeout) {
        clearTimeout(reconnectTimeout);
        reconnectTimeout = null;
      }
      
      // Clear all stored timeouts
      if (window.NESGuard_timeoutIds && Array.isArray(window.NESGuard_timeoutIds)) {
        window.NESGuard_timeoutIds.forEach(id => {
          if (id) clearTimeout(id);
        });
        window.NESGuard_timeoutIds = [];
      }
      
      // Clear all stored intervals
      if (window.NESGuard_intervalIds && Array.isArray(window.NESGuard_intervalIds)) {
        window.NESGuard_intervalIds.forEach(id => {
          if (id) clearInterval(id);
        });
        window.NESGuard_intervalIds = [];
      }
    });

    if (config.debug) {
      console.log("[NESGuard] Initialization complete.");
    }
  } // End initialize

  // --- Drupal Behavior ---
  Drupal.behaviors.nesguardClient = {
    /**
     * Initializes the NesGuard client once per page load.
     * @param {HTMLElement} context - The context for attaching behaviors.
     * @param {object} settings - Drupal settings object.
     */
    attach: function (context, settings) {
      try {
        // Run initialization only once on the initial document load.
        // Check context and the global initialized flag. Also check settings again.
        if (
          context === document &&
          !window.NESGUARD_CLIENT?.initialized &&
          settings?.nesguard
        ) {
          // Use setTimeout to ensure this runs after other critical scripts
          // Increased timeout to 500ms to ensure DOM is fully loaded
          setTimeout(function() {
            try {
              initialize();
            } catch (e) {
              console.error("[NESGuard] Error during initialization:", e);
            }
          }, 500);
          // Initialization sets window.NESGUARD_CLIENT.initialized = true now.
        } else if (context === document && !settings?.nesguard) {
          // Log error if settings are missing when behavior runs on document.
          console.error(
            "[NESGuard Behavior] Cannot initialize: settings.nesguard missing on document context."
          );
        }
      } catch (e) {
        console.error("[NESGuard] Error in behavior attach:", e);
      }
    },
  };
})(jQuery, Drupal, drupalSettings); // Pass jQuery, Drupal and drupalSettings to the IIFE.