<?php

namespace Drupal\nesguard_drupal\Controller;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\State\StateInterface;
use Drupal\nesguard_drupal\Service\NesguardAuthService;
use Drupal\nesguard_drupal\Service\NesguardSriService;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Drupal\Core\Config\ConfigFactoryInterface;
use Psr\Log\LoggerInterface;
use Drupal\Core\Lock\LockBackendInterface;
use Drupal\Core\Cache\CacheTagsInvalidatorInterface;
use Drupal\nesguard_drupal\Service\NesguardSubscriptionService;

/**
 * Controller for NESGuard hash updates and API callbacks.
 */
class NesguardController extends ControllerBase {

  /**
   * The state service.
   *
   * @var \Drupal\Core\State\StateInterface
   */
  protected $state;

  /**
   * The NESGuard auth service.
   *
   * @var \Drupal\nesguard_drupal\Service\NesguardAuthService
   */
  protected $authService;

  /**
   * The NESGuard SRI service.
   *
   * @var \Drupal\nesguard_drupal\Service\NesguardSriService
   */
  protected $sriService;

  /**
   * The config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The logger channel.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * The lock service.
   *
   * @var \Drupal\Core\Lock\LockBackendInterface
   */
  protected $lock;

  /**
   * The cache tags invalidator.
   *
   * @var \Drupal\Core\Cache\CacheTagsInvalidatorInterface
   */
  protected $cacheTagsInvalidator;

  /**
   * The NESGuard subscription service.
   *
   * @var \Drupal\nesguard_drupal\Service\NesguardSubscriptionService
   */
  protected $subscriptionService;

  /**
   * Creates a new NesguardController instance.
   */
  public function __construct(
    StateInterface $state,
    NesguardAuthService $auth_service,
    NesguardSriService $sri_service,
    ConfigFactoryInterface $config_factory,
    LoggerInterface $logger,
    LockBackendInterface $lock,
    CacheTagsInvalidatorInterface $cache_tags_invalidator,
    NesguardSubscriptionService $subscription_service,
  ) {
    $this->state = $state;
    $this->authService = $auth_service;
    $this->sriService = $sri_service;
    $this->configFactory = $config_factory;
    $this->logger = $logger;
    $this->lock = $lock;
    $this->cacheTagsInvalidator = $cache_tags_invalidator;
    $this->subscriptionService = $subscription_service;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('state'),
      $container->get('nesguard_drupal.auth'),
      $container->get('nesguard_drupal.sri'),
      $container->get('config.factory'),
      $container->get('logger.channel.nesguard_drupal'),
      $container->get('lock'),
      $container->get('cache_tags.invalidator'),
      $container->get('nesguard_drupal.subscription')
    );
  }

  /**
   * Access callback for the hash update route.
   *
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The account requesting access.
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The current request.
   *
   * @return \Drupal\Core\Access\AccessResult
   *   The access result.
   */
  public function accessCheck(AccountInterface $account, Request $request): AccessResult {
    // 1. Check the API key/token in the header
    $apiKeyHeader = $request->headers->get('X-NESGuard-API-Key');
    $apiTokenHeader = $request->headers->get('X-NESGuard-Auth-Token');

    // If direct API key is provided, verify it.
    if (!empty($apiKeyHeader)) {
      return $this->authService->verifyApiKey($apiKeyHeader)
        ? AccessResult::allowed()
        : AccessResult::forbidden('Invalid API key');
    }

    // If token is provided, verify it.
    if (!empty($apiTokenHeader)) {
      return $this->authService->verifyToken($apiTokenHeader)
        ? AccessResult::allowed()
        : AccessResult::forbidden('Invalid auth token');
    }

    // No valid authentication provided.
    return AccessResult::forbidden('Missing authentication');
  }

  /**
   * Endpoint to receive SRI hash updates.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The current request.
   *
   * @return \Symfony\Component\HttpFoundation\Response
   *   The response object.
   */
  public function hashUpdate(Request $request): Response {
    // Get content and parse JSON.
    $content = $request->getContent();
    $data = json_decode($content, TRUE);

    if (empty($data) || !is_array($data)) {
      return new JsonResponse(['error' => 'Invalid request format'], 400);
    }

    // Process hash updates - expected format:
    // { url: string, hash/newHash: string, oldHash?: string }.
    if ((!empty($data['url']) && (!empty($data['hash']) || !empty($data['newHash'])))) {
      $url = trim($data['url']);
      $newHash = !empty($data['hash']) ? trim($data['hash']) : trim($data['newHash']);
      $oldHash = !empty($data['oldHash']) ? trim($data['oldHash']) : NULL;

      // Log the hash update for debugging.
      if ($this->configFactory->get('nesguard_drupal.settings')->get('debug_mode')) {
        $this->logger->debug('Client-side notification of hash change: URL: @url, Hash: @hash', [
          '@url' => $url,
          '@hash' => $newHash,
        ]);
      }

      // Update the hash in our SRI service if available.
      if (method_exists($this->sriService, 'updateScriptHash')) {
        $this->sriService->updateScriptHash($url, $newHash, $oldHash);
      }

      // Most importantly, invalidate the library cache to ensure new page loads get the updated hash.
      try {
        // Acquire a lock before invalidating cache to prevent race conditions.
        $lock_key = 'nesguard_drupal.cache_invalidation_lock';

        if ($this->lock->acquire($lock_key, 5)) {
          try {
            $this->cacheTagsInvalidator->invalidateTags(['library_info']);

            // For Stripe URLs, also update special state vars for emergency recovery.
            if (strpos($url, 'stripe.com') !== FALSE) {
              $normalizedHash = function_exists('nesguard_drupal_normalize_hash') ?
                nesguard_drupal_normalize_hash($newHash) : $newHash;

              $this->state->set('nesguard_drupal.stripe_latest_hash', $normalizedHash);
              $this->state->set('nesguard_drupal.stripe_hash_update_time', time());

              if ($this->configFactory->get('nesguard_drupal.settings')->get('debug_mode')) {
                $this->logger->debug('Updated stripe_latest_hash state var for emergency recovery: @hash', [
                  '@hash' => $normalizedHash,
                ]);
              }
            }

            if ($this->configFactory->get('nesguard_drupal.settings')->get('debug_mode')) {
              $this->logger->debug('Cache tags invalidated for hash change: URL: @url', [
                '@url' => $url,
              ]);
            }

            // Store the update timestamp for reference.
            $this->state->set('nesguard_drupal.last_hash_update', [
              'url' => $url,
              'hash' => $newHash,
              'time' => time(),
              'source' => 'client-notification',
            ]);
          }
          finally {
            // Always release the lock.
            $this->lock->release($lock_key);
          }
        }
        else {
          // Another process is already invalidating the cache.
          if ($this->configFactory->get('nesguard_drupal.settings')->get('debug_mode')) {
            $this->logger->debug('Skipped cache invalidation (lock not acquired) for: @url', ['@url' => $url]);
          }
        }
      }
      catch (\Exception $e) {
        $this->logger->error('Failed to invalidate cache tags: @error', ['@error' => $e->getMessage()]);
      }

      // Return 1x1 transparent GIF with 204 status.
      return $this->createGifResponse(204);
    }

    // Handle batch updates if provided.
    if (!empty($data['updates']) && is_array($data['updates'])) {
      $updatedCount = 0;

      foreach ($data['updates'] as $update) {
        if (!empty($update['url']) && !empty($update['newHash'])) {
          $url = trim($update['url']);
          $newHash = trim($update['newHash']);
          $oldHash = !empty($update['oldHash']) ? trim($update['oldHash']) : NULL;

          $this->sriService->updateScriptHash($url, $newHash, $oldHash);
          $updatedCount++;
        }
      }

      if ($updatedCount > 0) {
        // Return 1x1 transparent GIF with 204 status.
        return $this->createGifResponse(204);
      }
    }

    return new JsonResponse(['error' => 'No valid hash updates found'], 400);
  }

  /**
   * Creates a 1x1 transparent GIF response.
   *
   * @param int $status
   *   The HTTP status code.
   *
   * @return \Symfony\Component\HttpFoundation\Response
   *   The response object.
   */
  protected function createGifResponse(int $status = 200): Response {
    // 1x1 transparent GIF.
    $gifContent = base64_decode('R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAAALAAAAAABAAEAAAIBRAA7');
    $response = new Response($gifContent, $status);
    $response->headers->set('Content-Type', 'image/gif');
    $response->headers->set('Content-Length', strlen($gifContent));
    $response->headers->set('Cache-Control', 'no-store, no-cache, must-revalidate, max-age=0');

    return $response;
  }

  /**
   * Endpoint to check for latest hash values.
   *
   * This serves as a fallback when WebSocket connection isn't working.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The current request.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   JSON response with hash updates.
   */
  public function hashCheck(Request $request): JsonResponse {
    // Check if the module is enabled.
    $config = $this->config('nesguard_drupal.settings');
    if (!$config->get('enabled')) {
      return new JsonResponse(['success' => FALSE, 'error' => 'Module disabled'], 403);
    }

    // Get request content and parse.
    $content = $request->getContent();
    $data = json_decode($content, TRUE);

    if (empty($data) || !is_array($data) || empty($data['scripts']) || !is_array($data['scripts'])) {
      return new JsonResponse(['success' => FALSE, 'error' => 'Invalid request format'], 400);
    }

    // Check subscription status.
    if ($this->subscriptionService->isActive() !== TRUE) {
      return new JsonResponse(['success' => FALSE, 'error' => 'Subscription inactive'], 403);
    }

    // Get latest hashes for the requested scripts.
    $debug = $config->get('debug_mode') ?? FALSE;
    $requestedScripts = $data['scripts'];
    $latestHashes = [];

    foreach ($requestedScripts as $scriptUrl) {
      $normalizedUrl = function_exists('nesguard_drupal_normalize_url')
        ? nesguard_drupal_normalize_url($scriptUrl)
        : $scriptUrl;

      if (!$normalizedUrl) {
        continue;
      }

      // Get hash for this script from our local cache first.
      $scriptData = $this->state->get('nesguard_drupal.sri_script_data');
      $existingHash = NULL;
      $existingPrevious = NULL;

      if (!empty($scriptData['hashes'][$normalizedUrl])) {
        $existingHash = $scriptData['hashes'][$normalizedUrl]['current'] ?? NULL;
        $existingPrevious = $scriptData['hashes'][$normalizedUrl]['previous'] ?? NULL;
      }

      // Now get the fresh hash
      // Force fresh lookup.
      $hashData = $this->sriService->getIntegrityHash($normalizedUrl, TRUE);

      if ($hashData !== NULL) {
        // If the hash has changed, we need to preserve the old one as "previous"
        // This is critical for handling transition periods.
        if ($existingHash && $hashData['current'] !== $existingHash) {
          // The hash has changed - use the existing hash as previous.
          $latestHashes[$normalizedUrl] = [
            'current' => $hashData['current'],
            'previous' => $existingHash,
          ];

          if ($this->configFactory->get('nesguard_drupal.settings')->get('debug_mode')) {
            $this->logger->debug('Hash check endpoint: Hash changed for @url. New: @new, Previous: @prev', [
              '@url' => $normalizedUrl,
              '@new' => $hashData['current'],
              '@prev' => $existingHash,
            ]);
          }
        }
        else {
          // Hash unchanged or new - use the API-provided previous if available.
          $latestHashes[$normalizedUrl] = [
            'current' => $hashData['current'],
            'previous' => $hashData['previous'] ?? $existingPrevious,
          ];
        }

        if ($debug) {
          $this->logger->debug('Hash check endpoint: Retrieved hash for @url: @hash', [
            '@url' => $normalizedUrl,
            '@hash' => $hashData['current'],
          ]);
        }
      }
    }

    // Also store the result in a cache for future reference
    // This helps us handle cases where getIntegrityHash might temporarily fail.
    $this->state->set('nesguard_drupal.hash_check_cache', [
      'hashes' => $latestHashes,
      'timestamp' => time(),
    ]);
    $this->state->set('nesguard_drupal.hash_check_last_run', time());

    // For Stripe URLs, also update our emergency recovery variables.
    foreach ($latestHashes as $url => $hashData) {
      if (strpos($url, 'stripe.com') !== FALSE && !empty($hashData['current'])) {
        $this->state->set('nesguard_drupal.stripe_latest_hash', $hashData['current']);
        $this->state->set('nesguard_drupal.stripe_hash_update_time', time());

        if ($debug) {
          $this->logger->debug('Hash check endpoint: Updated emergency recovery hash for Stripe: @hash', [
            '@hash' => $hashData['current'],
          ]);
        }

        // No need to check more URLs after finding Stripe.
        break;
      }
    }

    return new JsonResponse([
      'success' => TRUE,
      'hashes' => $latestHashes,
      'timestamp' => time(),
    ]);
  }

  /**
   * Endpoint to refresh access tokens for WebSocket connections.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   JSON response with new access token or error.
   */
  public function tokenRefresh(): JsonResponse {
    // Check if the module is enabled.
    $config = $this->config('nesguard_drupal.settings');
    if (!$config->get('enabled')) {
      return new JsonResponse(['success' => FALSE, 'error' => 'Module disabled'], 403);
    }

    // Force a token refresh.
    $accessToken = $this->authService->getAccessToken(TRUE);

    if (empty($accessToken)) {
      return new JsonResponse(
        ['success' => FALSE, 'error' => 'Failed to refresh access token'],
        401
      );
    }

    $this->logger->info('WebSocket token refresh successful');

    return new JsonResponse([
      'success' => TRUE,
      'accessToken' => $accessToken,
    ]);
  }

}