<?php

namespace Drupal\nesguard_drupal\Form;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Site\Settings;
use Drupal\Core\State\StateInterface;
use Drupal\Component\Utility\Html;
use Drupal\nesguard_drupal\Service\NesguardAuthService;
use Drupal\nesguard_drupal\Service\NesguardSriService;
use Drupal\nesguard_drupal\Service\NesguardSubscriptionService;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RequestStack;
use Drupal\Core\Cache\CacheTagsInvalidatorInterface;

/**
 * Configure NESGuard SRI Monitor settings using API Key/Agency ID authentication.
 */
class NesguardSettingsForm extends ConfigFormBase {

  /**
   * The state service.
   *
   * @var \Drupal\Core\State\StateInterface
   */
  protected $state;

  /**
   * The site settings service.
   *
   * @var \Drupal\Core\Site\Settings
   */
  protected $settings;

  /**
   * The current request.
   *
   * @var \Symfony\Component\HttpFoundation\Request
   */
  protected $currentRequest;

  /**
   * The logger channel.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * The NESGuard subscription service.
   *
   * @var \Drupal\nesguard_drupal\Service\NesguardSubscriptionService
   */
  protected $subscriptionService;

  /**
   * The NESGuard SRI service.
   *
   * @var \Drupal\nesguard_drupal\Service\NesguardSriService
   */
  protected $sriService;

  /**
   * The NESGuard authentication service.
   *
   * @var \Drupal\nesguard_drupal\Service\NesguardAuthService
   */
  protected $authService;

  /**
   * The cache tags invalidator.
   *
   * @var \Drupal\Core\Cache\CacheTagsInvalidatorInterface
   */
  protected $cacheTagsInvalidator;

  /**
   * Constructs a NesguardSettingsForm object.
   */
  public function __construct(
    ConfigFactoryInterface $config_factory,
    StateInterface $state,
    Settings $settings,
    RequestStack $request_stack,
    LoggerInterface $logger,
    NesguardSubscriptionService $subscription_service,
    NesguardSriService $sri_service,
    NesguardAuthService $auth_service,
    CacheTagsInvalidatorInterface $cache_tags_invalidator,
  ) {
    parent::__construct($config_factory);
    $this->state = $state;
    $this->settings = $settings;
    $this->currentRequest = $request_stack->getCurrentRequest();
    $this->logger = $logger;
    $this->subscriptionService = $subscription_service;
    $this->sriService = $sri_service;
    $this->authService = $auth_service;
    $this->cacheTagsInvalidator = $cache_tags_invalidator;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new self(
      $container->get('config.factory'),
      $container->get('state'),
      $container->get('settings'),
      $container->get('request_stack'),
      $container->get('logger.channel.nesguard_drupal'),
      $container->get('nesguard_drupal.subscription'),
      $container->get('nesguard_drupal.sri'),
      $container->get('nesguard_drupal.auth'),
      $container->get('cache_tags.invalidator')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId(): string {
    return 'nesguard_drupal_admin_settings_form';
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames(): array {
    return ['nesguard_drupal.settings'];
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state): array {
    $config = $this->config('nesguard_drupal.settings');

    if ($this->currentRequest && !$this->currentRequest->isSecure()) {
      $this->messenger()->addWarning($this->t('NESGuard relies on HTTPS for secure operation. Please configure your site to use HTTPS.'));
    }

    $encryption_key_set = !empty($this->settings->get('nesguard_encryption_key'));

    $form['status'] = [
      '#type' => 'details',
      '#title' => $this->t('General Settings'),
      '#open' => TRUE,
    ];
    $form['status']['enabled'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Enable NESGuard Integration'),
      '#default_value' => $config->get('enabled'),
      '#description' => $this->t('Enable SRI protection and API/WebSocket features.'),
    ];

    // Credentials Section (API Key / Agency ID).
    $form['credentials'] = [
      '#type' => 'details',
      '#title' => $this->t('API Credentials (API Key / Agency ID)'),
      '#open' => TRUE,
      '#description' => $this->t('Credentials for your <a href=":url" target="_blank">NESGuard.cloud</a> account using API Key authentication.', [':url' => 'https://nesguard.cloud']),
      '#states' => ['visible' => [':input[name="enabled"]' => ['checked' => TRUE]]],
    ];
    $form['credentials']['api_email_display'] = [
      '#type' => 'email',
      '#title' => $this->t('NESGuard Account Email (Reference)'),
      '#default_value' => $config->get('api_email'),
      '#description' => $this->t('The email associated with the account (used for reference, not login).'),
    ];
    $form['credentials']['agency_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Agency Identifier'),
      '#default_value' => $config->get('agency_id'),
      '#required' => TRUE,
      '#states' => ['required' => [':input[name="enabled"]' => ['checked' => TRUE]]],
      '#description' => $this->t('Your agency identifier (e.g., "NES-SIMON") required for authentication and WebSocket connection.'),
    ];

    // API Key Handling.
    $current_api_key_status_msg = $this->t('Not configured.');
    if (!empty($config->get('api_key'))) {
      $current_api_key_status_msg = $encryption_key_set ? $this->t('Configured (Encrypted).') : $this->t('Configured, but cannot be decrypted (Encryption key missing in settings.php).');
    }
    $form['credentials']['api_key_status'] = [
      '#type' => 'item',
      '#title' => $this->t('API Key Status'),
      '#markup' => $current_api_key_status_msg,
    ];
    if (!$encryption_key_set) {
      $form['credentials']['encryption_key_warning'] = [
        '#type' => 'markup',
        '#markup' => '<div class="messages messages--error">' . $this->t('<strong>Required for API Key:</strong> Add <code>$settings["nesguard_encryption_key"] = "YOUR_SECURE_RANDOM_KEY";</code> to your site\'s <strong>settings.php</strong> file to save/use an API Key.') . '</div>',
      ];
      $form['credentials']['encryption_key_help'] = [
        '#type' => 'markup',
        '#markup' => '<p>' . $this->t('Generate a secure key using PHP CLI: <code>echo base64_encode(random_bytes(32));</code>') . '</p>',
      ];
    }
    $form['credentials']['api_key_new'] = [
      '#type' => 'password',
      '#title' => $this->t('Enter/Update API Key'),
      '#description' => $this->t('Enter the API Key here. Leave blank to keep the current key. It will be encrypted on save.'),
      '#disabled' => !$encryption_key_set,
    ];

    // SRI & Subscription Status.
    $form['sri_status'] = [
      '#type' => 'details',
      '#title' => $this->t('SRI & Subscription Status'),
      '#open' => TRUE,
      '#states' => ['visible' => [':input[name="enabled"]' => ['checked' => TRUE]]],
    ];
    $subscription_active = $this->subscriptionService->isActive();
    $sri_status_markup = $this->t('Status Unknown (Check API credentials and connection)');
    if ($subscription_active === TRUE) {
      $sri_status_markup = $this->t('Active NESGuard subscription found. SRI protection will be applied.');
    }
    elseif ($subscription_active === FALSE) {
      $sri_status_markup = $this->t('No active NESGuard subscription found. SRI protection will NOT be applied.');
      if ($config->get('enabled')) {
        $this->messenger()->addWarning($this->t('SRI protection is disabled because no active NESGuard subscription was found.'));
      }
    }
    $form['sri_status']['status_display'] = [
      '#type' => 'item',
      '#title' => $this->t('Subscription Status'),
      '#markup' => $sri_status_markup,
      '#description' => $this->t('SRI protection is automatically enabled/disabled based on your NESGuard subscription status checked via API.'),
    ];

    $form['sri_status']['enforce_sri'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Enforce SRI (Block scripts with invalid hashes)'),
      '#default_value' => $config->get('enforce_sri') ?? FALSE,
      '#description' => $this->t('When enabled, scripts with missing or invalid integrity hashes will be blocked by the browser. When disabled, SRI monitoring continues but scripts will always load. <strong>Disable this option if you\'re experiencing "script not defined" errors in production (e.g., with Stripe.js).</strong>'),
    ];
    $monitored_urls = $this->sriService->getMonitoredScriptUrls();
    $form['sri_status']['monitored_scripts_display'] = [
      '#type' => 'item',
      '#title' => $this->t('Monitored Scripts (from API)'),
      '#markup' => !empty($monitored_urls) ? '<ul><li>' . implode('</li><li>', array_map([Html::class, 'escape'], $monitored_urls)) . '</li></ul>' : $this->t('No scripts found or API connection failed.'),
      '#description' => $this->t('This list is automatically fetched from the NESGuard API based on your account configuration.'),
    ];

    // CSP Settings.
    $form['csp'] = [
      '#type' => 'details',
      '#title' => $this->t('Content Security Policy (CSP) Management (Optional)'),
      '#open' => $config->get('csp_enabled'),
      '#description' => $this->t('<strong>Warning:</strong> Enabling CSP management here can conflict with other modules (e.g., CSP module), server configurations (Apache/Nginx headers), or meta tags. Use with caution and test thoroughly. If another system manages your CSP, leave this disabled.'),
      '#states' => ['visible' => [':input[name="enabled"]' => ['checked' => TRUE]]],
    ];
    $form['csp']['csp_enabled'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Enable CSP Header Management by this Module'),
      '#default_value' => $config->get('csp_enabled') ?? FALSE,
      '#description' => $this->t('If checked, this module will add a <code>Content-Security-Policy</code> header based on the settings below. <strong>Leave unchecked if CSP is handled elsewhere.</strong>'),
    ];
    $form['csp']['csp_report_only'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Use Report-Only Mode'),
      '#default_value' => $config->get('csp_report_only') ?? FALSE,
      '#description' => $this->t('Log CSP violations instead of blocking them.'),
      '#states' => ['visible' => [':input[name="csp_enabled"]' => ['checked' => TRUE]]],
    ];
    $form['csp']['common_sources_fs'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Common Source Allowances'),
      '#states' => ['visible' => [':input[name="csp_enabled"]' => ['checked' => TRUE]]],
    ];
    $form['csp']['common_sources_fs']['csp_allow_self'] = [
      '#type' => 'checkbox',
      '#title' => $this->t("Allow 'self'"),
      '#default_value' => $config->get('csp_allow_self') ?? TRUE,
      '#disabled' => TRUE,
    ];
    $form['csp']['common_sources_fs']['csp_allow_unsafe_inline_style'] = [
      '#type' => 'checkbox',
      '#title' => $this->t("Allow 'unsafe-inline' for Styles"),
      '#default_value' => $config->get('csp_allow_unsafe_inline_style') ?? TRUE,
      '#description' => $this->t("Needed for themes/modules using inline styles (reduces security)."),
    ];
    $form['csp']['common_sources_fs']['csp_allow_data_img'] = [
      '#type' => 'checkbox',
      '#title' => $this->t("Allow data: URIs for Images"),
      '#default_value' => $config->get('csp_allow_data_img') ?? TRUE,
      '#description' => $this->t("Needed for inline images."),
    ];
    $form['csp']['service_sources_fs'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Specific Service Allowances'),
      '#states' => ['visible' => [':input[name="csp_enabled"]' => ['checked' => TRUE]]],
    ];
    $form['csp']['service_sources_fs']['csp_allow_google_fonts'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Allow Google Fonts'),
      '#default_value' => $config->get('csp_allow_google_fonts') ?? TRUE,
      '#description' => $this->t('Adds fonts.googleapis.com (style-src) and fonts.gstatic.com (font-src).'),
    ];
    $form['csp']['service_sources_fs']['csp_allow_stripe'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Allow Stripe'),
      '#default_value' => $config->get('csp_allow_stripe') ?? TRUE,
      '#description' => $this->t('Adds js.stripe.com (script-src) and *.stripe.com (frame-src, connect-src, img-src).'),
    ];
    $form['csp']['service_sources_fs']['csp_allow_nesguard'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Allow NESGuard API/WS'),
      '#default_value' => $config->get('csp_allow_nesguard') ?? TRUE,
      '#description' => $this->t('Adds nesguard.cloud and wss://*.nesguard.cloud (connect-src).'),
    ];
    $form['csp']['extra_sources_fs'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Additional Custom Sources (One per line)'),
      '#states' => ['visible' => [':input[name="csp_enabled"]' => ['checked' => TRUE]]],
    ];
    $common_textarea = ['#type' => 'textarea', '#rows' => 3];
    $form['csp']['extra_sources_fs']['csp_extra_script_src'] = $common_textarea + ['#title' => $this->t('script-src'), '#default_value' => $config->get('csp_extra_script_src')];
    $form['csp']['extra_sources_fs']['csp_extra_style_src'] = $common_textarea + ['#title' => $this->t('style-src'), '#default_value' => $config->get('csp_extra_style_src')];
    $form['csp']['extra_sources_fs']['csp_extra_font_src'] = $common_textarea + ['#title' => $this->t('font-src'), '#default_value' => $config->get('csp_extra_font_src')];
    $form['csp']['extra_sources_fs']['csp_extra_connect_src'] = $common_textarea + ['#title' => $this->t('connect-src'), '#default_value' => $config->get('csp_extra_connect_src')];
    $form['csp']['extra_sources_fs']['csp_extra_img_src'] = $common_textarea + ['#title' => $this->t('img-src'), '#default_value' => $config->get('csp_extra_img_src')];
    $form['csp']['extra_sources_fs']['csp_extra_frame_src'] = $common_textarea + ['#title' => $this->t('frame-src'), '#default_value' => $config->get('csp_extra_frame_src')];

    // Advanced Settings.
    $form['advanced'] = [
      '#type' => 'details',
      '#title' => $this->t('Advanced Module Settings'),
      '#open' => FALSE,
      '#states' => ['visible' => [':input[name="enabled"]' => ['checked' => TRUE]]],
    ];
    $form['advanced']['cache_lifetime'] = [
      '#type' => 'number',
      '#title' => $this->t('API Data Cache Lifetime'),
      '#default_value' => $config->get('cache_lifetime') ?: 3600,
      '#min' => 60,
      '#step' => 60,
      '#field_suffix' => $this->t('seconds'),
      '#description' => $this->t('Duration to cache API results (subscription status, script list/hashes). Default: 3600 (1 hour).'),
    ];
    $form['advanced']['cron_update_interval'] = [
      '#type' => 'number',
      '#title' => $this->t('Cron Cache Refresh Interval'),
      '#default_value' => $config->get('cron_update_interval') ?: 21600,
      '#min' => 300,
      '#step' => 60,
      '#field_suffix' => $this->t('seconds'),
      '#description' => $this->t('How often cron should force a refresh of cached API data. Default: 21600 (6 hours).'),
    ];
    $form['advanced']['debug_mode'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Enable Debug Mode'),
      '#default_value' => $config->get('debug_mode'),
      '#description' => $this->t('Log detailed information to Drupal logs and browser console.'),
    ];

    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state): void {
    $values = $form_state->getValues();
    $enabled = $values['enabled'];
    $new_api_key = $values['api_key_new'];

    if (!empty($new_api_key) && empty($this->settings->get('nesguard_encryption_key'))) {
      $form_state->setErrorByName('api_key_new', $this->t('Cannot save API Key: The NESGuard encryption key ($settings["nesguard_encryption_key"]) is missing in settings.php.'));
    }

    if ($enabled) {
      if (empty($values['agency_id'])) {
        $form_state->setErrorByName('agency_id', $this->t('Agency Identifier is required when enabling NESGuard integration.'));
      }
      if (empty($new_api_key) && empty($this->config('nesguard_drupal.settings')->get('api_key'))) {
        $form_state->setErrorByName('api_key_new', $this->t('API Key is required for initial setup when enabling NESGuard integration.'));
      }
    }

    parent::validateForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state): void {
    $config = $this->config('nesguard_drupal.settings');
    $values = $form_state->getValues();
    $new_api_key_plaintext = $values['api_key_new'];
    $old_agency_id = $config->get('agency_id');
    $new_agency_id = $values['agency_id'];

    $config
      ->set('enabled', $values['enabled'])
      ->set('api_email', $values['api_email_display'])
      ->set('agency_id', $new_agency_id)
      ->set('enforce_sri', $values['enforce_sri'])
      ->set('csp_enabled', $values['csp_enabled'])
      ->set('csp_report_only', $values['csp_report_only'])
      ->set('csp_allow_self', $values['csp_allow_self'])
      ->set('csp_allow_unsafe_inline_style', $values['csp_allow_unsafe_inline_style'])
      ->set('csp_allow_data_img', $values['csp_allow_data_img'])
      ->set('csp_allow_google_fonts', $values['csp_allow_google_fonts'])
      ->set('csp_allow_stripe', $values['csp_allow_stripe'])
      ->set('csp_allow_nesguard', $values['csp_allow_nesguard'])
      ->set('csp_extra_script_src', $values['csp_extra_script_src'])
      ->set('csp_extra_style_src', $values['csp_extra_style_src'])
      ->set('csp_extra_font_src', $values['csp_extra_font_src'])
      ->set('csp_extra_connect_src', $values['csp_extra_connect_src'])
      ->set('csp_extra_img_src', $values['csp_extra_img_src'])
      ->set('csp_extra_frame_src', $values['csp_extra_frame_src'])
      ->set('cache_lifetime', $values['cache_lifetime'])
      ->set('cron_update_interval', $values['cron_update_interval'])
      ->set('debug_mode', $values['debug_mode']);

    // Handle API key encryption/saving.
    if (!empty($new_api_key_plaintext)) {
      $encryption_key = $this->settings->get('nesguard_encryption_key');
      $encrypted_key = NULL;
      try {
        $cipher = 'aes-256-cbc';
        $key_material = substr(hash('sha256', $encryption_key, TRUE), 0, 32);
        $iv_length = openssl_cipher_iv_length($cipher);
        $iv = openssl_random_pseudo_bytes($iv_length);
        // Use OPENSSL_RAW_DATA for binary output, then base64 encode.
        $ciphertext = openssl_encrypt($new_api_key_plaintext, $cipher, $key_material, OPENSSL_RAW_DATA, $iv);

        if ($ciphertext === FALSE) {
          $this->logger->error('API Key encryption failed. Error: @error', ['@error' => openssl_error_string()]);
        }
        else {
          $encrypted_key = base64_encode($iv . $ciphertext);
        }
      }
      catch (\Throwable $e) {
        $this->logger->error('Exception during API key encryption: @error', ['@error' => $e->getMessage()]);
      }

      if ($encrypted_key !== NULL) {
        $config->set('api_key', $encrypted_key);
        $this->messenger()->addStatus($this->t('API Key saved and encrypted.'));
        $this->authService->clearTokens();
      }
      else {
        $this->messenger()->addError($this->t('Failed to encrypt the API Key. Key was NOT saved. Check logs and encryption key in settings.php.'));
      }
    }

    if ($new_agency_id !== $old_agency_id) {
      $this->authService->clearTokens();
      $this->messenger()->addStatus($this->t('Agency Identifier updated. Authentication tokens cleared.'));
    }

    $config->save();

    // Clear service caches as settings may affect API calls.
    $this->subscriptionService->clearCache();
    $this->sriService->clearCache();

    $this->cacheTagsInvalidator->invalidateTags(['library_info']);

    parent::submitForm($form, $form_state);
    $this->messenger()->addStatus($this->t('NESGuard settings saved.'));
  }

}
