<?php

namespace Drupal\nesguard_drupal\Service;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\State\StateInterface;
use Psr\Log\LoggerInterface;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Exception\RequestException;
use Drupal\Core\Site\Settings;
use Drupal\Core\Lock\LockBackendInterface;

/**
 * Handles authentication with the NESGuard API using API Key / Agency ID.
 */
class NesguardAuthService {

  public const API_BASE_URL = 'https://nesgaurd.cloud/api/v1/';
  protected const TOKEN_STATE_KEY = 'nesguard_drupal.api_key_token_data';
  protected const AUTH_LOCK_KEY = 'nesguard_drupal.api_key_auth_lock';
  protected const REFRESH_LOCK_KEY = 'nesguard_drupal.api_key_refresh_lock';
  protected const REFRESH_COOLDOWN_KEY = 'nesguard_drupal.last_token_refresh';
  protected const REFRESH_COOLDOWN_SECONDS = 600;
  // 12 hours.
  protected const ASSUMED_TOKEN_VALIDITY_SECONDS = 43200;
  protected const RATE_LIMIT_LOGIN_KEY = 'nesguard_drupal.auth.login.limit_until';
  protected const RATE_LIMIT_REFRESH_KEY = 'nesguard_drupal.auth.refresh.limit_until';
  protected const RATE_LIMIT_DEFAULT_DURATION = 60;

  /**
   * The config object.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected $config;

  /**
   * The state service.
   *
   * @var \Drupal\Core\State\StateInterface
   */
  protected $state;

  /**
   * The logger channel.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * The HTTP client.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected $httpClient;

  /**
   * The site settings service.
   *
   * @var \Drupal\Core\Site\Settings
   */
  protected $settings;

  /**
   * The lock service.
   *
   * @var \Drupal\Core\Lock\LockBackendInterface
   */
  protected $lock;

  /**
   * Constructs a NesguardAuthService object.
   */
  public function __construct(
    ConfigFactoryInterface $config_factory,
    StateInterface $state,
    LoggerInterface $logger,
    ClientInterface $http_client,
    Settings $settings,
    LockBackendInterface $lock,
  ) {
    $this->config = $config_factory->get('nesguard_drupal.settings');
    $this->state = $state;
    $this->logger = $logger;
    $this->httpClient = $http_client;
    $this->settings = $settings;
    $this->lock = $lock;
  }

  /**
   * Checks if a specific action is currently rate limited.
   */
  protected function isRateLimited(string $limit_key): bool {
    $limited_until = $this->state->get($limit_key, 0);
    if (time() < $limited_until) {
      if ($this->config->get('debug_mode')) {
        $this->logger->debug('Action (@key) is rate limited until @time.', [
          '@key' => $limit_key,
          // Formatting the timestamp can be helpful for debugging.
          '@time' => date('Y-m-d H:i:s T', $limited_until),
        ]);
      }
      return TRUE;
    }

    return FALSE;
  }

  /**
   * Sets a rate limit for a specific action.
   */
  protected function setRateLimit(string $limit_key, int $duration = self::RATE_LIMIT_DEFAULT_DURATION): void {
    $limited_until = time() + $duration;
    $this->state->set($limit_key, $limited_until);
    $this->logger->warning('Rate limit set for action (@key) until @time (@duration s).', [
      '@key' => $limit_key,
      '@time' => date('Y-m-d H:i:s', $limited_until),
      '@duration' => $duration,
    ]);
  }

  /**
   * Clears a rate limit for a specific action.
   */
  protected function clearRateLimit(string $limit_key): void {
    if ($this->state->get($limit_key)) {
      $this->state->delete($limit_key);
      if ($this->config->get('debug_mode')) {
        $this->logger->debug('Cleared rate limit for action (@key).', ['@key' => $limit_key]);
      }
    }
  }

  /**
   * Checks if API credentials (Agency ID / API Key) are configured.
   */
  public function hasCredentials(): bool {
    return !empty($this->config->get('agency_id')) && !empty($this->config->get('api_key'));
  }

  /**
   * Decrypts the stored API key.
   *
   * @return string|null
   *   The decrypted API key or NULL on failure.
   */
  protected function decryptApiKey(): ?string {
    $encrypted_api_key = $this->config->get('api_key');
    if (empty($encrypted_api_key)) {
      return NULL;
    }

    // Check both old and new key names for backward compatibility.
    $encryption_key = $this->settings->get('nesguard_encryption_key')
    // Corrected spelling.
      ?? $this->settings->get('nesguard_encryption_key')
      ?? $this->state->get('nesguard_drupal.encryption_key')
    // Corrected spelling.
      ?? $this->state->get('nesguard_drupal.encryption_key');

    if (empty($encryption_key)) {
      $this->logger->error('Cannot decrypt API key: Encryption key missing from settings.php and state.');
      return NULL;
    }

    // If the old key name is used, issue a deprecation warning.
    if ($this->settings->get('nesguard_encryption_key') && !$this->settings->get('nesguard_encryption_key')) {
      $this->logger->warning('Using deprecated encryption key name "nesguard_encryption_key". Please update to "nesguard_encryption_key" in settings.php.');
    }

    // If key came from state, issue warning.
    if (!$this->settings->get('nesguard_encryption_key') && !$this->settings->get('nesguard_encryption_key') &&
        ($this->state->get('nesguard_drupal.encryption_key') || $this->state->get('nesguard_drupal.encryption_key'))) {
      $this->logger->warning('Using encryption key from State API for decryption. Move to settings.php for better security.');
    }

    // Basic check for potentially unencrypted legacy keys.
    if (strlen($encrypted_api_key) < 64 && !str_contains($encrypted_api_key, '+') && !str_contains($encrypted_api_key, '/') && !str_contains($encrypted_api_key, '=')) {
      $this->logger->notice('Using potentially plaintext API key (short, no base64 chars).');
      return $encrypted_api_key;
    }

    try {
      $cipher = 'aes-256-cbc';
      $key_material = substr(hash('sha256', $encryption_key, TRUE), 0, 32);
      // Use strict decoding.
      $decoded_data = base64_decode($encrypted_api_key, TRUE);

      if ($decoded_data === FALSE) {
        $this->logger->error('Invalid encrypted API key format. Base64 decoding failed.');
        return NULL;
      }

      $iv_length = openssl_cipher_iv_length($cipher);
      if (strlen($decoded_data) <= $iv_length) {
        $this->logger->error('Encrypted API key data too short (length: @len) to contain IV (length: @iv_len).', ['@len' => strlen($decoded_data), '@iv_len' => $iv_length]);
        return NULL;
      }

      $iv = substr($decoded_data, 0, $iv_length);
      $ciphertext = substr($decoded_data, $iv_length);

      if ($iv === FALSE || $ciphertext === FALSE) {
        $this->logger->error('Failed to extract IV or ciphertext from encrypted data.');
        return NULL;
      }

      // Use OPENSSL_RAW_DATA for compatibility with binary IV/ciphertext.
      $decrypted = openssl_decrypt($ciphertext, $cipher, $key_material, OPENSSL_RAW_DATA, $iv);

      if ($decrypted === FALSE) {
        $error_msg = openssl_error_string();
        $this->logger->error('Failed to decrypt API key. Check encryption key. OpenSSL Error: @error', ['@error' => $error_msg]);

        // For more user-friendly debugging, set a state variable with the error.
        $this->state->set('nesguard_drupal.last_decryption_error', [
          'time' => time(),
          'error' => $error_msg,
          'using_legacy_key' => (bool) $this->settings->get('nesguard_encryption_key'),
        ]);

        return NULL;
      }

      // Successful decryption - clear any previous error.
      $this->state->delete('nesguard_drupal.last_decryption_error');
      return $decrypted;
    }
    catch (\Throwable $e) {
      $error_message = $e->getMessage();
      $this->logger->error('Exception during API key decryption: @error', ['@error' => $error_message]);

      // Store detailed error information for status reporting.
      $this->state->set('nesguard_drupal.last_decryption_error', [
        'time' => time(),
        'error' => $error_message,
        'exception' => get_class($e),
        'using_legacy_key' => (bool) $this->settings->get('nesguard_encryption_key'),
      ]);

      return NULL;
    }
  }

  /**
   * Gets currently stored authentication tokens.
   */
  protected function getStoredTokens(): ?array {
    $data = $this->state->get(self::TOKEN_STATE_KEY);
    if (is_array($data) && isset($data['access'], $data['refresh'], $data['updated_at'])) {
      $data['is_likely_valid'] = (time() - $data['updated_at']) < self::ASSUMED_TOKEN_VALIDITY_SECONDS;
      return $data;
    }
    return NULL;
  }

  /**
   * Stores authentication tokens.
   */
  protected function storeTokens(string $access_token, ?string $refresh_token): void {
    $data = [
      'access' => $access_token,
      'refresh' => $refresh_token ?? '',
      'updated_at' => time(),
    ];
    $this->state->set(self::TOKEN_STATE_KEY, $data);
  }

  /**
   * Clears stored authentication tokens and cooldown timer.
   */
  public function clearTokens(): void {
    $this->state->delete(self::TOKEN_STATE_KEY);
    $this->state->delete(self::REFRESH_COOLDOWN_KEY);
    $this->logger->info('Cleared stored authentication tokens.');
  }

  /**
   * Gets a valid access token, handling expiry, refresh, locking, rate limits.
   */
  public function getAccessToken(bool $force_refresh = FALSE): ?string {
    if (!$this->config->get('enabled') || !$this->hasCredentials()) {
      return NULL;
    }

    $stored_tokens = $this->getStoredTokens();

    if (!$force_refresh && $stored_tokens && $stored_tokens['is_likely_valid']) {
      return $stored_tokens['access'];
    }

    // Attempt Refresh first if needed/possible, checking rate limit.
    if ($stored_tokens && !empty($stored_tokens['refresh'])) {
      $now = time();
      $last_refresh_attempt = $this->state->get(self::REFRESH_COOLDOWN_KEY, 0);
      if (($force_refresh || !$stored_tokens['is_likely_valid']) && ($now - $last_refresh_attempt > self::REFRESH_COOLDOWN_SECONDS)) {
        if ($this->isRateLimited(self::RATE_LIMIT_REFRESH_KEY)) {
          $this->logger->warning('Token refresh skipped due to active rate limit.');
        }
        else {
          $refreshed_token = $this->performTokenRefreshWithLock($stored_tokens['refresh']);
          if ($refreshed_token !== NULL) {
            return $refreshed_token;
          }
        }
      }
      elseif ($stored_tokens['is_likely_valid']) {
        return $stored_tokens['access'];
      }
    }

    // Attempt Login if needed, checking rate limit.
    $current_tokens = $this->getStoredTokens();
    if ($current_tokens === NULL || !$current_tokens['is_likely_valid']) {
      if ($this->isRateLimited(self::RATE_LIMIT_LOGIN_KEY)) {
        $this->logger->error('Login skipped due to active rate limit.');
        return NULL;
      }
      return $this->performLoginWithLock();
    }
    else {
      return $current_tokens['access'];
    }
  }

  /**
   * Performs login with locking.
   */
  protected function performLoginWithLock(): ?string {
    if (!$this->lock->acquire(self::AUTH_LOCK_KEY, 30)) {
      if ($this->lock->wait(self::AUTH_LOCK_KEY, 5)) {
        $tokens = $this->getStoredTokens();
        if ($tokens && $tokens['is_likely_valid']) {
          return $tokens['access'];
        }
      }
      else {
        $this->logger->error('Auth lock timeout during login attempt.');
        return NULL;
      }
    }
    $new_token = NULL;
    try {
      $new_token = $this->performLogin();
    }
    finally {
      $this->lock->release(self::AUTH_LOCK_KEY);
    }
    return $new_token;
  }

  /**
   * Performs token refresh with locking and cooldown update.
   */
  protected function performTokenRefreshWithLock(string $refresh_token): ?string {
    $this->state->set(self::REFRESH_COOLDOWN_KEY, time());

    if (!$this->lock->acquire(self::REFRESH_LOCK_KEY, 30)) {
      if ($this->lock->wait(self::REFRESH_LOCK_KEY, 3)) {
        $tokens = $this->getStoredTokens();
        if ($tokens && $tokens['is_likely_valid']) {
          return $tokens['access'];
        }
      }
      else {
        $this->logger->error('Refresh lock timeout during refresh attempt.');
        return NULL;
      }
    }
    $new_token = NULL;
    try {
      $new_token = $this->performTokenRefresh($refresh_token);
    }
    finally {
      $this->lock->release(self::REFRESH_LOCK_KEY);
    }
    return $new_token;
  }

  /**
   * Performs the actual login API call using API Key/Agency ID.
   */
  protected function performLogin(): ?string {
    $api_key = $this->decryptApiKey();
    $agency_id = $this->config->get('agency_id');
    if (empty($api_key) || empty($agency_id)) {
      $this->logger->error('Cannot perform login: Agency ID or decrypted API Key is missing.');
      return NULL;
    }
    // Re-check rate limit just before call.
    if ($this->isRateLimited(self::RATE_LIMIT_LOGIN_KEY)) {
      return NULL;
    }

    $this->logger->info('Attempting API login via API Key for agency @agency', ['@agency' => $agency_id]);
    try {
      $response = $this->httpClient->request('POST', self::API_BASE_URL . 'auth/login/', [
        'json' => ['api_key' => $api_key, 'agency_id' => $agency_id],
        'headers' => ['Accept' => 'application/json', 'Content-Type' => 'application/json'],
        'timeout' => 15,
        'connect_timeout' => 5,
        'http_errors' => FALSE,
      ]);

      $status_code = $response->getStatusCode();
      $body = $response->getBody()->getContents();
      $data = json_decode($body, TRUE);

      // Validate JSON parsing succeeded.
      if ($data === NULL && json_last_error() !== JSON_ERROR_NONE) {
        $json_error = json_last_error_msg();
        $this->logger->error('Failed to parse API login response JSON: @error. Response body: @body', [
          '@error' => $json_error,
          '@body' => substr($body, 0, 255) . (strlen($body) > 255 ? '...[truncated]' : ''),
        ]);
        // Return null on parse failure.
        return NULL;
      }

      $access_token_new = $data['data']['token'] ?? ($data['access_token'] ?? ($data['access'] ?? NULL));
      $refresh_token_new = $data['data']['refresh_token'] ?? ($data['refresh'] ?? NULL);

      if ($status_code == 200 && $access_token_new !== NULL) {
        $this->storeTokens($access_token_new, $refresh_token_new);
        $this->clearRateLimit(self::RATE_LIMIT_LOGIN_KEY);
        $this->logger->info('API login successful for @agency.', ['@agency' => $agency_id]);
        return $access_token_new;
      }
      elseif ($status_code == 429) {
        $this->logger->warning('API login failed for @agency: Rate Limited (429).', ['@agency' => $agency_id]);
        $this->setRateLimit(self::RATE_LIMIT_LOGIN_KEY);
        return NULL;
      }
      else {
        $error_message = $data['error']['message'] ?? ($data['message'] ?? $body);
        $this->logger->error(
          'API login failed for @agency: Status @status - @message',
          [
            '@agency' => $agency_id,
            '@status' => $status_code,
            '@message' => $error_message,
          ]
        );
        $this->clearTokens();
        return NULL;
      }
    }
    catch (RequestException $e) {
      $this->logger->error(
        'API login network error for @agency: @error',
        ['@agency' => $agency_id, '@error' => $e->getMessage()]
      );
      return NULL;
    }
    catch (\Throwable $e) {
      $this->logger->error(
        'API login general exception for @agency: @error',
        ['@agency' => $agency_id, '@error' => $e->getMessage()]
      );
      return NULL;
    }
  }

  /**
   * Performs the actual token refresh API call using API Key/Agency ID flow.
   */
  protected function performTokenRefresh(string $refresh_token): ?string {
    $agency_id = $this->config->get('agency_id');
    if (empty($agency_id)) {
      $this->logger->error('Cannot refresh token: Agency ID missing.');
      return NULL;
    }
    // Re-check rate limit just before call.
    if ($this->isRateLimited(self::RATE_LIMIT_REFRESH_KEY)) {
      return NULL;
    }

    $this->logger->info('Attempting API token refresh for agency @agency', ['@agency' => $agency_id]);
    try {
      $response = $this->httpClient->request('POST', self::API_BASE_URL . 'auth/token/refresh/', [
        'json' => ['refresh_token' => $refresh_token, 'agency_id' => $agency_id],
        'headers' => ['Accept' => 'application/json', 'Content-Type' => 'application/json'],
        'timeout' => 15,
        'connect_timeout' => 5,
        'http_errors' => FALSE,
      ]);

      $status_code = $response->getStatusCode();
      $body = $response->getBody()->getContents();
      $data = json_decode($body, TRUE);

      // Validate JSON parsing succeeded.
      if ($data === NULL && json_last_error() !== JSON_ERROR_NONE) {
        $json_error = json_last_error_msg();
        $this->logger->error('Failed to parse API token refresh response JSON: @error. Response body: @body', [
          '@error' => $json_error,
          '@body' => substr($body, 0, 255) . (strlen($body) > 255 ? '...[truncated]' : ''),
        ]);
        // Return null on parse failure.
        return NULL;
      }

      $access_token_new = $data['data']['token'] ?? ($data['access_token'] ?? ($data['access'] ?? NULL));
      $refresh_token_new = $data['data']['refresh_token'] ?? ($data['refresh'] ?? NULL);

      if ($status_code == 200 && $access_token_new !== NULL) {
        $this->storeTokens($access_token_new, $refresh_token_new ?: $refresh_token);
        $this->clearRateLimit(self::RATE_LIMIT_REFRESH_KEY);
        $this->logger->info('API token refresh successful for @agency.', ['@agency' => $agency_id]);
        return $access_token_new;
      }
      elseif ($status_code == 429) {
        $this->logger->warning('API token refresh failed for @agency: Rate Limited (429).', ['@agency' => $agency_id]);
        $this->setRateLimit(self::RATE_LIMIT_REFRESH_KEY);
        return NULL;
      }
      else {
        $error_message = $data['error']['message'] ?? ($data['message'] ?? $body);
        $this->logger->warning(
          'API token refresh failed for @agency: Status @status - @message.',
          [
            '@agency' => $agency_id,
            '@status' => $status_code,
            '@message' => $error_message,
          ]
        );
        if ($status_code == 401 || $status_code == 403) {
          $this->logger->error('Clearing tokens due to refresh failure (401/403).');
          $this->clearTokens();
        }
        return NULL;
      }
    }
    catch (RequestException $e) {
      $this->logger->error(
        'API token refresh network error for @agency: @error.',
        ['@agency' => $agency_id, '@error' => $e->getMessage()]
      );
      return NULL;
    }
    catch (\Throwable $e) {
      $this->logger->error(
        'API token refresh general exception for @agency: @error.',
        ['@agency' => $agency_id, '@error' => $e->getMessage()]
      );
      return NULL;
    }
  }

}
