<?php

namespace Drupal\nesguard_drupal\Service;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Config\ImmutableConfig;
use Drupal\Core\State\StateInterface;
// Use self-namespaced services.
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Exception\ConnectException;
use GuzzleHttp\Exception\RequestException;
// Use constant for options.
use GuzzleHttp\RequestOptions;
use Psr\Log\LoggerInterface;

/**
 * Manages SRI data (monitored scripts, hashes) from the NesGuard API.
 *
 * Fetches the list of monitored scripts and their current/previous hashes for
 * an agency, handling authentication, caching, and rate limiting. Includes an
 * emergency fallback for Stripe JS hash calculation.
 *
 * @internal
 * This service is intended for internal use by the NesGuard Drupal module.
 * It relies on NesguardAuthService for authentication and configuration
 * provided via nesguard_drupal.settings.
 */
class NesguardSriService {

  // Constants for API, caching, rate limiting, and fallback.
  // IMPORTANT: Must match the exact spelling used in NesguardAuthService and NesguardSubscriptionService.
  public const API_BASE_URL = 'https://nesgaurd.cloud/api/v1/';
  public const STRIPE_JS_URL = 'https://js.stripe.com/v3';

  protected const SCRIPT_DATA_CACHE_KEY = 'nesguard_drupal.sri_script_data';
  // 4 hours (increased from 1 hour to reduce API calls)
  protected const DEFAULT_CACHE_LIFETIME_SECONDS = 14400;
  protected const RATE_LIMIT_FETCH_KEY = 'nesguard_drupal.sri.fetch.limit_until';
  // 60 seconds.
  protected const RATE_LIMIT_DEFAULT_DURATION = 60;
  protected const STRIPE_FALLBACK_HASH_KEY = 'nesguard_drupal.stripe_fallback_hash';
  protected const STRIPE_FALLBACK_TIME_KEY = 'nesguard_drupal.stripe_fallback_time';
  // 10 minutes for fallback hash (increased from 2 minutes)
  protected const STRIPE_FALLBACK_CACHE_SECONDS = 600;
  // Number of retries for fallback fetch.
  protected const FALLBACK_FETCH_RETRIES = 1;
  // Delay between fallback retries.
  protected const FALLBACK_RETRY_DELAY_MS = 500;

  /**
   * The immutable configuration object ('nesguard_drupal.settings').
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected ImmutableConfig $config;

  /**
   * The state service.
   *
   * @var \Drupal\Core\State\StateInterface
   */
  protected StateInterface $state;

  /**
   * The logger service for the 'nesguard_drupal' channel.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected LoggerInterface $logger;

  /**
   * The HTTP client service.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected ClientInterface $httpClient;

  /**
   * The NesGuard authentication service.
   *
   * @var \Drupal\nesguard_drupal\Service\NesguardAuthService
   */
  protected NesguardAuthService $authService;

  /**
   * The NesGuard subscription service.
   *
   * @var \Drupal\nesguard_drupal\Service\NesguardSubscriptionService
   */
  protected NesguardSubscriptionService $subscriptionService;

  /**
   * Constructs a NesguardSriService object.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The configuration factory service.
   * @param \Drupal\Core\State\StateInterface $state
   *   The state service.
   * @param \Psr\Log\LoggerInterface $logger
   *   The logger service, pre-configured for the 'nesguard_drupal' channel.
   * @param \GuzzleHttp\ClientInterface $http_client
   *   The HTTP client service.
   * @param \Drupal\nesguard_drupal\Service\NesguardAuthService $auth_service
   *   The NesGuard authentication service.
   * @param \Drupal\nesguard_drupal\Service\NesguardSubscriptionService $subscription_service
   *   The NesGuard subscription service.
   */
  public function __construct(
    ConfigFactoryInterface $config_factory,
    StateInterface $state,
    LoggerInterface $logger,
    ClientInterface $http_client,
    NesguardAuthService $auth_service,
    NesguardSubscriptionService $subscription_service,
  ) {
    $this->config = $config_factory->get('nesguard_drupal.settings');
    $this->state = $state;
    $this->logger = $logger;
    $this->httpClient = $http_client;
    $this->authService = $auth_service;
    $this->subscriptionService = $subscription_service;
  }

  // --- Rate Limiting Helpers ---

  /**
   * Checks if a specific action is currently rate limited.
   *
   * @param string $limit_key
   *   The state key used for tracking the rate limit timestamp.
   *
   * @return bool
   *   TRUE if rate limited, FALSE otherwise.
   */
  protected function isRateLimited(string $limit_key): bool {
    $limited_until = $this->state->get($limit_key, 0);
    $is_limited = time() < $limited_until;
    if ($is_limited && $this->config->get('debug_mode')) {
      $this->logger->debug('Action (@key) is rate limited until @time.', [
        '@key' => $limit_key,
        '@time' => date('Y-m-d H:i:s T', $limited_until),
      ]);
    }
    return $is_limited;
  }

  /**
   * Sets a rate limit for a specific action.
   *
   * @param string $limit_key
   *   The state key to store the rate limit timestamp.
   * @param int $duration
   *   The duration of the rate limit in seconds.
   */
  protected function setRateLimit(string $limit_key, int $duration = self::RATE_LIMIT_DEFAULT_DURATION): void {
    $limited_until = time() + $duration;
    $this->state->set($limit_key, $limited_until);
    $this->logger->warning('Rate limit set for action (@key) until @time (@duration s).', [
      '@key' => $limit_key,
      '@time' => date('Y-m-d H:i:s T', $limited_until),
      '@duration' => $duration,
    ]);
  }

  /**
   * Clears a rate limit for a specific action.
   *
   * @param string $limit_key
   *   The state key for the rate limit to clear.
   */
  protected function clearRateLimit(string $limit_key): void {
    if ($this->state->get($limit_key)) {
      $this->state->delete($limit_key);
      if ($this->config->get('debug_mode')) {
        $this->logger->debug('Cleared rate limit for action (@key).', ['@key' => $limit_key]);
      }
    }
  }

  // --- Hashing Helper ---

  /**
   * Calculates the SHA-384 hash of content for SRI.
   *
   * @param string|null $content
   *   The content to hash.
   *
   * @return string|null
   *   The 'sha384-' prefixed, base64-encoded hash, or NULL on failure.
   */
  protected function calculateSha384(?string $content): ?string {
    if ($content === NULL) {
      return NULL;
    }
    // Calculate raw binary hash.
    $binary_hash = hash('sha384', $content, TRUE);
    if ($binary_hash === FALSE) {
      $this->logger->error('SHA384 hash calculation failed.');
      return NULL;
    }
    // Base64 encode the raw binary hash.
    $base64_hash = base64_encode($binary_hash);
    $normalized_hash = 'sha384-' . $base64_hash;

    // Validate the final format (redundant check if normalization function exists).
    if (function_exists('nesguard_drupal_normalize_hash')) {
      // Reuse the global helper for final validation/normalization.
      return nesguard_drupal_normalize_hash($normalized_hash);
    }

    // Basic fallback validation if helper doesn't exist.
    return preg_match('/^sha384-[A-Za-z0-9+\/=]{64}$/', $normalized_hash) ? $normalized_hash : NULL;
  }

  // --- Core Service Methods ---

  /**
   * Fetches monitored scripts and their hashes (current and previous) from API.
   *
   * Handles caching, rate limiting, and retries once on 401 Unauthorized.
   *
   * @param bool $bypass_cache
   *   If TRUE, forces a fetch from the API, ignoring cached data.
   *
   * @return array|null
   *   An associative array mapping normalized URL => ['current' => HASH, 'previous' => HASH|null].
   *   Returns NULL on critical failure (e.g., no creds, initial auth fail).
   *   Returns potentially stale cache data if rate limited or API fails after retry.
   */
  public function getMonitoredScriptHashes(bool $bypass_cache = FALSE): ?array {
    $debug = $this->config->get('debug_mode') ?? FALSE;
    if ($debug) {
      $this->logger->debug('SRI Service: getMonitoredScriptHashes called (bypass_cache: @bypass).', ['@bypass' => $bypass_cache ? 'Yes' : 'No']);
    }

    if (!$this->config->get('enabled') || !$this->authService->hasCredentials()) {
      if ($debug) {
        $log_reason = !$this->config->get('enabled') ? 'Module disabled' : 'No credentials';
        $this->logger->debug('SRI Service: getMonitoredScriptHashes returning NULL ({$log_reason}).', ['@reason' => $log_reason]);
      }
      return NULL;
    }

    $cache_lifetime = $this->config->get('cache_lifetime') ?: self::DEFAULT_CACHE_LIFETIME_SECONDS;
    $cache_key = self::SCRIPT_DATA_CACHE_KEY;
    $cached_data = NULL;
    // Initialize stale hashes.
    $stale_hashes = NULL;

    if (!$bypass_cache) {
      $cached_data = $this->state->get($cache_key);
      if ($cached_data !== NULL && isset($cached_data['timestamp'], $cached_data['hashes']) && (time() - $cached_data['timestamp'] < $cache_lifetime)) {
        if ($debug) {
          $this->logger->debug('SRI Service: getMonitoredScriptHashes returning valid cache data (@count items).', [
            '@count' => is_array($cached_data['hashes']) ? count($cached_data['hashes']) : 0,
          ]);
        }
        // Return fresh cache data.
        return $cached_data['hashes'];
      }
      // Keep potentially stale hashes if cache exists but is expired.
      $stale_hashes = $cached_data['hashes'] ?? NULL;
      if ($debug && $cached_data !== NULL) {
        $this->logger->debug('SRI Service: getMonitoredScriptHashes cache expired or invalid.');
      }
    }
    // If bypassing cache, still check if stale data exists for fallback.
    elseif ($cached_data = $this->state->get($cache_key)) {
      $stale_hashes = $cached_data['hashes'] ?? NULL;
    }

    // Check rate limiting after cache check.
    if ($this->isRateLimited(self::RATE_LIMIT_FETCH_KEY)) {
      $this->logger->warning('SRI script/hash fetch skipped due to rate limit. Returning stale cache if available.');
      // Return stale data if rate limited.
      return $stale_hashes;
    }

    // Perform API Call with Retry on 401.
    $max_attempts = 2;
    $attempt = 0;
    $hash_map = NULL;
    $agency_id = $this->config->get('agency_id');
    if (empty($agency_id)) {
      $this->logger->error('Cannot fetch SRI scripts/hashes: Agency ID missing in configuration.');
      // Cannot proceed without Agency ID.
      return NULL;
    }

    while ($attempt < $max_attempts && $hash_map === NULL) {
      $attempt++;
      if ($debug) {
        $this->logger->debug('SRI Service: Fetch attempt #@attempt for agency @agency.', ['@attempt' => $attempt, '@agency' => $agency_id]);
      }

      // Force token refresh only on retry (attempt > 1).
      $force_refresh = ($attempt > 1);
      $access_token = $this->authService->getAccessToken($force_refresh);

      if (!$access_token) {
        $this->logger->warning('Cannot fetch SRI scripts/hashes (Attempt @attempt): Authentication failed.', ['@attempt' => $attempt]);
        // Do not retry if auth fails; return stale data immediately.
        return $stale_hashes;
      }

      try {
        $api_base_url = self::API_BASE_URL;
        $endpoint_url = rtrim($api_base_url, '/') . '/sri/client/scripts/';
        if ($debug) {
          $this->logger->debug('SRI Service: Calling @url (Attempt @attempt)', ['@url' => $endpoint_url, '@attempt' => $attempt]);
        }

        $drupal_version = \Drupal::VERSION;
        $response = $this->httpClient->request('GET', $endpoint_url, [
          RequestOptions::HEADERS => [
            'Authorization' => 'Bearer ' . $access_token,
            'X-Agency-ID' => $agency_id,
            'Accept' => 'application/json',
            'User-Agent' => 'NesGuard Drupal/' . $drupal_version,
          ],
          // Total request timeout.
          // Reduce from 15s to 10s to avoid long hanging requests.
          RequestOptions::TIMEOUT => 10,
          // Connection timeout - reduce from 5s to 3s.
          RequestOptions::CONNECT_TIMEOUT => 3,
          // Handle errors manually.
          RequestOptions::HTTP_ERRORS => FALSE,
        ]);

        $status_code = $response->getStatusCode();
        $body = $response->getBody()->getContents();
        $data = json_decode($body, TRUE);

        // Validate JSON parsing succeeded.
        if ($data === NULL && json_last_error() !== JSON_ERROR_NONE) {
          $json_error = json_last_error_msg();
          $this->logger->error('SRI Service: Failed to parse API response JSON: @error. Response body: @body', [
            '@error' => $json_error,
            '@body' => substr($body, 0, 255) . (strlen($body) > 255 ? '...[truncated]' : ''),
          ]);
          // Return stale data on parse failure.
          return $stale_hashes;
        }

        if ($debug) {
          $this->logger->debug('SRI Service: Attempt @attempt API Status: @status', ['@attempt' => $attempt, '@status' => $status_code]);
          // Optionally log response body in debug, be careful with size/sensitivity.
          // $this->logger->debug('SRI Service: Attempt @attempt API Body: @body', ['@attempt' => $attempt, '@body' => $body]);.
        }

        // --- Process Response ---
        if ($status_code === 200 && ($data['success'] ?? FALSE) === TRUE && isset($data['data']['scripts']) && is_array($data['data']['scripts'])) {
          // Get existing hashes from state if available
          // This allows us to maintain previous hashes through API updates.
          $existing_hash_map = [];
          $cached_data = $this->state->get($cache_key);
          if ($cached_data && !empty($cached_data['hashes']) && is_array($cached_data['hashes'])) {
            $existing_hash_map = $cached_data['hashes'];
          }

          $hash_map = [];
          foreach ($data['data']['scripts'] as $script_info) {
            // Ensure required fields exist.
            if (isset($script_info['url']) && (isset($script_info['currentHash']) || isset($script_info['current_hash']) || isset($script_info['hash']))) {
              $normalized_url = function_exists('nesguard_drupal_normalize_url') ? nesguard_drupal_normalize_url($script_info['url']) : $script_info['url'];
              $current_hash_raw = $script_info['currentHash'] ?? $script_info['current_hash'] ?? $script_info['hash'] ?? NULL;
              $previous_hash_raw = $script_info['previousHash'] ?? $script_info['previous_hash'] ?? $script_info['old_hash'] ?? NULL;

              // Normalize hashes.
              $normalized_current = function_exists('nesguard_drupal_normalize_hash') ? nesguard_drupal_normalize_hash($current_hash_raw) : $current_hash_raw;
              $normalized_previous = function_exists('nesguard_drupal_normalize_hash') ? nesguard_drupal_normalize_hash($previous_hash_raw) : $previous_hash_raw;

              if ($normalized_url && $normalized_current) {
                // Check if this URL already has a hash in our existing data.
                $existing_hash_data = $existing_hash_map[$normalized_url] ?? NULL;
                $existing_current = $existing_hash_data['current'] ?? NULL;

                // If we already have a hash and it's different than the new one,
                // the current hash becomes the previous hash (transition handling)
                if ($existing_current && $existing_current !== $normalized_current) {
                  // New hash is different than what we had - use existing as previous.
                  $hash_map[$normalized_url] = [
                    'current' => $normalized_current,
                    'previous' => $existing_current,
                    'updated' => time(),
                    'source' => 'api',
                  ];

                  if ($debug) {
                    $this->logger->info('SRI Service: Hash changed for @url. New: @new, Previous: @old', [
                      '@url' => $normalized_url,
                      '@new' => $normalized_current,
                      '@old' => $existing_current,
                    ]);
                  }
                }
                else {
                  // Either no existing hash or same hash - use API provided previous.
                  $hash_map[$normalized_url] = [
                    'current' => $normalized_current,
                    'previous' => ($normalized_previous && $normalized_previous !== $normalized_current) ? $normalized_previous : NULL,
                    'updated' => time(),
                    'source' => 'api',
                  ];
                }
              }
              else {
                $this->logger->warning('Invalid URL (@url) or current hash (@hash) received from API after normalization.', ['@url' => $script_info['url'] ?? '?', '@hash' => $current_hash_raw ?? '?']);
              }
            }
            else {
              $this->logger->warning('Skipping invalid script entry from API: @entry', ['@entry' => json_encode($script_info)]);
            }
          } // End foreach script.

          // Special handling for Stripe - make sure it's always in the hash map.
          $stripe_url_normalized = function_exists('nesguard_drupal_normalize_url') ?
            nesguard_drupal_normalize_url(self::STRIPE_JS_URL) : self::STRIPE_JS_URL;

          // If Stripe isn't in the map but we have a previous hash, preserve it.
          if ($stripe_url_normalized && !isset($hash_map[$stripe_url_normalized]) && isset($existing_hash_map[$stripe_url_normalized])) {
            $hash_map[$stripe_url_normalized] = $existing_hash_map[$stripe_url_normalized];

            if ($debug) {
              $this->logger->debug('SRI Service: Preserved existing Stripe hash in map: @hash', [
                '@hash' => $existing_hash_map[$stripe_url_normalized]['current'] ?? 'unknown',
              ]);
            }
          }

          // If we have Stripe in the map, update the emergency recovery variables.
          if (isset($hash_map[$stripe_url_normalized]) && !empty($hash_map[$stripe_url_normalized]['current'])) {
            $this->state->set('nesguard_drupal.stripe_latest_hash', $hash_map[$stripe_url_normalized]['current']);
            $this->state->set('nesguard_drupal.stripe_hash_update_time', time());

            if ($debug) {
              $this->logger->debug('SRI Service: Updated emergency Stripe hash from API fetch: @hash', [
                '@hash' => $hash_map[$stripe_url_normalized]['current'],
              ]);
            }
          }

          // Cache the successfully fetched data.
          $this->state->set($cache_key, [
            'hashes' => $hash_map,
            'timestamp' => time(),
            'source' => 'api',
          ]);
          $this->clearRateLimit(self::RATE_LIMIT_FETCH_KEY);
          $this->logger->info('Successfully fetched and cached @count script entries for @agency.', ['@count' => count($hash_map), '@agency' => $agency_id]);
          // Success, exit while loop.
          break;
        }
        // --- Handle Errors ---
        elseif ($status_code === 401 && $attempt < $max_attempts) {
          $this->logger->warning('SRI fetch received 401 (Attempt @attempt). Forcing token refresh and retrying.', ['@attempt' => $attempt]);
          // Loop will continue, triggering getAccessToken(TRUE).
        }
        elseif ($status_code === 429) {
          $this->logger->warning('SRI script/hash fetch failed for @agency: Rate Limited (429). Using stale cache if available.', ['@agency' => $agency_id]);
          $this->setRateLimit(self::RATE_LIMIT_FETCH_KEY);
          // Use stale cache.
          $hash_map = $stale_hashes;
          // Do not retry on 429.
          break;
        }
        else {
          // Other errors (4xx, 5xx, unexpected success=false).
          $error_message = $data['error']['message'] ?? (is_string($body) && strlen($body) < 256 ? $body : 'Unknown error structure or long body');
          $this->logger->error('Failed to fetch SRI scripts/hashes for @agency (Attempt @attempt): Status @status - @message', [
            '@agency' => $agency_id,
            '@attempt' => $attempt,
            '@status' => $status_code,
            '@message' => $error_message,
          ]);
          // Use stale cache.
          $hash_map = $stale_hashes;
          // Do not retry on other errors.
          break;
        }
      }
      // --- Catch Exceptions ---
      catch (ConnectException $e) {
        $this->logger->error('Network connection error fetching SRI scripts/hashes for @agency (Attempt @attempt): @error', ['@agency' => $agency_id, '@attempt' => $attempt, '@error' => $e->getMessage()]);
        // Use stale cache, don't retry connection errors immediately.
        $hash_map = $stale_hashes;
        // Add rate limiting to prevent excessive retries on network issues
        // 5 minutes.
        $this->setRateLimit(self::RATE_LIMIT_FETCH_KEY, 300);
        break;
      }
      catch (RequestException $e) {
        $this->logger->error('Network request error fetching SRI scripts/hashes for @agency (Attempt @attempt): @error', ['@agency' => $agency_id, '@attempt' => $attempt, '@error' => $e->getMessage()]);
        // Use stale cache, don't retry.
        $hash_map = $stale_hashes;
        // Add rate limiting to prevent excessive retries on network issues
        // 3 minutes.
        $this->setRateLimit(self::RATE_LIMIT_FETCH_KEY, 180);
        break;
      }
      catch (\Throwable $e) {
        $this->logger->error('General exception fetching SRI scripts/hashes for @agency (Attempt @attempt): @error', ['@agency' => $agency_id, '@attempt' => $attempt, '@error' => $e->getMessage()]);
        // Use stale cache, don't retry.
        $hash_map = $stale_hashes;
        break;
      }

      // Add a small delay before retrying after a 401 error.
      if ($attempt < $max_attempts && $hash_map === NULL) {
        // 250ms delay.
        usleep(250000);
      }

    } // End While loop.

    // Return the final result (could be fresh map, NULL, or stale cache).
    return $hash_map;
  }

  /**
   * Fetches Stripe JS content directly and calculates its CURRENT hash.
   *
   * This is an emergency fallback used if the API fails or doesn't provide
   * the Stripe hash. Uses a short-term cache for the calculated hash.
   *
   * @return string|null
   *   The calculated sha384 hash ('sha384-...') or NULL on failure.
   */
  protected function fetchAndCalculateStripeHashFallback(): ?string {
    $debug = $this->config->get('debug_mode') ?? FALSE;
    if ($debug) {
      $this->logger->debug('SRI Service FALLBACK: Entering fetchAndCalculateStripeHashFallback.');
    }

    $cache_key = self::STRIPE_FALLBACK_HASH_KEY;
    $cache_time_key = self::STRIPE_FALLBACK_TIME_KEY;
    $cache_ttl = self::STRIPE_FALLBACK_CACHE_SECONDS;

    // Check short-term cache first.
    $cached_hash = $this->state->get($cache_key);
    $cache_time = $this->state->get($cache_time_key, 0);

    if ($cached_hash && (time() - $cache_time < $cache_ttl)) {
      if ($debug) {
        $this->logger->debug('SRI Service FALLBACK: Using cached Stripe JS fallback hash: @hash', ['@hash' => $cached_hash]);
      }
      return $cached_hash;
    }
    if ($debug) {
      $this->logger->debug('SRI Service FALLBACK: No valid cached fallback hash found, attempting fetch.');
    }

    $this->logger->warning('Attempting emergency fallback: Fetching Stripe JS directly to calculate hash.');
    $stripe_js_url = self::STRIPE_JS_URL;
    $content = NULL;
    $last_exception = NULL;
    $max_attempts = 1 + self::FALLBACK_FETCH_RETRIES;

    for ($attempt = 1; $attempt <= $max_attempts; $attempt++) {
      if ($debug) {
        $this->logger->debug('SRI Service FALLBACK: Stripe fetch attempt @attempt / @max.', ['@attempt' => $attempt, '@max' => $max_attempts]);
      }
      $last_exception = NULL;
      try {
        $response = $this->httpClient->request('GET', $stripe_js_url, [
          RequestOptions::HEADERS => [
        // Use a common bot UA.
            'User-Agent' => 'Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)',
            'Accept' => 'application/javascript, */*; q=0.8',
            'Accept-Language' => 'en-US,en;q=0.5',
            'Cache-Control' => 'no-cache, no-store, must-revalidate',
            'Pragma' => 'no-cache',
            'Expires' => '0',
          ],
          RequestOptions::TIMEOUT => 10,
          RequestOptions::CONNECT_TIMEOUT => 4,
          // Ensure SSL verification.
          RequestOptions::VERIFY => TRUE,
          RequestOptions::HTTP_ERRORS => FALSE,
          RequestOptions::ALLOW_REDIRECTS => ['max' => 3, 'strict' => TRUE, 'referer' => FALSE, 'protocols' => ['https']],
        ]);
        $status_code = $response->getStatusCode();
        if ($debug) {
          $this->logger->debug('SRI Service FALLBACK: Stripe direct fetch attempt @attempt status: @status', ['@attempt' => $attempt, '@status' => $status_code]);
        }

        if ($status_code === 200) {
          $content = $response->getBody()->getContents();
          // Basic check for non-empty content.
          if (empty(trim($content))) {
            $this->logger->warning('Stripe JS fallback fetch attempt @attempt returned 200 but content seems empty.', ['@attempt' => $attempt]);
            // Treat as failure.
            $content = NULL;
          }
          else {
            // Add cache-busting for future script requests.
            if ($debug) {
              $this->logger->debug('Success fetching Stripe JS content. Adding cache-busting parameters for future requests.');
            }

            // Update the URL in the state to include cache-busting for future page loads.
            $cacheBustUrl = $stripe_js_url;
            if (strpos($cacheBustUrl, '?') === FALSE) {
              $cacheBustUrl .= '?v=' . time();
            }
            else {
              $cacheBustUrl = preg_replace('/([?&]v=)[^&]*/', '$1' . time(), $cacheBustUrl);
              if (strpos($cacheBustUrl, 'v=') === FALSE) {
                $cacheBustUrl .= '&v=' . time();
              }
            }
            $this->state->set('nesguard_drupal.stripe_cache_bust_url', $cacheBustUrl);

            // Success! Exit loop.
            break;
          }
        }
        else {
          $this->logger->error('Stripe JS fallback fetch attempt @attempt failed: Status @status', ['@attempt' => $attempt, '@status' => $status_code]);
          // Don't retry common client errors (except timeout/rate limit which are exceptions).
          if ($status_code >= 400 && $status_code < 500 && $status_code !== 408 && $status_code !== 429) {
            // Stop retrying on 403, 404 etc.
            break;
          }
        }
      }
      // Catch specific exceptions for better logging.
      catch (ConnectException $e) {
        $last_exception = $e;
        $this->logger->warning('Stripe JS fallback network connection error attempt @attempt: @error', ['@attempt' => $attempt, '@error' => $e->getMessage()]);
      }
      catch (RequestException $e) {
        $last_exception = $e;
        $this->logger->warning('Stripe JS fallback network request error attempt @attempt: @error', ['@attempt' => $attempt, '@error' => $e->getMessage()]);
      }
      catch (\Throwable $e) {
        // Stop on general errors.
        $last_exception = $e;
        $this->logger->error('Stripe JS fallback general exception attempt @attempt: @error', ['@attempt' => $attempt, '@error' => $e->getMessage()]);
        break;
      }

      // Wait before retrying if fetch failed and more attempts remain.
      if ($content === NULL && $attempt < $max_attempts) {
        usleep(self::FALLBACK_RETRY_DELAY_MS * 1000);
      }
    } // End retry loop.

    // --- Process fetched content ---
    if ($content !== NULL) {
      $calculated_hash = $this->calculateSha384($content);
      if ($debug) {
        $this->logger->debug('SRI Service FALLBACK: Stripe fallback hash calculation result: @hash', ['@hash' => $calculated_hash ?? 'NULL']);
      }
      if ($calculated_hash) {
        $this->logger->info('Stripe JS fallback successful. Calculated hash: @hash', ['@hash' => $calculated_hash]);
        // Cache the calculated hash and timestamp.
        $this->state->set($cache_key, $calculated_hash);
        $this->state->set($cache_time_key, time());
        return $calculated_hash;
      }
      else {
        $this->logger->error('Stripe JS fallback failed: Could not calculate hash from fetched content.');
      }
    }
    else {
      // Log final failure after all attempts.
      $this->logger->error('Stripe JS fallback failed: Could not fetch content after @attempts attempts. Last error: @error', [
      // Attempts completed.
        '@attempts' => $attempt - 1,
        '@error' => $last_exception ? $last_exception->getMessage() : 'Non-200 status or empty content',
      ]);
    }

    // Fallback ultimately failed.
    return NULL;
  }

  /**
   * Gets the list of monitored script URLs.
   *
   * Returns URLs from API/cache if available and subscription is active.
   * Falls back to returning just the Stripe URL if API fails/is empty but
   * subscription is active. Returns empty array otherwise.
   *
   * @return array<int, string>
   *   List of normalized monitored script URLs.
   */
  public function getMonitoredScriptUrls(): array {
    $debug = $this->config->get('debug_mode') ?? FALSE;
    if ($debug) {
      $this->logger->debug('SRI Service: getMonitoredScriptUrls called.');
    }

    if (!$this->config->get('enabled')) {
      if ($debug) {
        $this->logger->debug('SRI Service: getMonitoredScriptUrls returning [] (Module disabled).');
      }
      return [];
    }

    // Only proceed if subscription is strictly active.
    if ($this->subscriptionService->isActive() !== TRUE) {
      if ($debug) {
        $this->logger->debug('SRI Service: getMonitoredScriptUrls returning [] (Subscription not active).');
      }
      return [];
    }

    // Get hash map (uses API/cache).
    // FALSE = Use cache if valid.
    $hash_map = $this->getMonitoredScriptHashes(FALSE);

    // Case 1: Successfully got a non-empty hash map from API/cache.
    if ($hash_map !== NULL && !empty($hash_map)) {
      $urls = array_keys($hash_map);
      if ($debug) {
        $this->logger->debug('SRI Service: getMonitoredScriptUrls returning @count URLs from API/cache.', ['@count' => count($urls)]);
      }
      return $urls;
    }

    // Case 2: API/Cache failed ($hash_map === NULL) OR returned empty ($hash_map === []).
    // Fall back to monitoring only Stripe JS since subscription is active.
    $stripe_url_normalized = function_exists('nesguard_drupal_normalize_url') ? nesguard_drupal_normalize_url(self::STRIPE_JS_URL) : self::STRIPE_JS_URL;

    if ($stripe_url_normalized) {
      $log_reason = ($hash_map === NULL) ? 'API data unavailable' : 'API returned 0 scripts';
      $this->logger->warning('@reason, falling back to monitoring Stripe JS only.', ['@reason' => $log_reason]);
      if ($debug) {
        $this->logger->debug('SRI Service: getMonitoredScriptUrls returning Stripe fallback URL: @url', ['@url' => $stripe_url_normalized]);
      }
      // Return array with just Stripe URL.
      return [$stripe_url_normalized];
    }
    else {
      $this->logger->error('Failed to normalize the fallback Stripe JS URL. Cannot determine monitored scripts.');
      if ($debug) {
        $this->logger->debug('SRI Service: getMonitoredScriptUrls returning [] (Stripe fallback normalization failed).');
      }
      // Return empty if even Stripe URL fails normalization.
      return [];
    }
  }

  /**
   * Gets the integrity hash data (current and previous) for a script URL.
   *
   * Uses cached API data first. If the requested URL is Stripe's and it's not
   * found in the API data (or API failed), attempts Stripe JS fallback hash
   * calculation for the *current* hash.
   *
   * @param string $script_url
   *   The URL to look up (will be normalized).
   * @param bool $bypass_cache
   *   If TRUE, forces a refresh of the underlying API script/hash map cache
   *   before performing the lookup. Does not affect the Stripe fallback cache.
   *
   * @return array|null
   *   An array ['current' => HASH, 'previous' => HASH|null] or NULL if no
   *   hash can be determined. 'previous' hash will be NULL for Stripe fallback.
   */
  public function getIntegrityHash(string $script_url, bool $bypass_cache = FALSE): ?array {
    $debug = $this->config->get('debug_mode') ?? FALSE;
    if ($debug) {
      $this->logger->debug('SRI Service: getIntegrityHash called for @url (bypass_cache: @bypass).', ['@url' => $script_url, '@bypass' => $bypass_cache ? 'Yes' : 'No']);
    }

    if (!$this->config->get('enabled')) {
      // Module disabled.
      return NULL;
    }

    $normalized_url = function_exists('nesguard_drupal_normalize_url') ? nesguard_drupal_normalize_url($script_url) : $script_url;
    if (!$normalized_url) {
      $this->logger->warning('SRI Service: getIntegrityHash received invalid URL for normalization: @url', ['@url' => $script_url]);
      return NULL;
    }

    // Get hash map (uses API/cache, handles $bypass_cache internally).
    $hash_map = $this->getMonitoredScriptHashes($bypass_cache);

    // Step 1: Check the main hash map (API/Cache data).
    if ($hash_map !== NULL && isset($hash_map[$normalized_url])) {
      if ($debug) {
        $this->logger->debug('SRI Service: getIntegrityHash found data in map for @url.', ['@url' => $normalized_url]);
      }
      // Return consistent structure.
      return [
        'current' => $hash_map[$normalized_url]['current'],
        'previous' => $hash_map[$normalized_url]['previous'] ?? NULL,
      ];
    }

    // Step 2: If not found OR map failed, check if it's the Stripe URL.
    $stripe_url_normalized = function_exists('nesguard_drupal_normalize_url') ? nesguard_drupal_normalize_url(self::STRIPE_JS_URL) : self::STRIPE_JS_URL;

    if ($normalized_url === $stripe_url_normalized && $stripe_url_normalized !== NULL) {
      // Only attempt fallback if subscription is active (otherwise SRI is off).
      if ($this->subscriptionService->isActive() === TRUE) {
        $log_reason = ($hash_map === NULL) ? 'API map unavailable' : 'Stripe hash missing from API map';
        $this->logger->warning('@reason, attempting Stripe hash fallback calculation for @url.', ['@reason' => $log_reason, '@url' => $normalized_url]);

        // Fallback calculates only the current hash. Uses its own short cache.
        $fallback_hash = $this->fetchAndCalculateStripeHashFallback();

        if ($fallback_hash !== NULL) {
          if ($debug) {
            $this->logger->debug('SRI Service: getIntegrityHash Stripe fallback returned: @hash', ['@hash' => $fallback_hash]);
          }
          // Return in the expected structure, previous is NULL for fallbacks.
          return ['current' => $fallback_hash, 'previous' => NULL];
        }
        else {
          // Fallback calculation itself failed.
          if ($debug) {
            $this->logger->debug('SRI Service: getIntegrityHash Stripe fallback failed for @url.', ['@url' => $normalized_url]);
          }
          // Give up if fallback fails.
          return NULL;
        }
      }
      elseif ($debug) {
        $this->logger->debug('SRI Service: Skipping Stripe fallback for @url because subscription is inactive.', ['@url' => $normalized_url]);
      }
    }

    // Step 3: URL is not Stripe, and was not found in the hash map (or map failed).
    if ($debug) {
      $log_suffix = ($hash_map === NULL) ? 'API map was unavailable' : 'not found in monitored list';
      $this->logger->debug('SRI Service: Integrity hash requested for non-Stripe URL @url, but @reason.', ['@url' => $normalized_url, '@reason' => $log_suffix]);
    }

    // Hash cannot be determined.
    return NULL;
  }

  /**
   * Clears the cached SRI script data and the Stripe fallback hash/timestamp.
   */
  public function clearCache(): void {
    $this->state->delete(self::SCRIPT_DATA_CACHE_KEY);
    $this->state->delete(self::STRIPE_FALLBACK_HASH_KEY);
    $this->state->delete(self::STRIPE_FALLBACK_TIME_KEY);
    $this->logger->info('Cleared cached SRI script data and Stripe fallback hash.');
  }

  /**
   * Updates a script hash in the cache based on real-time information.
   *
   * This method is called when a hash change is detected by the client-side
   * WebSocket, allowing immediate hash updates without waiting for cron.
   *
   * @param string $url
   *   The URL of the script to update.
   * @param string $newHash
   *   The new hash value.
   * @param string|null $oldHash
   *   The old hash value to replace (optional).
   *
   * @return bool
   *   TRUE if the update was successful, FALSE otherwise.
   */
  public function updateScriptHash(string $url, string $newHash, ?string $oldHash = NULL): bool {
    $debug = $this->config->get('debug_mode') ?? FALSE;
    if ($debug) {
      $this->logger->debug('SRI Service: updateScriptHash called for URL: @url', ['@url' => $url]);
    }

    // Normalize the URL and hashes.
    $normalizedUrl = nesguard_drupal_normalize_url($url);
    $normalizedNewHash = nesguard_drupal_normalize_hash($newHash);
    $normalizedOldHash = $oldHash ? nesguard_drupal_normalize_hash($oldHash) : NULL;

    if (!$normalizedUrl || !$normalizedNewHash) {
      $this->logger->warning('SRI Service: Failed to normalize URL or hash for updateScriptHash');
      return FALSE;
    }

    // Get the current hash map from cache.
    $scriptData = $this->state->get(self::SCRIPT_DATA_CACHE_KEY);
    if (!$scriptData || !isset($scriptData['hashes']) || !is_array($scriptData['hashes'])) {
      if ($debug) {
        $this->logger->debug('SRI Service: No existing hash map found, creating new one');
      }
      $scriptData = [
        'hashes' => [],
        'timestamp' => time(),
      ];
    }

    // Add or update the entry for this URL.
    $scriptData['hashes'][$normalizedUrl] = [
      'current' => $normalizedNewHash,
      'previous' => $normalizedOldHash,
      'last_update' => time(),
      'source' => 'real-time',
    ];

    // Update the timestamp.
    $scriptData['timestamp'] = time();

    // Save the updated map back to state.
    $this->state->set(self::SCRIPT_DATA_CACHE_KEY, $scriptData);

    // For Stripe URLs, also update the URL with cache-busting parameters.
    if (strpos($normalizedUrl, 'stripe.com') !== FALSE) {
      $cacheBustUrl = $normalizedUrl;
      if (strpos($cacheBustUrl, '?') === FALSE) {
        $cacheBustUrl .= '?v=' . time();
      }
      else {
        $cacheBustUrl = preg_replace('/([?&]v=)[^&]*/', '$1' . time(), $cacheBustUrl);
        if (strpos($cacheBustUrl, 'v=') === FALSE) {
          $cacheBustUrl .= '&v=' . time();
        }
      }

      // Also store the raw hash directly in a separate state key for immediate client-side recovery.
      $this->state->set('nesguard_drupal.stripe_latest_hash', $normalizedNewHash);
      $this->state->set('nesguard_drupal.stripe_hash_update_time', time());

      // Store the cache-busted URL for future page loads.
      $this->state->set('nesguard_drupal.stripe_cache_bust_url', $cacheBustUrl);

      // Log the update for tracking purposes.
      $this->logger->info('Updated Stripe hash in state: @hash from source: @source', [
        '@hash' => $normalizedNewHash,
        '@source' => $oldHash ? 'update' : 'new',
      ]);

      if ($debug) {
        $this->logger->debug('SRI Service: Updated cache-busted URL for Stripe: @url', [
          '@url' => $cacheBustUrl,
        ]);
      }
    }

    if ($debug) {
      $this->logger->debug('SRI Service: Updated hash for @url to @hash', [
        '@url' => $normalizedUrl,
        '@hash' => $normalizedNewHash,
      ]);
    }

    return TRUE;
  }

}