<?php

namespace Drupal\nesguard_drupal\Service;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\State\StateInterface;
use Psr\Log\LoggerInterface;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Exception\RequestException;

/**
 * Checks user subscription status via NESGuard API. Includes retry on 401.
 */
class NesguardSubscriptionService {

  // Constants for API, caching, rate limiting.
  public const API_BASE_URL = 'https://nesgaurd.cloud/api/v1/';
  protected const SUBSCRIPTION_CACHE_KEY = 'nesguard_drupal.subscription_status';
  // Cache status for 1 hour.
  protected const CACHE_LIFETIME_SECONDS = 3600;
  protected const RATE_LIMIT_CHECK_KEY = 'nesguard_drupal.sub.check.limit_until';
  // Seconds.
  protected const RATE_LIMIT_DEFAULT_DURATION = 60;

  /**
   * The immutable configuration object.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected $config;

  /**
   * The state service.
   *
   * @var \Drupal\Core\State\StateInterface
   */
  protected $state;

  /**
   * The logger service.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * The HTTP client service.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected $httpClient;

  /**
   * The authentication service.
   *
   * @var \Drupal\nesguard_drupal\Service\NesguardAuthService
   */
  protected $authService;

  /**
   * Constructs a NesguardSubscriptionService object.
   */
  public function __construct(
    ConfigFactoryInterface $config_factory,
    StateInterface $state,
    LoggerInterface $logger,
    ClientInterface $http_client,
    NesguardAuthService $auth_service,
  ) {
    $this->config = $config_factory->get('nesguard_drupal.settings');
    $this->state = $state;
    $this->logger = $logger;
    $this->httpClient = $http_client;
    $this->authService = $auth_service;
  }

  // --- Rate Limiting Helpers (Corrected Implementation) ---

  /**
   * Checks if a specific action is currently rate limited.
   */
  protected function isRateLimited(string $limit_key): bool {
    $limited_until = $this->state->get($limit_key, 0);
    $is_limited = time() < $limited_until;
    if ($is_limited && $this->config->get('debug_mode')) {
      $this->logger->debug('Action (@key) is rate limited until @time.', [
        '@key' => $limit_key,
        '@time' => date('Y-m-d H:i:s T', $limited_until),
      ]);
    }
    // *** Ensure boolean is always returned ***
    return $is_limited;
  }

  /**
   * Sets a rate limit for a specific action.
   */
  protected function setRateLimit(string $limit_key, int $duration = self::RATE_LIMIT_DEFAULT_DURATION): void {
    $limited_until = time() + $duration;
    $this->state->set($limit_key, $limited_until);
    $this->logger->warning('Rate limit set for action (@key) until @time (@duration s).', [
      '@key' => $limit_key,
      '@time' => date('Y-m-d H:i:s T', $limited_until),
      '@duration' => $duration,
    ]);
  }

  /**
   * Clears a rate limit for a specific action.
   */
  protected function clearRateLimit(string $limit_key): void {
    // Check if key exists before deleting for efficiency/logging.
    if ($this->state->get($limit_key)) {
      $this->state->delete($limit_key);
      if ($this->config->get('debug_mode')) {
        $this->logger->debug('Cleared rate limit for action (@key).', ['@key' => $limit_key]);
      }
    }
  }

  // --- Core Service Method ---

  /**
   * Checks if the user associated with the API credentials has an active subscription.
   *
   * Includes caching, rate limiting, and retry on 401 Unauthorized.
   */
  public function isActive(): ?bool {
    $debug = $this->config->get('debug_mode');
    if ($debug) {
      $this->logger->debug('NESGUARD DEBUG (Sub): isActive called.');
    }

    if (!$this->config->get('enabled') || !$this->authService->hasCredentials()) {
      if ($debug) {
        $this->logger->debug('NESGUARD DEBUG (Sub): isActive returning NULL (Module disabled or no creds).');
      }
      return NULL;
    }

    // Check cache first.
    $cached_status = $this->state->get(self::SUBSCRIPTION_CACHE_KEY);
    if ($cached_status !== NULL && isset($cached_status['timestamp']) && (time() - $cached_status['timestamp'] < self::CACHE_LIFETIME_SECONDS)) {
      if ($debug) {
        $this->logger->debug('NESGUARD DEBUG (Sub): isActive returning from cache: @status', ['@status' => $cached_status['status'] ? 'Active' : 'Inactive']);
      }
      return $cached_status['status'];
    }

    // Check rate limit before API call.
    if ($this->isRateLimited(self::RATE_LIMIT_CHECK_KEY)) {
      $this->logger->warning('Subscription status check skipped due to rate limit.');
      // Return stale cache if rate limited.
      return $cached_status['status'] ?? NULL;
    }

    // Perform API Call with Retry on 401.
    // Initial attempt + 1 retry after token refresh.
    $max_attempts = 2;
    $attempt = 0;
    $access_token = NULL;
    $is_active = NULL;
    $agency_id = $this->config->get('agency_id');

    while ($attempt < $max_attempts && $is_active === NULL) {
      $attempt++;
      if ($debug) {
        $this->logger->debug('NESGUARD DEBUG (Sub): isActive API call attempt #@attempt.', ['@attempt' => $attempt]);
      }

      $force_refresh = ($attempt > 1);
      $access_token = $this->authService->getAccessToken($force_refresh);

      if (!$access_token) {
        $this->logger->warning('Cannot check subscription status (Attempt @attempt): Authentication failed.', ['@attempt' => $attempt]);
        return $cached_status['status'] ?? NULL;
      }
      if (empty($agency_id)) {
        $this->logger->error('Cannot check subscription status (Attempt @attempt): Agency ID missing.', ['@attempt' => $attempt]);
        return NULL;
      }

      try {
        $endpoint_url = self::API_BASE_URL . 'subscriptions/user/';
        if ($debug) {
          $this->logger->debug('NESGUARD DEBUG (Sub): Calling @url (Attempt @attempt)', ['@url' => $endpoint_url, '@attempt' => $attempt]);
        }

        $response = $this->httpClient->request('GET', $endpoint_url, [
          'headers' => [
            'Authorization' => 'Bearer ' . $access_token,
            'X-Agency-ID' => $agency_id,
            'Accept' => 'application/json',
            'User-Agent' => 'NESGuard Drupal/' . \Drupal::VERSION,
          ],
          'timeout' => 10,
          'connect_timeout' => 5,
          'http_errors' => FALSE,
        ]);

        $status_code = $response->getStatusCode();
        $body = $response->getBody()->getContents();
        $data = json_decode($body, TRUE);

        // Validate JSON parsing succeeded.
        if ($data === NULL && json_last_error() !== JSON_ERROR_NONE) {
          $json_error = json_last_error_msg();
          $this->logger->error('Failed to parse subscription API response JSON: @error. Response body: @body', [
            '@error' => $json_error,
            '@body' => substr($body, 0, 255) . (strlen($body) > 255 ? '...[truncated]' : ''),
          ]);
          $attempt++;
          // Try again on JSON parse failure.
          continue;
        }

        if ($debug) {
          $this->logger->debug('NESGUARD DEBUG (Sub): Attempt @attempt API Status: @status', ['@attempt' => $attempt, '@status' => $status_code]);
        }

        if ($status_code == 200 && ($data['success'] ?? FALSE) && isset($data['data']['status'])) {
          $is_active = ($data['data']['status'] === 'active');
          $this->logger->info('Subscription status for @agency checked: @status', ['@agency' => $agency_id, '@status' => $is_active ? 'Active' : 'Inactive/Other']);
          $this->clearRateLimit(self::RATE_LIMIT_CHECK_KEY);
          // Success.
          break;
        }
        elseif ($status_code == 404) {
          $is_active = FALSE;
          $this->logger->info('No active subscription found for user/agency @agency (404).', ['@agency' => $agency_id]);
          $this->clearRateLimit(self::RATE_LIMIT_CHECK_KEY);
          // Success (found no subscription).
          break;
        }
        elseif ($status_code == 401 && $attempt < $max_attempts) {
          $this->logger->warning('Subscription check received 401 (Attempt @attempt). Forcing token refresh and retrying.', ['@attempt' => $attempt]);
          // Continue loop.
        }
        elseif ($status_code == 429) {
          $this->logger->warning('Subscription status check failed for @agency: Rate Limited (429).', ['@agency' => $agency_id]);
          $this->setRateLimit(self::RATE_LIMIT_CHECK_KEY);
          $is_active = $cached_status['status'] ?? NULL;
          // Do not retry.
          break;
        }
        else {
          // Other error, or 401 on final attempt.
          $error_message = $data['error']['message'] ?? $body;
          $this->logger->error('Failed to check subscription status for @agency (Attempt @attempt): Status @status - @message', [
            '@agency' => $agency_id,
            '@attempt' => $attempt,
            '@status' => $status_code,
            '@message' => $error_message,
          ]);
          $is_active = $cached_status['status'] ?? NULL;
          // Do not retry.
          break;
        }
      }
      // End Try block.
      catch (RequestException $e) {
        $this->logger->error('Network error checking subscription status for @agency (Attempt @attempt): @error', ['@agency' => $agency_id, '@attempt' => $attempt, '@error' => $e->getMessage()]);
        $is_active = $cached_status['status'] ?? NULL;
        // Do not retry.
        break;
      }
      catch (\Throwable $e) {
        $this->logger->error('General exception checking subscription status for @agency (Attempt @attempt): @error', ['@agency' => $agency_id, '@attempt' => $attempt, '@error' => $e->getMessage()]);
        $is_active = $cached_status['status'] ?? NULL;
        // Do not retry.
        break;
      }

    } // End While loop.

    // Cache the result only if we determined a definitive status.
    if ($is_active !== NULL) {
      $this->state->set(self::SUBSCRIPTION_CACHE_KEY, ['status' => $is_active, 'timestamp' => time()]);
    }

    return $is_active;
  }

  /**
   * Clears the cached subscription status.
   */
  public function clearCache(): void {
    $this->state->delete(self::SUBSCRIPTION_CACHE_KEY);
    $this->logger->info('Cleared cached subscription status.');
  }

}
