<?php

namespace Drupal\Tests\nesguard_drupal\Functional;

use Drupal\Tests\BrowserTestBase;
use Drupal\Core\Site\Settings; // To set settings temporarily if needed

/**
 * Tests the NESGuard settings form.
 *
 * @group nesguard_drupal
 */
class NesguardSettingsFormTest extends BrowserTestBase {

  /**
   * {@inheritdoc}
   */
  protected static $modules = ['nesguard_drupal']; // Enable module

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stark'; // Use minimal theme for testing

  /**
   * A user with permission to administer site configuration.
   *
   * @var \Drupal\user\UserInterface
   */
  protected $adminUser;

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    // Ensure the encryption key is set for testing saving API key.
    // Note: This modifies settings for the test run only.
    $settings = Settings::getAll();
    $settings['nesguard_encryption_key'] = 'a-secure-key-for-testing-only-123';
    new Settings($settings);

    // Create and log in an administrative user.
    $this->adminUser = $this->drupalCreateUser(['administer site configuration']);
    $this->drupalLogin($this->adminUser);
  }

  /**
   * Tests that the settings form loads and basic elements exist.
   */
  public function testFormLoad(): void {
    $this->drupalGet('/admin/config/system/nesguard-sri');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContains('NESGuard SRI Settings');
    $this->assertSession()->fieldExists('enabled');
    $this->assertSession()->fieldExists('agency_id');
    $this->assertSession()->fieldExists('api_key_new');
    $this->assertSession()->fieldExists('csp_enabled');
  }

  /**
   * Tests submitting the form and saving configuration.
   */
  public function testFormSubmit(): void {
    // Define form values to submit.
    $edit = [
      'enabled' => TRUE,
      'api_email_display' => 'test@example.com',
      'agency_id' => 'TEST-NES-123',
      'api_key_new' => 'test-api-key-value-12345', // Will be encrypted
      'csp_enabled' => TRUE,
      'csp_report_only' => TRUE,
      'debug_mode' => TRUE,
      // Add other values as needed...
    ];

    $this->drupalGet('/admin/config/system/nesguard-sri');
    $this->submitForm($edit, 'Save configuration');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContains('NESGuard settings saved.');

    // Verify saved configuration.
    $config = $this->config('nesguard_drupal.settings');
    $this->assertTrue($config->get('enabled'));
    $this->assertEquals('test@example.com', $config->get('api_email'));
    $this->assertEquals('TEST-NES-123', $config->get('agency_id'));
    $this->assertTrue($config->get('csp_enabled'));
    $this->assertTrue($config->get('csp_report_only'));
    $this->assertTrue($config->get('debug_mode'));

    // Verify API key was saved (and is not plaintext).
    $saved_api_key = $config->get('api_key');
    $this->assertNotEmpty($saved_api_key);
    $this->assertNotEquals('test-api-key-value-12345', $saved_api_key);
    // We could try decrypting it here if needed, using the known test key.
  }

  // Add tests for validation (e.g., missing required fields when enabled).
  // Add tests for disabling the module via the form.

}