<?php

namespace Drupal\Tests\nesguard_drupal\FunctionalJavascript;

use Drupal\FunctionalJavascriptTests\WebDriverTestBase;
use Drupal\Core\Url;

/**
 * Tests the client-side JavaScript behavior of NESGuard SRI.
 *
 * @group nesguard_drupal
 */
class NesguardClientTest extends WebDriverTestBase {

  /**
   * {@inheritdoc}
   */
  protected static $modules = ['nesguard_drupal', 'node']; // Add necessary modules

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stark';

  /**
   * A user with permission to access content.
   *
   * @var \Drupal\user\UserInterface
   */
  protected $testUser;

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    // Configure the module - enable it, set API keys etc.
    // This likely requires setting $settings['nesguard_encryption_key'] first.
    // It's often easier to use config import or API calls in setUp().
    $this->config('nesguard_drupal.settings')
      ->set('enabled', TRUE)
      // Mock necessary config - API key needs encryption step normally.
      // ->set('api_key', 'some_encrypted_key_for_test')
      ->set('agency_id', 'TEST-AGENCY-JS')
      // Mock subscription as active (bypass API call if possible for test).
      // This might require mocking the subscription service in the container.
      // For now, assume it might work or needs mocking.
      ->set('debug_mode', TRUE) // Enable debug for console logs
      ->save();

    // Create content or set up page where JS should load.
    $this->drupalCreateNode(['title' => 'Test Page', 'type' => 'page']);
    $this->testUser = $this->drupalCreateUser(['access content']);
    $this->drupalLogin($this->testUser);
  }

  /**
   * Tests that the library and settings are attached.
   */
  public function testJavaScriptAttachment(): void {
    $this->drupalGet('/node/1');
    $session = $this->getSession();
    $page = $session->getPage();
    $web_assert = $this->assertSession();

    // 1. Check if drupalSettings.nesguard exists.
    $settings = $this->getDrupalSettings();
    $this->assertArrayHasKey('nesguard', $settings, 'nesguard key found in drupalSettings.');
    $this->assertTrue($settings['nesguard']['sriEnabled'], 'SRI is marked as enabled in settings.');
    $this->assertEquals('TEST-AGENCY-JS', $settings['nesguard']['agencyId'], 'Agency ID found in settings.');
    // We don't want to assert the access token value here.
    $this->assertArrayHasKey('accessToken', $settings['nesguard'], 'Access token key exists in settings.');

    // 2. Check if the library's JS file (or aggregate) is loaded.
    // This checks the raw HTML source. Look near </body>.
    // Note: The exact filename will change with aggregation. Check for presence.
    $web_assert->responseContains('nesguard-client.js'); // Check if filename present somewhere.
    // More robust: Check if the NESGUARD_CLIENT global is created.
    $result = $session->evaluateScript('typeof window.NESGUARD_CLIENT');
    $this->assertEquals('object', $result, 'NESGUARD_CLIENT global object exists.');

    // 3. Check for console logs (requires debug mode enabled).
     $logs = $session->evaluateScript("window.console.history && window.console.history.logs ? JSON.stringify(window.console.history.logs) : '[]'");
     // Note: Capturing console logs is tricky and browser/driver dependent.
     // This example assumes some mechanism like console.history exists.
     $this->assertStringContainsString('[NESGuard] Initializing client', $logs, 'Initialization log found.');
     // Check for connection attempt log etc.
  }

  /**
   * Tests if SRI attributes are potentially applied (basic check).
   *
   * NOTE: This cannot easily test the *correctness* of the hash without
   * knowing the expected hash or mocking the external script.
   * It also runs AFTER the JS may have modified the DOM.
   */
   public function testSriAttributePresence(): void {
     // Need to load a page that includes a script *known* to be monitored
     // AND loaded via hook_library_info_alter or targeted by JS.
     // Example assumes Stripe is loaded via library on this path.
     // Adjust path and selector as needed.
     $this->drupalGet('/some-page-with-stripe'); // Replace with actual path
     $session = $this->getSession();
     $web_assert = $this->assertSession();

     // Wait briefly for JS execution if needed.
     $session->wait(1000, 'typeof window.NESGUARD_CLIENT !== "undefined" && window.NESGUARD_CLIENT.initialized');

     // Check if the Stripe script tag has the attributes.
     // This selector might need adjustment.
     $stripe_script = $page->find('css', 'script[src*="js.stripe.com/v3"]');
     $this->assertNotNull($stripe_script, 'Stripe script tag found.');
     if ($stripe_script) {
        $this->assertTrue($stripe_script->hasAttribute('integrity'), 'Stripe script has integrity attribute.');
        $this->assertStringStartsWith('sha384-', $stripe_script->getAttribute('integrity'), 'Integrity attribute starts with sha384-.');
        $this->assertEquals('anonymous', $stripe_script->getAttribute('crossorigin'), 'Stripe script has crossorigin attribute.');
     }
   }

  // Add more tests:
  // - Test behavior when subscription is inactive (sriEnabled should be false, no SRI attributes).
  // - Test CSP headers if CSP is enabled in config.
  // - More advanced: Mock WebSocket communication (very complex in WebDriver tests).

}