<?php

namespace Drupal\Tests\nesguard_drupal\Unit\Service;

use Drupal\Core\Cache\MemoryCache\MemoryBackend; // Correct MemoryBackend namespace
use Drupal\Core\Config\Config;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\DependencyInjection\ContainerBuilder;
use Drupal\Core\Lock\LockBackendInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\State\State;
use Drupal\Core\Site\Settings;
use Drupal\Tests\UnitTestCase;
use Drupal\nesguard_drupal\Service\NesguardAuthService;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Psr7\Response;
use GuzzleHttp\Exception\RequestException;
use GuzzleHttp\Psr7\Request;
use Prophecy\Argument;
use Prophecy\PhpUnit\ProphecyTrait;

/**
 * @coversDefaultClass \Drupal\nesguard_drupal\Service\NesguardAuthService
 * @group nesguard_drupal
 */
class NesguardAuthServiceTest extends UnitTestCase {

  use ProphecyTrait;

  protected $authService;
  protected $configFactory;
  protected $config;
  protected $state;
  protected $logger;
  protected $httpClient;
  protected $settings;
  protected $lock;
  protected $apiBaseUrl = NesguardAuthService::API_BASE_URL;
  protected $encryptionKey = 'test-encryption-key-1234567890';

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->config = $this->prophesize(Config::class);
    $this->config->get('enabled')->willReturn(TRUE);
    $this->config->get('agency_id')->willReturn('TEST-AGENCY');
    $this->config->get('api_key')->willReturn($this->encrypt('test-api-key'));
    $this->config->get('debug_mode')->willReturn(FALSE);

    $this->configFactory = $this->prophesize(ConfigFactoryInterface::class);
    $this->configFactory->get('nesguard_drupal.settings')->willReturn($this->config->reveal());

    // Use actual State object with MemoryBackend for testing state interactions.
    $this->state = new State(new MemoryBackend('state'));

    $this->logger = $this->prophesize(LoggerChannelInterface::class);

    $this->httpClient = $this->prophesize(ClientInterface::class);

    $this->settings = $this->prophesize(Settings::class);
    $this->settings->get('nesguard_encryption_key')->willReturn($this->encryptionKey);

    $this->lock = $this->prophesize(LockBackendInterface::class);
    // Default happy path for lock.
    $this->lock->acquire(Argument::any(), Argument::any())->willReturn(TRUE);
    $this->lock->wait(Argument::any(), Argument::any())->willReturn(TRUE);
    // Use willBeCalled() for methods expected ONLY on successful acquire.
    $this->lock->release(Argument::any())->will(function() {}); // Allow but don't strictly require release in all paths initially.

    $this->authService = new NesguardAuthService(
      $this->configFactory->reveal(),
      $this->state,
      $this->logger->reveal(),
      $this->httpClient->reveal(),
      $this->settings->reveal(),
      $this->lock->reveal()
    );

    // Mock container for logger.
    $container = new ContainerBuilder();
    $logger_factory = $this->prophesize(LoggerChannelFactoryInterface::class);
    $logger_factory->get('nesguard_drupal')->willReturn($this->logger->reveal());
    $container->set('logger.factory', $logger_factory->reveal());
    \Drupal::setContainer($container);
  }

  /**
   * Helper to simulate encryption (matches service logic).
   */
  protected function encrypt(string $value): string {
    $cipher = 'aes-256-cbc';
    $key_material = substr(hash('sha256', $this->encryptionKey, TRUE), 0, 32);
    $iv_length = openssl_cipher_iv_length($cipher);
    $iv = openssl_random_pseudo_bytes($iv_length);
    $ciphertext = openssl_encrypt($value, $cipher, $key_material, OPENSSL_RAW_DATA, $iv);
    return base64_encode($iv . $ciphertext);
  }

  /**
   * Helper to simulate decryption failure by providing bad key.
   */
  protected function simulateDecryptionFailure() {
    // Mock settings to return a different key during decryption attempt.
    $this->settings->get('nesguard_encryption_key')->willReturn('wrong-key-for-decryption');
    // Re-create service with modified settings mock if necessary, or ensure
    // the service re-fetches config if config changes mid-request (unlikely).
    // For simplicity, assume the service uses the key available when called.
  }

  /**
   * @covers ::hasCredentials
   * @dataProvider providerHasCredentials
   */
  public function testHasCredentials(string $agency_id, string $api_key, bool $expected): void {
    $this->config->get('agency_id')->willReturn($agency_id);
    $this->config->get('api_key')->willReturn($api_key);
    $this->assertEquals($expected, $this->authService->hasCredentials());
  }

  public function providerHasCredentials(): array { /* ... Same as before ... */ }

  /** @covers ::getAccessToken */
  public function testGetAccessTokenModuleDisabled(): void { /* ... Same as before ... */ }

  /** @covers ::getAccessToken */
  public function testGetAccessTokenNoCredentials(): void { /* ... Same as before ... */ }

  /** @covers ::getAccessToken */
  public function testGetAccessTokenValidCache(): void { /* ... Same as before ... */ }

  /** @covers ::getAccessToken */
  public function testGetAccessTokenExpiredCacheSuccessfulRefresh(): void { /* ... Same as before ... */ }

  /** @covers ::getAccessToken */
  public function testGetAccessTokenNoCacheSuccessfulLogin(): void { /* ... Same as before ... */ }

  /** @covers ::getAccessToken */
  public function testGetAccessTokenRefreshFailsThenLoginSucceeds(): void {
    $expired_token_data = [
      'access' => 'expired-access-token',
      'refresh' => 'valid-refresh-token',
      'updated_at' => time() - (NesguardAuthService::ASSUMED_TOKEN_VALIDITY_SECONDS + 3600),
    ];
    $this->state->set(NesguardAuthService::TOKEN_STATE_KEY, $expired_token_data);
    $this->state->set(NesguardAuthService::REFRESH_COOLDOWN_KEY, 0);

    // Mock refresh failure (e.g., 401 Unauthorized).
    $mock_refresh_response = new Response(401, [], json_encode(['success' => false, 'error' => ['code' => 'UNAUTHORIZED', 'message' => 'Invalid refresh token']]));
    $this->httpClient->post($this->apiBaseUrl . 'auth/token/refresh/', Argument::any())
      ->shouldBeCalledOnce()
      ->willReturn($mock_refresh_response);

    // Mock successful login.
    $login_response_body = json_encode(['access' => 'new-token-via-login', 'refresh' => 'new-refresh-via-login']);
    $mock_login_response = new Response(200, [], $login_response_body);
    $this->httpClient->post($this->apiBaseUrl . 'auth/login/', Argument::any())
      ->shouldBeCalledOnce() // Called after refresh fails.
      ->willReturn($mock_login_response);

    // Assertions.
    $token = $this->authService->getAccessToken();
    $this->assertEquals('new-token-via-login', $token);
    $this->logger->warning(Argument::containingString('API token refresh failed'), Argument::any())->shouldHaveBeenCalled();
    $this->logger->error(Argument::containingString('Clearing tokens due to refresh failure'), Argument::any())->shouldHaveBeenCalled();
    $new_state = $this->state->get(NesguardAuthService::TOKEN_STATE_KEY);
    $this->assertEquals('new-token-via-login', $new_state['access']);
  }

  /** @covers ::getAccessToken */
  public function testGetAccessTokenLoginRateLimited(): void { /* ... Same as before ... */ }

  /** @covers ::getAccessToken */
  public function testGetAccessTokenRefreshRateLimited(): void {
    $expired_token_data = [ /* ... Expired token data with refresh token ... */ ];
    $this->state->set(NesguardAuthService::TOKEN_STATE_KEY, $expired_token_data);
    $this->state->set(NesguardAuthService::REFRESH_COOLDOWN_KEY, 0);
    // Set rate limit for refresh.
    $this->state->set(NesguardAuthService::RATE_LIMIT_REFRESH_KEY, time() + 300);

    // Login should be attempted as refresh is skipped.
    $mock_login_response = new Response(500, [], 'Server Error'); // Simulate login also failing.
     $this->httpClient->post($this->apiBaseUrl . 'auth/login/', Argument::any())
       ->shouldBeCalledOnce()
       ->willReturn($mock_login_response);

    $token = $this->authService->getAccessToken();
    $this->assertNull($token);
    $this->logger->warning('Token refresh skipped due to active rate limit.')->shouldHaveBeenCalled();
    $this->httpClient->post($this->apiBaseUrl . 'auth/token/refresh/', Argument::any())->shouldNotHaveBeenCalled();
  }

  /** @covers ::getAccessToken */
  public function testGetAccessTokenRefreshApi429ThenLoginSucceeds(): void {
    $this->state->set(NesguardAuthService::TOKEN_STATE_KEY, $expired_token_data);
    $this->state->set(NesguardAuthService::REFRESH_COOLDOWN_KEY, 0);
    $this->state->set(NesguardAuthService::REFRESH_COOLDOWN_KEY, 0);

    // Mock refresh 429 response.
    $mock_refresh_response = new Response(429, [], 'Too Many Requests');
    $this->httpClient->post($this->apiBaseUrl . 'auth/token/refresh/', Argument::any())
      ->shouldBeCalledOnce()
      ->willReturn($mock_refresh_response);

    // Mock successful login after refresh fails.
    $login_response_body = json_encode(['access' => 'login-token-after-429', 'refresh' => 'login-refresh-after-429']);
    $mock_login_response = new Response(200, [], $login_response_body);
     $this->httpClient->post($this->apiBaseUrl . 'auth/login/', Argument::any())
       ->shouldBeCalledOnce()
       ->willReturn($mock_login_response);

    $token = $this->authService->getAccessToken();
    $this->assertEquals('login-token-after-429', $token);

    // Check refresh rate limit was set.
    $this->assertGreaterThan(time(), $this->state->get(NesguardAuthService::RATE_LIMIT_REFRESH_KEY, 0));
    $this->logger->warning(Argument::containingString('Rate limit set for action'), Argument::containingString(NesguardAuthService::RATE_LIMIT_REFRESH_KEY))->shouldHaveBeenCalled();
  }

   /** @covers ::getAccessToken */
  public function testGetAccessTokenLoginApi403(): void {
    // No token data in state initially.

    // Mock login failure (e.g., 403 Forbidden).
    $mock_response = new Response(403, [], json_encode(['success' => false, 'error' => ['code' => 'FORBIDDEN', 'message' => 'Invalid credentials']]));
    $this->httpClient->post($this->apiBaseUrl . 'auth/login/', Argument::any())
      ->shouldBeCalledOnce()
      ->willReturn($mock_response);

    $token = $this->authService->getAccessToken();
    $this->assertNull($token);
    $this->logger->error(Argument::containingString('API login failed'), Argument::any())->shouldHaveBeenCalled();
    // Ensure tokens are cleared on login failure.
    $this->assertNull($this->state->get(NesguardAuthService::TOKEN_STATE_KEY));
  }

   /** @covers ::getAccessToken */
   public function testGetAccessTokenLoginNetworkError(): void {
    // No token data in state initially.
    $request = new Request('POST', $this->apiBaseUrl . 'auth/login/');
    $this->httpClient->post($this->apiBaseUrl . 'auth/login/', Argument::any())
      ->shouldBeCalledOnce()
      ->willThrow(new RequestException("Network error", $request));

    $token = $this->authService->getAccessToken();
    $this->assertNull($token);
    $this->logger->error(Argument::containingString('API login network error'), Argument::any())->shouldHaveBeenCalled();
  }

   /** @covers ::getAccessToken */
  public function testGetAccessTokenDecryptionFails(): void {
    // No token data in state initially.
    $this->simulateDecryptionFailure(); // Use helper to make decryption fail.

    // Mock login API call expectation (it might still be called but should fail).
    // We test the decryption failure by asserting NULL return and log message.
    $this->httpClient->post($this->apiBaseUrl . 'auth/login/', Argument::any())
      ->shouldNotHaveBeenCalled(); // If decrypt fails early, don't call API.

    $token = $this->authService->getAccessToken();
    $this->assertNull($token);
    // Check for decryption error log (difficult to check exact message via prophecy).
    // Instead, verify HTTP client wasn't called, implying failure before that.
    $this->logger->error(Argument::containingString('Cannot perform login: Agency ID or decrypted API Key is missing'), Argument::any())->shouldHaveBeenCalled();
  }

   /** @covers ::getAccessToken */
  public function testGetAccessTokenLockFails(): void {
    // No token data in state initially.
    $this->lock->acquire(Argument::any(), Argument::any())->willReturn(FALSE);
    $this->lock->wait(Argument::any(), Argument::any())->willReturn(FALSE);

    $token = $this->authService->getAccessToken();
    $this->assertNull($token); // Expect null if lock fails.
    $this->logger->error('Auth lock timeout during login attempt.')->shouldHaveBeenCalled();
    $this->httpClient->post(Argument::any(), Argument::any())->shouldNotHaveBeenCalled();
  }

  /** @covers ::clearTokens */
  public function testClearTokens(): void {
     $this->state->set(NesguardAuthService::TOKEN_STATE_KEY, ['access' => 'a', 'refresh' => 'b', 'updated_at' => time()]);
     $this->state->set(NesguardAuthService::REFRESH_COOLDOWN_KEY, time());

     $this->authService->clearTokens();

     $this->assertNull($this->state->get(NesguardAuthService::TOKEN_STATE_KEY));
     $this->assertNull($this->state->get(NesguardAuthService::REFRESH_COOLDOWN_KEY));
     $this->logger->info('Cleared stored authentication tokens.')->shouldHaveBeenCalled();
  }

}