<?php

namespace Drupal\Tests\nesguard_drupal\Unit\Service;

use Drupal\Core\Cache\MemoryCache\MemoryBackend;
use Drupal\Core\Config\Config;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\DependencyInjection\ContainerBuilder;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\State\State;
use Drupal\Tests\UnitTestCase;
use Drupal\nesguard_drupal\Service\NesguardAuthService;
use Drupal\nesguard_drupal\Service\NesguardSubscriptionService;
use Drupal\nesguard_drupal\Service\NesguardSriService;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Psr7\Response;
use GuzzleHttp\Exception\RequestException;
use GuzzleHttp\Psr7\Request;
use Prophecy\Argument;
use Prophecy\PhpUnit\ProphecyTrait;

/**
 * @coversDefaultClass \Drupal\nesguard_drupal\Service\NesguardSriService
 * @group nesguard_drupal
 */
class NesguardSriServiceTest extends UnitTestCase {

  use ProphecyTrait;

  protected $sriService;
  protected $configFactory;
  protected $config;
  protected $state;
  protected $logger;
  protected $httpClient;
  protected $authService;
  protected $subscriptionService;
  protected $apiBaseUrl = NesguardSriService::API_BASE_URL;
  protected $stripeUrl;

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->config = $this->prophesize(Config::class);
    $this->config->get('enabled')->willReturn(TRUE);
    $this->config->get('agency_id')->willReturn('TEST-AGENCY');
    $this->config->get('debug_mode')->willReturn(FALSE);
    // Cache lifetime.
    $this->config->get('cache_lifetime')->willReturn(3600);

    $this->configFactory = $this->prophesize(ConfigFactoryInterface::class);
    $this->configFactory->get('nesguard_drupal.settings')->willReturn($this->config->reveal());

    $this->state = new State(new MemoryBackend('state'));

    $this->logger = $this->prophesize(LoggerChannelInterface::class);

    $this->httpClient = $this->prophesize(ClientInterface::class);

    $this->authService = $this->prophesize(NesguardAuthService::class);
    $this->authService->hasCredentials()->willReturn(TRUE);
    $this->authService->getAccessToken()->willReturn('dummy-access-token');
    // Ensure constant is accessible.
    $this->authService->API_BASE_URL = NesguardAuthService::API_BASE_URL;


    $this->subscriptionService = $this->prophesize(NesguardSubscriptionService::class);
    $this->subscriptionService->isActive()->willReturn(TRUE); // Assume active by default.

    $this->sriService = new NesguardSriService(
      $this->configFactory->reveal(),
      $this->state,
      $this->logger->reveal(),
      $this->httpClient->reveal(),
      $this->authService->reveal(),
      $this->subscriptionService->reveal()
    );

    // Mock container for logger.
    $container = new ContainerBuilder();
    $logger_factory = $this->prophesize(LoggerChannelFactoryInterface::class);
    $logger_factory->get('nesguard_drupal')->willReturn($this->logger->reveal());
    $container->set('logger.factory', $logger_factory->reveal());
    \Drupal::setContainer($container);

    // Mock global normalize functions or ensure they exist/are loaded.
    if (!function_exists('nesguard_drupal_normalize_url')) {
        eval('namespace { function nesguard_drupal_normalize_url(?string $url): ?string { return $url ? trim($url) : null; } }');
    }
    if (!function_exists('nesguard_drupal_normalize_hash')) {
       eval('namespace { function nesguard_drupal_normalize_hash(?string $hash): ?string { return $hash ? trim($hash) : null; } }');
    }
    $this->stripeUrl = nesguard_drupal_normalize_url(NesguardSriService::STRIPE_JS_URL);
  }

  /**
   * Creates a mock API response for /sri/client/scripts/.
   */
  protected function createMockScriptsResponse(array $scripts, int $status = 200, bool $success = TRUE): Response {
      $data = ['success' => $success];
      if ($success) {
          $data['data'] = ['scripts' => $scripts];
      } else {
          $data['error'] = ['code' => 'TEST_ERROR', 'message' => 'Test Error'];
      }
      return new Response($status, [], json_encode($data));
  }

  /**
   * @covers ::getMonitoredScriptHashes
   */
  public function testGetMonitoredScriptHashesFromCache(): void {
    $expected_map = [$this->stripeUrl => 'sha384-cached-hash'];
    $this->state->set(NesguardSriService::SCRIPT_DATA_CACHE_KEY, ['hashes' => $expected_map, 'timestamp' => time() - 100]);

    $result = $this->sriService->getMonitoredScriptHashes(FALSE); // Don't bypass cache.

    $this->assertEquals($expected_map, $result);
    $this->httpClient->get(Argument::any(), Argument::any())->shouldNotHaveBeenCalled();
  }

  /**
   * @covers ::getMonitoredScriptHashes
   */
  public function testGetMonitoredScriptHashesCacheExpiredApiSuccess(): void {
    $this->state->set(NesguardSriService::SCRIPT_DATA_CACHE_KEY, ['hashes' => ['old-url' => 'old-hash'], 'timestamp' => time() - 7200]); // Expired.

    $api_scripts = [
      ['url' => $this->stripeUrl, 'hash' => 'sha384-new-stripe-hash'],
      ['url' => 'https://example.com/script2.js', 'hash' => 'sha384-script2-hash'],
    ];
    $mock_response = $this->createMockScriptsResponse($api_scripts);
    $this->httpClient->get($this->apiBaseUrl . 'sri/client/scripts/', Argument::any())
      ->shouldBeCalledOnce()->willReturn($mock_response);

    $expected_map = [
        $this->stripeUrl => 'sha384-new-stripe-hash',
        'https://example.com/script2.js' => 'sha384-script2-hash',
    ];
    $result = $this->sriService->getMonitoredScriptHashes(FALSE); // Trigger fetch.
    $this->assertEquals($expected_map, $result);

    // Check cache update.
    $new_cache = $this->state->get(NesguardSriService::SCRIPT_DATA_CACHE_KEY);
    $this->assertEquals($expected_map, $new_cache['hashes']);
  }

  /**
   * @covers ::getMonitoredScriptHashes
   */
  public function testGetMonitoredScriptHashesApiReturnsEmpty(): void {
    $mock_response = $this->createMockScriptsResponse([]); // Empty scripts array.
    $this->httpClient->get(Argument::any(), Argument::any())->shouldBeCalledOnce()->willReturn($mock_response);

    $result = $this->sriService->getMonitoredScriptHashes(TRUE); // Bypass cache.
    $this->assertEquals([], $result);
    // Ensure cache is updated with empty array.
    $new_cache = $this->state->get(NesguardSriService::SCRIPT_DATA_CACHE_KEY);
    $this->assertEquals([], $new_cache['hashes']);
  }

  /**
   * @covers ::getMonitoredScriptHashes
   */
  public function testGetMonitoredScriptHashesApiFailsReturnsStaleCache(): void {
    $stale_map = ['stale-url' => 'stale-hash'];
    $this->state->set(NesguardSriService::SCRIPT_DATA_CACHE_KEY, ['hashes' => $stale_map, 'timestamp' => time() - 7200]); // Expired.

    $mock_response = $this->createMockScriptsResponse([], 500, FALSE); // API error.
    $this->httpClient->get(Argument::any(), Argument::any())->shouldBeCalledOnce()->willReturn($mock_response);

    // Should return the stale cache data.
    $result = $this->sriService->getMonitoredScriptHashes(FALSE);
    $this->assertEquals($stale_map, $result);
    $this->logger->error(Argument::containingString('Failed to fetch SRI scripts/hashes'), Argument::any())->shouldHaveBeenCalled();
  }

   /**
   * @covers ::getMonitoredScriptHashes
   */
  public function testGetMonitoredScriptHashesApi429ReturnsStaleCache(): void {
    $stale_map = ['stale-url' => 'stale-hash'];
    $this->state->set(NesguardSriService::SCRIPT_DATA_CACHE_KEY, ['hashes' => $stale_map, 'timestamp' => time() - 100]); // Cache valid but bypass = true later.

    $mock_response = new Response(429, [], 'Too Many Requests');
    $this->httpClient->get(Argument::any(), Argument::any())->shouldBeCalledOnce()->willReturn($mock_response);

    // Bypass cache to trigger API call.
    $result = $this->sriService->getMonitoredScriptHashes(TRUE);
    // Should return NULL because cache was bypassed and API failed.
    // Correction: should return stale cache data according to current logic. Let's test that.
    // Need to prime cache to test return value.
     $this->state->set(NesguardSriService::SCRIPT_DATA_CACHE_KEY, ['hashes' => $stale_map, 'timestamp' => time() - 7200]); // Ensure stale cache exists.
     $result = $this->sriService->getMonitoredScriptHashes(FALSE); // Call again without bypass. Should hit 429 -> return stale cache.
     $this->assertEquals($stale_map, $result);

    $this->logger->warning(Argument::containingString('Rate Limited (429)'), Argument::any())->shouldHaveBeenCalled();
    $this->assertGreaterThan(time(), $this->state->get(NesguardSriService::RATE_LIMIT_FETCH_KEY, 0));
  }


  /**
   * @covers ::getMonitoredScriptUrls
   */
  public function testGetMonitoredScriptUrlsSuccess(): void {
     $map = ['url1' => 'h1', 'url2' => 'h2'];
     // Need to mock getMonitoredScriptHashes to return this map.
     // Easiest way is to set the cache.
     $this->state->set(NesguardSriService::SCRIPT_DATA_CACHE_KEY, ['hashes' => $map, 'timestamp' => time() - 100]);
     $this->subscriptionService->isActive()->willReturn(TRUE);

     $urls = $this->sriService->getMonitoredScriptUrls();
     $this->assertEquals(['url1', 'url2'], $urls);
     $this->logger->warning(Argument::any(), Argument::any())->shouldNotHaveBeenCalled(); // No fallback warning.
  }

   /**
   * @covers ::getMonitoredScriptUrls
   */
  public function testGetMonitoredScriptUrlsApiReturnsEmptyWithFallback(): void {
     $this->state->set(NesguardSriService::SCRIPT_DATA_CACHE_KEY, ['hashes' => [], 'timestamp' => time() - 100]); // Empty map.
     $this->subscriptionService->isActive()->willReturn(TRUE);

     $urls = $this->sriService->getMonitoredScriptUrls();
     $this->assertEquals([$this->stripeUrl], $urls);
     $this->logger->warning(Argument::containingString('API returned 0 scripts, falling back'), Argument::any())->shouldHaveBeenCalled();
  }

  /**
   * @covers ::getMonitoredScriptUrls
   */
  public function testGetMonitoredScriptUrlsApiFailsWithFallback(): void {
     // Simulate API failure by ensuring getMonitoredScriptHashes returns NULL.
     // Clear cache and mock API error.
     $this->state->delete(NesguardSriService::SCRIPT_DATA_CACHE_KEY);
     $mock_response = $this->createMockScriptsResponse([], 500, FALSE);
     $this->httpClient->get(Argument::any(), Argument::any())->willReturn($mock_response);
     $this->subscriptionService->isActive()->willReturn(TRUE);

     $urls = $this->sriService->getMonitoredScriptUrls();
     $this->assertEquals([$this->stripeUrl], $urls);
     $this->logger->warning(Argument::containingString('API data unavailable, falling back'), Argument::any())->shouldHaveBeenCalled();
  }

  /**
   * @covers ::getMonitoredScriptUrls
   */
  public function testGetMonitoredScriptUrlsInactiveSubscription(): void {
     $this->subscriptionService->isActive()->willReturn(FALSE);
     $urls = $this->sriService->getMonitoredScriptUrls();
     $this->assertEquals([], $urls);
     $this->logger->warning(Argument::any(), Argument::any())->shouldNotHaveBeenCalled(); // No fallback warning.
  }


  /**
   * @covers ::getIntegrityHash
   */
  public function testGetIntegrityHashSuccessFromApiCache(): void {
      $map = [$this->stripeUrl => 'hash-from-api', 'url2' => 'h2'];
      $this->state->set(NesguardSriService::SCRIPT_DATA_CACHE_KEY, ['hashes' => $map, 'timestamp' => time() - 100]);

      $hash = $this->sriService->getIntegrityHash($this->stripeUrl);
      $this->assertEquals('hash-from-api', $hash);
      // Ensure fallback calculation was NOT called.
      $this->logger->warning(Argument::containingString('fallback'), Argument::any())->shouldNotHaveBeenCalled();
  }

  /**
   * @covers ::getIntegrityHash
   */
  public function testGetIntegrityHashNotMonitored(): void {
       $map = [$this->stripeUrl => 'hash-from-api'];
       $this->state->set(NesguardSriService::SCRIPT_DATA_CACHE_KEY, ['hashes' => $map, 'timestamp' => time() - 100]);

       $hash = $this->sriService->getIntegrityHash('https://example.com/other.js');
       $this->assertNull($hash);
       // Ensure fallback calculation was NOT called.
       $this->logger->warning(Argument::containingString('fallback'), Argument::any())->shouldNotHaveBeenCalled();
  }

  /**
   * @covers ::getIntegrityHash
   * @covers ::fetchAndCalculateStripeHashFallback
   */
  public function testGetIntegrityHashStripeApiMissTriggersFallbackSuccess(): void {
       $map = ['url2' => 'h2']; // Stripe missing from API response.
       $this->state->set(NesguardSriService::SCRIPT_DATA_CACHE_KEY, ['hashes' => $map, 'timestamp' => time() - 100]);
       // Ensure fallback cache is empty/stale.
       $this->state->delete(NesguardSriService::STRIPE_FALLBACK_HASH_KEY);

       // Mock direct fetch to Stripe.
       $mock_stripe_response = new Response(200, [], 'stripe-js-content');
       $this->httpClient->get(NesguardSriService::STRIPE_JS_URL, Argument::any())
            ->shouldBeCalledOnce()
            ->willReturn($mock_stripe_response);

       // Mock hash calculation - use real one via service? Or mock? Mocking is more unit-testy.
       // Let's assume calculateSha384 works and test the flow.
       $expected_fallback_hash = 'sha384-calculated-fallback-hash';
       // We can't easily mock protected methods with Prophecy. Test by checking result.
       // We know calculateSha384 needs 'stripe-js-content'.
       // $this->sriService->calculateSha384('stripe-js-content')->willReturn($expected_fallback_hash); // Cannot mock protected.

       // Call the public method.
       $hash = $this->sriService->getIntegrityHash(NesguardSriService::STRIPE_JS_URL);

       // Need to manually calculate expected hash for assertion if not mocking.
       $expected_fallback_hash = $this->sriService->calculateSha384('stripe-js-content');

       $this->assertEquals($expected_fallback_hash, $hash);
       $this->logger->warning(Argument::containingString('Stripe JS hash not found in NESGuard API data, attempting Stripe JS fallback'), Argument::any())->shouldHaveBeenCalled();
       $this->logger->info('Stripe JS fallback successful. Calculated hash: @hash', ['@hash' => $expected_fallback_hash])->shouldHaveBeenCalled();
       // Check fallback cache was set.
       $this->assertEquals($expected_fallback_hash, $this->state->get(NesguardSriService::STRIPE_FALLBACK_HASH_KEY));
  }

   /**
   * @covers ::getIntegrityHash
   * @covers ::fetchAndCalculateStripeHashFallback
   */
   public function testGetIntegrityHashStripeFallbackUsesCache(): void {
       $map = ['url2' => 'h2']; // Stripe missing.
       $this->state->set(NesguardSriService::SCRIPT_DATA_CACHE_KEY, ['hashes' => $map, 'timestamp' => time() - 100]);
       // Set valid fallback cache.
       $cached_fallback_hash = 'sha384-cached-fallback-hash';
       $this->state->set(NesguardSriService::STRIPE_FALLBACK_HASH_KEY, $cached_fallback_hash);
       $this->state->set(NesguardSriService::STRIPE_FALLBACK_TIME_KEY, time() - 60); // Cached 1 min ago.

       $hash = $this->sriService->getIntegrityHash(NesguardSriService::STRIPE_JS_URL);

       $this->assertEquals($cached_fallback_hash, $hash);
       $this->logger->warning(Argument::containingString('Stripe JS hash not found'), Argument::any())->shouldHaveBeenCalled();
       // Ensure direct fetch to Stripe was NOT called.
       $this->httpClient->get(NesguardSriService::STRIPE_JS_URL, Argument::any())->shouldNotHaveBeenCalled();
   }

   /**
   * @covers ::getIntegrityHash
   * @covers ::fetchAndCalculateStripeHashFallback
   */
  public function testGetIntegrityHashStripeFallbackFails(): void {
       $map = ['url2' => 'h2']; // Stripe missing.
       $this->state->set(NesguardSriService::SCRIPT_DATA_CACHE_KEY, ['hashes' => $map, 'timestamp' => time() - 100]);
       // Ensure fallback cache is empty/stale.
       $this->state->delete(NesguardSriService::STRIPE_FALLBACK_HASH_KEY);

       // Mock direct fetch to Stripe failing.
       $mock_stripe_response = new Response(503, [], 'Service Unavailable');
       $this->httpClient->get(NesguardSriService::STRIPE_JS_URL, Argument::any())
            ->shouldBeCalledTimes(1 + NesguardSriService::FALLBACK_FETCH_RETRIES) // Initial + retries.
            ->willReturn($mock_stripe_response);

       $hash = $this->sriService->getIntegrityHash(NesguardSriService::STRIPE_JS_URL);

       $this->assertNull($hash);
       $this->logger->warning(Argument::containingString('Stripe JS hash not found'), Argument::any())->shouldHaveBeenCalled();
       $this->logger->error(Argument::containingString('Stripe JS fallback failed: Could not fetch content'), Argument::any())->shouldHaveBeenCalled();
  }

  /** @covers ::clearCache */
  public function testClearCache(): void { /* ... Same as subscription service test ... */ }

}