<?php

namespace Drupal\Tests\nesguard_drupal\Unit\Service;

use Drupal\Core\Cache\MemoryCache\MemoryBackend;
use Drupal\Core\Config\Config;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\DependencyInjection\ContainerBuilder;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\State\State;
use Drupal\Tests\UnitTestCase;
use Drupal\nesguard_drupal\Service\NesguardAuthService;
use Drupal\nesguard_drupal\Service\NesguardSubscriptionService;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Psr7\Response;
use GuzzleHttp\Exception\RequestException;
use GuzzleHttp\Psr7\Request;
use Prophecy\Argument;
use Prophecy\PhpUnit\ProphecyTrait;

/**
 * @coversDefaultClass \Drupal\nesguard_drupal\Service\NesguardSubscriptionService
 * @group nesguard_drupal
 */
class NesguardSubscriptionServiceTest extends UnitTestCase {

  use ProphecyTrait;

  protected $subscriptionService;
  protected $configFactory;
  protected $config;
  protected $state;
  protected $logger;
  protected $httpClient;
  protected $authService;
  protected $apiBaseUrl = NesguardSubscriptionService::API_BASE_URL;

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->config = $this->prophesize(Config::class);
    $this->config->get('enabled')->willReturn(TRUE);
    $this->config->get('agency_id')->willReturn('TEST-AGENCY');
    $this->config->get('debug_mode')->willReturn(FALSE);

    $this->configFactory = $this->prophesize(ConfigFactoryInterface::class);
    $this->configFactory->get('nesguard_drupal.settings')->willReturn($this->config->reveal());

    $this->state = new State(new MemoryBackend('state'));

    $this->logger = $this->prophesize(LoggerChannelInterface::class);

    $this->httpClient = $this->prophesize(ClientInterface::class);

    $this->authService = $this->prophesize(NesguardAuthService::class);
    // Default: Auth service provides credentials and a token.
    $this->authService->hasCredentials()->willReturn(TRUE);
    $this->authService->getAccessToken()->willReturn('dummy-access-token');

    $this->subscriptionService = new NesguardSubscriptionService(
      $this->configFactory->reveal(),
      $this->state,
      $this->logger->reveal(),
      $this->httpClient->reveal(),
      $this->authService->reveal()
    );

    $container = new ContainerBuilder();
    $logger_factory = $this->prophesize(LoggerChannelFactoryInterface::class);
    $logger_factory->get('nesguard_drupal')->willReturn($this->logger->reveal());
    $container->set('logger.factory', $logger_factory->reveal());
    \Drupal::setContainer($container);
  }

  /**
   * @covers ::isActive
   */
  public function testIsActiveModuleDisabled(): void {
    $this->config->get('enabled')->willReturn(FALSE);
    $this->assertNull($this->subscriptionService->isActive());
  }

  /**
   * @covers ::isActive
   */
  public function testIsActiveNoCredentials(): void {
    $this->authService->hasCredentials()->willReturn(FALSE);
    $this->assertNull($this->subscriptionService->isActive());
  }

  /**
   * @covers ::isActive
   * @dataProvider providerIsActiveCache
   */
  public function testIsActiveFromCache(bool $cached_status, bool $expected): void {
    $cache_key = NesguardSubscriptionService::SUBSCRIPTION_CACHE_KEY;
    $this->state->set($cache_key, ['status' => $cached_status, 'timestamp' => time() - 100]); // Valid cache.

    $this->assertEquals($expected, $this->subscriptionService->isActive());
    $this->httpClient->get(Argument::any(), Argument::any())->shouldNotHaveBeenCalled();
  }

  public function providerIsActiveCache(): array {
    return [
      'Cached Active' => [TRUE, TRUE],
      'Cached Inactive' => [FALSE, FALSE],
    ];
  }

  /**
   * @covers ::isActive
   */
  public function testIsActiveCacheExpiredApiActive(): void {
    $cache_key = NesguardSubscriptionService::SUBSCRIPTION_CACHE_KEY;
    $this->state->set($cache_key, ['status' => FALSE, 'timestamp' => time() - 7200]); // Expired cache.

    $response_body = json_encode(['success' => TRUE, 'data' => ['status' => 'active']]);
    $mock_response = new Response(200, [], $response_body);
    $this->httpClient->get($this->apiBaseUrl . 'subscriptions/user/', Argument::any())
      ->shouldBeCalledOnce()->willReturn($mock_response);

    $this->assertTrue($this->subscriptionService->isActive());
    // Verify cache was updated.
    $new_cache = $this->state->get($cache_key);
    $this->assertTrue($new_cache['status']);
    $this->assertGreaterThan(time() - 5, $new_cache['timestamp']); // Timestamp updated recently.
  }

  /**
   * @covers ::isActive
   */
  public function testIsActiveCacheExpiredApiInactive(): void {
    $this->state->set(NesguardSubscriptionService::SUBSCRIPTION_CACHE_KEY, ['status' => TRUE, 'timestamp' => time() - 7200]); // Expired.

    $response_body = json_encode(['success' => TRUE, 'data' => ['status' => 'inactive']]);
    $mock_response = new Response(200, [], $response_body);
    $this->httpClient->get(Argument::any(), Argument::any())->shouldBeCalledOnce()->willReturn($mock_response);

    $this->assertFalse($this->subscriptionService->isActive());
    $new_cache = $this->state->get(NesguardSubscriptionService::SUBSCRIPTION_CACHE_KEY);
    $this->assertFalse($new_cache['status']);
  }

   /**
   * @covers ::isActive
   */
  public function testIsActiveCacheExpiredApiNotFound404(): void {
    $this->state->set(NesguardSubscriptionService::SUBSCRIPTION_CACHE_KEY, ['status' => TRUE, 'timestamp' => time() - 7200]); // Expired.

    $mock_response = new Response(404, [], json_encode(['success' => false, 'error' => ['code' => 'NOT_FOUND']]));
    $this->httpClient->get(Argument::any(), Argument::any())->shouldBeCalledOnce()->willReturn($mock_response);

    $this->assertFalse($this->subscriptionService->isActive()); // 404 treated as inactive.
    $new_cache = $this->state->get(NesguardSubscriptionService::SUBSCRIPTION_CACHE_KEY);
    $this->assertFalse($new_cache['status']); // Cache should store false.
    $this->logger->info(Argument::containingString('No active subscription found'), Argument::any())->shouldHaveBeenCalled();
  }

  /**
   * @covers ::isActive
   */
  public function testIsActiveAuthFails(): void {
    $this->authService->getAccessToken()->willReturn(NULL); // Simulate auth failure.

    $this->assertNull($this->subscriptionService->isActive());
    $this->logger->warning('Cannot check subscription status: Authentication failed.')->shouldHaveBeenCalled();
    $this->httpClient->get(Argument::any(), Argument::any())->shouldNotHaveBeenCalled();
  }

  /**
   * @covers ::isActive
   */
  public function testIsActiveRateLimited(): void {
    $limit_key = NesguardSubscriptionService::RATE_LIMIT_CHECK_KEY;
    $this->state->set(NesguardSubscriptionService::SUBSCRIPTION_CACHE_KEY, ['status' => TRUE, 'timestamp' => time() - 100]); // Valid cache.
    $this->state->set(NesguardSubscriptionService::SUBSCRIPTION_CACHE_KEY, ['status' => TRUE, 'timestamp' => time() - 100]); // Valid cache.

    // Should return cached value.
    $this->assertTrue($this->subscriptionService->isActive());
    $this->logger->warning('Subscription status check skipped due to rate limit.')->shouldHaveBeenCalled();
    $this->httpClient->get(Argument::any(), Argument::any())->shouldNotHaveBeenCalled();
  }

  /**
   * @covers ::isActive
   */
  public function testIsActiveApi429(): void {
    $limit_key = NesguardSubscriptionService::RATE_LIMIT_CHECK_KEY;
    // No cache or expired cache.
    $this->state->delete(NesguardSubscriptionService::SUBSCRIPTION_CACHE_KEY);

    $mock_response = new Response(429, [], 'Too Many Requests');
    $this->httpClient->get(Argument::any(), Argument::any())->shouldBeCalledOnce()->willReturn($mock_response);

    // Should return NULL as cache was empty, and set rate limit.
    $this->assertNull($this->subscriptionService->isActive());
    $this->logger->warning(Argument::containingString('Rate limit set for action'), Argument::containingString($limit_key))->shouldHaveBeenCalled();
    $this->assertGreaterThan(time(), $this->state->get($limit_key, 0));
  }

  /**
   * @covers ::isActive
   */
  public function testIsActiveApi500(): void {
    // No cache or expired cache.

    $mock_response = new Response(500, [], 'Server Error');
    $this->httpClient->get(Argument::any(), Argument::any())->shouldBeCalledOnce()->willReturn($mock_response);

    $this->assertNull($this->subscriptionService->isActive());
    $this->logger->error(Argument::containingString('Failed to check subscription status'), Argument::any())->shouldHaveBeenCalled();
    // Ensure cache was not set on failure.
    $this->assertNull($this->state->get(NesguardSubscriptionService::SUBSCRIPTION_CACHE_KEY));
  }

  /**
   * @covers ::clearCache
   */
  public function testClearCache(): void {
    $cache_key = NesguardSubscriptionService::SUBSCRIPTION_CACHE_KEY;
    $this->state->set($cache_key, ['status' => TRUE, 'timestamp' => time()]);
    $this->subscriptionService->clearCache();
    $this->assertNull($this->state->get($cache_key));
    $this->logger->info('Cleared cached subscription status.')->shouldHaveBeenCalled();
  }

}